// txtmus - parse.js - ABC parse
//
// Copyright (C) 2014-2024 Jean-Francois Moine
//
// This file is part of txtmus-core.
//
// txtmus-core is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// txtmus-core is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with txtmus-core.  If not, see <http://www.gnu.org/licenses/>.

    var	a_gch,		// array of parsed chord symbols / annotations
	a_dcn = [],	// array of parsed decoration names
	maps = {}	// maps object - see set_map()

// set the source references of a symbol
function set_ref(s) {
	s.fname = parse.fname
	s.istart = parse.istart
	s.iend = parse.iend
} // set_ref()

// -- %% pseudo-comment

// clef definition (%%clef, K: and V:)
function new_clef(clef_def) {
    var	s = new_sym({
		typ: C.clef,
		clef_line: 2,
		clef_type: "t",
		v: curvoice.v,
		p_v: curvoice,
		time: curvoice.time
//		dur: 0
	}),
	i = 1

	set_ref(s)

	switch (clef_def[0]) {
	case '"':
		i = clef_def.indexOf('"', 1)
		s.clef_name = clef_def.slice(1, i)
		i++
		break
	case 'a':
		if (clef_def[1] == 'u') {	// auto
			s.clef_type = "a"
			s.clef_auto = 1
			i = 4
			break
		}
		i = 4				// alto
		// fall thru
	case 'C':
		s.clef_type = "c"
		s.clef_line = 3
		break
	case 'b':				// bass
		i = 4
		// fall thru
	case 'F':
		s.clef_type = "b"
		s.clef_line = 4
		break
	case 'n':				// none
		i = 4
		s.invis = 1
		s.clef_none = 1 //true
		break
	case 't':
		if (clef_def[1] == 'e') {	// tenor
			s.clef_type = "c"
			s.clef_line = 4
			break
		}
		i = 6
	case 'G':
//		s.clef_type = "t"		// treble
		break
	case 'p':
		i = 4
	case 'P':				// perc
		s.clef_type = "p"
		s.clef_line = 3
		break
	default:
		syntax(1, "Unknown clef '$1'", clef_def)
		return //undefined
	}
	if (clef_def[i] >= '1' && clef_def[i] <= '9') {
		s.clef_line = +clef_def[i]
		i++
	}

	// handle the octave (+/-8 - ^/_8)
	delete curvoice.snd_oct
	if (clef_def[i + 1] != '8'
	 && clef_def[i + 1] != '1')
		return s			// done

	switch (clef_def[i]) {			// octave
	case '^':
		s.clef_oct_transp = 1
	case '+':
		s.clef_octave = clef_def[i + 1] == '8' ? 7 : 14
		if (!s.clef_oct_transp)		// MIDI higher octave
			curvoice.snd_oct = clef_def[i + 1] == 8 ? 12 : 24
		break
	case '_':
		s.clef_oct_transp = 1
	case '-':
		s.clef_octave = clef_def[i + 1] == '8' ? -7 : -14
		if (!s.clef_oct_transp)		// MIDI lower octave
			curvoice.snd_oct = clef_def[i + 1] == 8 ? -12 : -24
		break
	}
	return s
} // new_clef()

// convert an interval to a base-40 interval
function get_interval(parm, score) {
    var	i, note, ia, pit1, pit2

	tm.reap.lastIndex = 0
	ia = tm.reap.exec(parm)
	if (!ia) {
		syntax(1, errs.bad_transp)
		return
	}
	note = get_acc_pit(ia)
	pit1 = tm.pab40(note.pit, note.acc)

	ia = tm.reap.exec(parm)
	if (!ia) {
		if (!score) {
			syntax(1, errs.bad_transp)
			return
		}
		pit2 = 242			// 'c' (C5)
	} else {
		note = get_acc_pit(ia)
		pit2 = tm.pab40(note.pit, note.acc)
	}
	return pit2 - pit1
} // get_interval()

// transpose a note of the score
// Return the transposed real accidental
function nt_trans(nt,
		  a) {			// real accidental
    var	ak, an, d, b40, n

	if (typeof a == "object") {		// if microtonal accidental
		n = a[0]			// numerator
		d = a[1]			// denominator
		a = n > 0 ? 1 : -1		// base accidental for transpose
	}

	b40 = tm.pab40(nt.pit, a)
		+ curvoice.tr_sco		// base-40 transposition

	nt.pit = tm.b40p(b40)			// new pitch
	an = tm.b40a(b40)			// new accidental

	if (!d) {				// if not a microtonal accidental
		if (an == -3)			// if triple sharp/flat
			return an
		a = an
		if (nt.acc) {			// if old accidental
			if (!a)
				a = 3		// required natural
		} else {
			if (!curvoice.ckey.k_none) // if normal key
				a = 0		// no accidental
		}
		nt.acc = a
		return an
	}

	// set the microtonal accidental after transposition
	switch (an) {
	case -2:
		if (n > 0)
			n -= d * 2
		else
			n -= d
		break
	case -1:
		if (n > 0)
			n -= d
		break
	case 0:
	case 3:
	if (n > 0)
			n -= d
		else
			n += d
		break
	case 1:
		if (n < 0)
			n += d
		break
	case 2:
		if (n < 0)
			n += d * 2
		else
			n += d
		break
	}
	nt.acc = [n, d]
	return an
} // nt_trans()

// set the linebreak character
function set_linebreak(parm) {
    var	i, c

	// remove the old definition(s)
	for (i = 0; i < 128; i++) {
		if (char_tb[i] == "$")
			char_tb[i] = nil	// remove old definition
	}

	parm = parm.split(/\s+/)
	for (i = 0; i < parm.length; i++) {
		c = parm[i]
		switch (c) {
		case '!':
		case '$':
		case '*':
		case ';':
		case '?':
		case '@':
			break
		case "<none>":
			continue
		case "<EOL>":
			c = '\n'
			break
		default:
			syntax(1, "Bad value '$1' in %%linebreak - ignored", c)
			continue
		}
		char_tb[c.charCodeAt(0)] = '$'
	}
} // set_linebreak()

// set a new user character (U: or %%user)
function set_user(parm) {
    var	k, c, v,
	a = parm.match(/(.)[=\s]*(\[[A-Za-z]:.+\]|".+"|!.+!)$/)

	if (!a) {
		syntax(1, 'Lack of starting [, ! or " in U: / %%user')
		return
	}
	c = a[1]
	v = a[2]
	if (c[0] == '\\') {
		if (c[1] == 't')
			c = '\t'
		else
			c = c[1] || ' '
	}

	k = c.charCodeAt(0)
	if (k >= 128) {
		syntax(1, errs.not_ascii)
		return
	}
	switch (char_tb[k][0]) {
	case '0':			// nil
	case 'd':			// free, possible decoration
	case '#':			// ignore
	case ' ':
		break
	case '"':
	case '!':
	case '[':
		if (char_tb[k].length > 1)
			break
		// fall thru
	default:
		syntax(1, "Bad user character '$1'", c)
		return
	}
	switch (v) {
	case "!beambreak!":
		v = " "
		break
	case "!ignore!":
		v = "#"
		break
	case "!nil!":
	case "!none!":
		v = "d"
		break
	}
	char_tb[k] = v
} // set_user()

// get a stafflines value
function get_st_lines(parm) {
	if (!parm)
		return
	if (/^[\]\[|.-]+$/.test(parm))
		return parm.replace(/\]/g, '[')

    var	n = +parm
	switch (n) {
	case 0: return "..."
	case 1: return "..|"
	case 2: return ".||"
	case 3: return ".|||"
	}
	if (isNaN(n) || n < 0 || n > 16)
		return //undefined
	return "||||||||||||||||".slice(0, n)
} // get_st_lines()

// create a block symbol in the tune body
function new_block(subtype) {
    var	s = new_sym({
		typ: C.block,
		subtype: subtype
//		dur: 0
	})

	sym_link(s)
	return s
} // new_block()

// set the voice parameters
// (possible hook)
Mus.prototype.set_vp = function(a) {
    var	s, item, pos, val, clefpit, tr_p, v,
	i = 0

	while (1) {
		item = a[i++]
		if (!item)
			break
		switch (item) {
		case "clef":
			if (a[i++] != '=')
				break
			s = a[i++]		// keep last clef
			continue
		case "clefpitch":
			if (a[i++] != '=')
				break
			item = a[i++]		// (<note><octave>)
			if (item) {
				val = C.ntb.indexOf(item[0])
				if (val >= 0) {
					switch (item[1]) {
					case "'":
						val += 7
						break
					case ',':
						val -= 7
						if (item[2] == ',')
							val -= 7
						break
					}
					clefpit = 4 - val	// 4 = 'G'
					continue
				}
			}
//			syntax(1, errs.bad_val, item)
			break
		case "octave":
			if (a[i++] != '=')
				break
			val = +a[i++]
			if (isNaN(val))
//				syntax(1, errs.bad_val, item)
				break
//			else
				curvoice.octave = val
			continue
		case "cue":
			if (a[i++] != '=')
				break
			curvoice.scale = a[i++] == 'on' ? .7 : 1
			continue
		case "instrument":
			if (a[i++] != '=')
				break

			// instrument=M/N => score=MN and sound=cN
			// (instrument=M == instrument=M/M)
			item = a[i++]
			val = item.indexOf('/')
			if (val < 0) {			// written pitch
				val = get_interval('c' + item)
				if (val == undefined)
					break
				curvoice.sound = val
				tr_p = 1 //true
				val = 0
			} else {			// sounding pitch
				val = get_interval('c' + item.slice(val + 1))
				if (val == undefined)
					break
				curvoice.sound = val
				tr_p = 1 //true
				val = get_interval(item.replace('/', ''))
				if (val == undefined)
					break
			}
			if (cfmt.sound == "concert"
			 && (curvoice.sound + val) % 40 == 0)
				curvoice.score = val	// octave transposing
			else
				curvoice.score = cfmt.sound ? curvoice.sound : val
			continue
		case "map":			// %%voicemap
			if (a[i++] != '=')
				break
			curvoice.map = a[i++]
			continue
		case "name":
		case "nm":
			if (a[i++] != '=')
				break
			curvoice.nm = a[i++]
			if (curvoice.nm[0] == '"')
				curvoice.nm = cnv_escape(curvoice.nm.slice(1, -1))
			curvoice.new_name = 1 //true
			continue
		case "stem":			// compatibility
		case "pos":			// from %%pos only
			if (a[i++] != '=')
				break
			if (item == "pos")
				item = a[i++]
					.slice(1, -1)	// always inside dble quotes
					.split(/\s+/)
			else
				item = ["stm", a[i++]]
			val = tm.posval[item[1]]
			if (val == undefined) {
//				syntax(1, errs.bad_val, "%%pos")
				item = "%%pos"
				break
			}
			switch (item[2]) {
			case "align": val |= C.SL_ALIGN; break
			case "center": val |= C.SL_CENTER; break
			case "close": val |= C.SL_CLOSE; break
			}
			if (!pos)
				pos = {}
			pos[item[0]] = val
			continue
		case "scale":			// %%voicescale
			if (a[i++] != '=')
				break
			val = +a[i++]
			if (isNaN(val) || val < .5 || val > 2)
				break
//				syntax(1, errs.bad_val, "%%voicescale")
//			else
				curvoice.scale = val
			continue
		case "score":
			if (a[i++] != '=')
				break
			if (cfmt.nedo) {
				syntax(1, errs.notransp)
				continue
			}
			// score=MN
			// (score=M == score=Mc)
			item = a[i++]
			if (cfmt.sound)
				break
			val = get_interval(item, 1)
			if (val != undefined)
				curvoice.score = val
			continue
		case "shift":
			if (a[i++] != '=')
				break
			if (cfmt.nedo) {
				syntax(1, errs.notransp)
				continue
			}
			val = get_interval(a[i++])
			if (val != undefined) {
				curvoice.shift = val
				tr_p = 1 //true
			}
			continue
		case "sound":
			if (a[i++] != '=')
				break
			if (cfmt.nedo) {
				syntax(1, errs.notransp)
				continue
			}
// concert-score display: apply sound=
// sounding-score display: apply sound= only if M != c/C
// sound: apply sound=
			val = get_interval(a[i++])
			if (val == undefined)
				continue
			curvoice.sound = val
			if (cfmt.sound)
				curvoice.score = val
			tr_p = 1 //true
			continue
		case "subname":
		case "sname":
		case "snm":
			if (a[i++] != '=')
				break
			curvoice.snm = a[i++]
			if (curvoice.snm[0] == '"')
				curvoice.snm = curvoice.snm.slice(1, -1)
			continue
		case "stafflines":
			if (a[i++] != '=')
				break
			val = get_st_lines(a[i++])
			if (val == undefined)
				break
//				syntax(1, "Bad %%stafflines value")
			else if (curvoice.st != undefined)
				par_sy.staves[curvoice.st].stafflines = val
			else
				curvoice.stafflines = val
			continue
		case "staffnonote":
			if (a[i++] != '=')
				break
			val = +a[i++]
			if (isNaN(val))
				break
//				syntax(1, "Bad %%staffnonote value")
//			else
				curvoice.staffnonote = val
			continue
		case "staffscale":
			if (a[i++] != '=')
				break
			val = +a[i++]
			if (isNaN(val) || val < .3 || val > 2)
				break
//				syntax(1, "Bad %%staffscale value")
//			else
				curvoice.staffscale = val
			continue
		case "sync":				// synchronize the voices
			get_staves("sync")		// with an empty %%staves
			continue
		case "tacet":
			if (a[i++] != '=')
				break
			val = a[i++]
			curvoice.tacet = val || undefined
			continue
		case "transpose":		// (abcMIDI compatibility)
			if (a[i++] != '=')
				break
			if (cfmt.nedo) {
				syntax(1, errs.notransp)
				continue
			}
			val = get_transp(a[i++])
			if (val == undefined)
				break
//				syntax(1, errs.bad_transp)
//			} else {
				curvoice.sound = val
				if (cfmt.sound)
					curvoice.score = val
				tr_p = 1 //true
//			}
			continue
		default:
			switch (item.slice(0, 4)) {
			case "treb":
			case "bass":
			case "alto":
			case "teno":
			case "perc":
				s = item
				break
			default:
				if ("GFC".indexOf(item[0]) >= 0)
					s = item		// clef fixme ??
//				else if (item.slice(-1) == '=')
//					a.shift()
				else if (a[i] == '=')
					i += 2
				break
			}
			continue
		}
		syntax(1, errs.bad_val, item)
		break
	}
	if (pos) {
		curvoice.pos = clone(curvoice.pos)
		for (item in pos)
			if (pos.hasOwnProperty(item))
				curvoice.pos[item] = pos[item]
	}

	if (s) {
		s = new_clef(s)
		if (s) {
			if (clefpit)
				s.clefpit = clefpit
			get_clef(s)
		}
	}

	// define the sound transposition
	// the score transposition curvoice.tr_sco is defined in key_trans()
	if (tr_p) {
		tr_p = (curvoice.sound | 0) + (curvoice.shift | 0)
		if (tr_p)
			curvoice.tr_snd = tm.b40m(tr_p + 122) - 36
							// semi-tone interval
		else if (curvoice.tr_snd)
			curvoice.tr_snd = 0
	}
} // set_vp()

// set the K: / V: parameters
function set_kv_parm(a) {	// array of items
	if (!curvoice.init) {	// add the global parameters if not done yet
		curvoice.init = 1 //true
		if (info.V) {
			if (info.V[curvoice.id])
				a = info.V[curvoice.id].concat(a)
			if (info.V['*'])
				a = info.V['*'].concat(a)
		}
	}
	if (a.length)
		self.set_vp(a)
} // set_kv_parm()

// memorize the K:/V: parameters
function memo_kv_parm(vid,	// voice ID (V:) / '*' (K:/V:*)
			a) {	// array of items
//	if (!a.length)
//		return
	if (!info.V)
		info.V = {}
	if (info.V[vid])
		Array.prototype.push.apply(info.V[vid], a)
	else
		info.V[vid] = a
} // memo_kv_parm()

// K: key signature
// return the key and the voice/clef parameters
//fixme: no key if empty (no sf/acc)
function new_key(parm) {
    var	i, key_end, c, tmp, nt,
	sf = "FCGDAEB".indexOf(parm[0]) - 1,
	mode = 0,
	s = new_sym({
		typ: C.key
//		dur: 0
	})

	set_ref(s)

	// tonic
	i = 1
    if (sf < -1) {
	switch (parm[0]) {
	case 'H':				// bagpipe
		key_end = 1 //true
		if (parm[1].toLowerCase() != 'p') {
			syntax(1, "Unknown bagpipe-like key")
			break
		}
		s.k_bagpipe = parm[1]
		sf = parm[1] == 'P' ? 0 : 2
		i++

		// initialize the temperament if not done yet
		if (!cfmt.temper)
	// detune in cents for just intonation in A
	// (from https://patrickmclaurin.com/wordpress/?page_id=2420)
	//  C    ^C     D    _E     E     F    ^F     G    _A     A    _B     B
	// 15.3 -14.0  -2.0 -10.0   1.9  13.3 -16.0 -31.8 -12.0   0.0  11.4   3.8
	// but 'A' bagpipe = 480Hz => raise Math.log2(480/440)*1200 = 151
			cfmt.temper = new Float32Array([
   //	1.66, 1.37, 1.49, 1.41, 1.53, 1.63, 1.35, 1.19, 1.39, 1.51, 1.62, 1.55
   //   C    ^C     D    _E     E     F    ^F     G    _A      A     _B      B
11.62, 12.55,
     1.66, 2.37, 3.49, 0,
     1.66, 2.37, 3.49, 4.41, 5.53, 0,
		 3.49, 4.41, 5.53, 6.63, 7.35,
		       4.41, 5.53, 6.63, 7.35, 8.19, 0,
				   6.63, 7.35, 8.19, 9.39, 10.51, 0,
					       8.19, 9.39, 10.51, 11.62, 12.55, 0,
							   10.51, 11.62, 12.55,
									     1.66, 1.66
			])
		break
	case 'P':
		syntax(1, "K:P is deprecated")
		sf = 0
		s.k_drum = 1 //true
		key_end = 1 //true
		break
	case 'n':				// none
		if (parm.indexOf("none") == 0) {
			sf = 0
			s.k_none = 1 //true
			i = 4
			break
		}
		// fall thru
	default:
		s.k_map = []
		s.k_mode = 0
		return [s, info_split(parm)]
	}
    }

	if (!key_end) {
		switch (parm[i]) {
		case '#': sf += 7; i++; break
		case 'b': sf -= 7; i++; break
		}
		parm = parm.slice(i).trim()
		switch (parm.slice(0, 3).toLowerCase()) {
		default:
			if (parm[0] != 'm'
			 || parm[1] > ' ') {
				key_end = 1 //true
				break
			}
			// fall thru ('m')
		case "aeo":
		case "m":
		case "min": sf -= 3
			mode = 5
			break
		case "dor": sf -= 2
			mode = 1
			break
		case "ion":
		case "maj": break
		case "loc": sf -= 5
			mode = 6
			break
		case "lyd": sf += 1
			mode = 3
			break
		case "mix": sf -= 1
			mode = 4
			break
		case "phr": sf -= 4
			mode = 2
			break
		}
		if (!key_end)
			parm = parm.replace(/\w+\s*/, '')

		// [exp] accidentals
		if (parm.indexOf("exp ") == 0) {
			parm = parm.replace(/\w+\s*/, '')
			if (!parm)
				syntax(1, "No accidental after 'exp'")
			s.exp = 1 //true
		}
		c = parm[0]
		if (c == '^' || c == '_' || c == '=') {
			s.k_a_acc = []
			tm.ren.lastIndex = 0
			do {
				c = tm.ren.exec(parm)
				if (!c)
					break
				nt = get_acc_pit(c)
				s.k_a_acc.push(nt)
				c = parm[tm.ren.lastIndex]
				while (c <= ' ')
					c = parm[++tm.ren.lastIndex]
			} while (c == '^' || c == '_' || c == '=')
			parm = parm.slice(tm.ren.lastIndex).trim()
		} else if (s.exp && parm.indexOf("none") == 0) {
			sf = 0
			parm = parm.replace(/\w+\s*/, '')
		}
	}

	if (sf < -7 || sf > 7) {
		syntax(1, "Key with double sharps/flats")
		if (sf > 7)
			sf -= 12
		else
			sf += 12
	}
	s.k_sf = sf

	// set the map of the notes with accidentals
	s.k_map = s.k_bagpipe && !sf
		? tm.keys[9]			// implicit F# and C#
		: tm.keys[s.k_sf + 7]
	if (s.k_a_acc) {
		s.k_map = new Int8Array(s.k_map)
		i = s.k_a_acc.length
		while (--i >= 0) {
			nt = s.k_a_acc[i]
			s.k_map[(nt.pit + 37) % 7] = nt.acc
		}
	}
	s.k_mode = mode

	// key note as base-40
	s.k_b40 = [1,24,7,30,13,36,19, 2 ,25,8,31,14,37,20,3][sf + 7]

	return [s, info_split(parm)]
} // new_key()

// M: meter
function new_meter(p) {
    var	p_v, val, v, wmeasure, in_parenth, j, m2, s2,
	s = new_sym({
		typ: C.meter,
//		dur: 0,
		a_meter: []
	}),
	meter = {},
	m1 = 0,
	i = 0

	set_ref(s)

	if (p.indexOf("none") == 0) {
		i = 4				// no meter
		wmeasure = 1	// simplify measure numbering and C.mrest conversion
	} else {
		wmeasure = 0
		while (i < p.length) {
			if (p[i] == '=')
				break
			switch (p[i]) {
			case 'C':
				meter.top = p[i++]
				if (!m1) {
					m1 = 4
					m2 = 4
				}
				break
			case 'c':
			case 'o':
				meter.top = p[i++]
				if (!m1) {
					if (p[i - 1] == 'c') {
						m1 = 2
						m2 = 4	// c = 2/4
					} else {
						m1 = 3
						m2 = 4	// o = 3/4
					}
					switch (p[i]) {
					case '|':
						m2 /= 2	// c| = 2/2, o| = 3/2
						break
					case '.':
						m1 *= 3
						m2 *= 2	// c. = 6/8, o. = 9/8
						break
					}
				}
				break
			case '.':
			case '|':
				m1 = 0
				meter.top = p[i++]
				break
			case '(':
				if (p[i + 1] == '(') {	// "M:5/4 ((2+3)/4)"
					in_parenth = 1
					meter.top = p[i++]
					s.a_meter.push(meter)
					meter = {}
				}
				j = i + 1
				while (j < p.length) {
					if (p[j] == ')' || p[j] == '/')
						break
					j++
				}
				if (p[j] == ')' && p[j + 1] == '/') { // "M:5/4 (2+3)/4"
					i++		// remove the parenthesis
					continue
				}			// "M:5 (2+3)"
				// fall thru
			case ')':
				in_parenth = p[i] == '('
				meter.top = p[i++]
				s.a_meter.push(meter)
				meter = {}
				continue
			default:
				if (p[i] <= '0' || p[i] > '9') {
					syntax(1, "Bad char '$1' in M:", p[i])
					return
				}
				m2 = 2			// default when no bottom value
				meter.top = p[i++]
				for (;;) {
					while (p[i] >= '0' && p[i] <= '9')
						meter.top += p[i++]
					if (p[i] == ')') {
						if (p[i + 1] != '/')
							break
						i++
					}
					if (p[i] == '/') {
						i++
						if (p[i] <= '0' || p[i] > '9') {
							syntax(1, "Bad char '$1' in M:",
								p[i])
							return
						}
						meter.bot = p[i++]
						while (p[i] >= '0' && p[i] <= '9')
							meter.bot += p[i++]
						break
					}
					if (p[i] > ' ' && p[i] != '+')
						break
					if (i >= p.length
					 || p[i + 1] == '(')	// "M:5 (2/4+3/4)"
						break
					meter.top += p[i++]
				}
				m1 = eval(meter.top)
				break
			}
			if (!in_parenth) {
				if (meter.bot)
					m2 = +meter.bot
				wmeasure += m1 * C.BLEN / m2
			}
			s.a_meter.push(meter)
			meter = {}
			while (p[i] <= ' ')
				i++
			if (p[i] == '+') {
				meter.top = p[i++]
				s.a_meter.push(meter)
				meter = {}
			}
		}
	}
	if (p[i] == '=') {
		val = p.substring(++i).match(/^(\d+)\/(\d+)$/)
		if (!val) {
			syntax(1, "Bad duration '$1' in M:", p.substring(i))
			return
		}
		s.meas_dur = p			// (for toabc or tomnx)
		wmeasure = C.BLEN * val[1] / val[2]
	}
	if (!wmeasure) {
		syntax(1, errs.bad_val, 'M:')
		return
	}
	s.wmeasure = wmeasure

	if (cfmt.writefields.indexOf('M') < 0)
		s.a_meter = []

	if (parse.state != 3) {			// global or in tune header
		info.M = p
//--fixme: to remove
//		if (!info.M)
//			info.M = []
//		info.M[0] = {}
//		strnew(info.M[0], p, "history")
		glovar.meter = s
		if (parse.state) {

			// in the tune header, change the unit note length
			if (!glovar.ulen) {
				if (wmeasure <= 1
				 || wmeasure >= C.BLEN * 3 / 4)
					glovar.ulen = C.BLEN / 8
				else
					glovar.ulen = C.BLEN / 16
			}
			for (v = 0; v < voice_tb.length; v++) {
				voice_tb[v].meter = s
				voice_tb[v].wmeasure = wmeasure
			}
		}
	} else {
		curvoice.wmeasure = wmeasure
		if (is_voice_sig())
			curvoice.meter = s
		else
			sym_link(s)

		// set the meter of the overlay voices
		for (p_v = curvoice.voice_down; p_v; p_v = p_v.voice_down)
			p_v.wmeasure = wmeasure
	}
} // new_meter()

// link P: or Q:
function link_pq(s, text) {
    var	p_v, s2

	if (curvoice.v == par_sy.top_voice) {
		sym_link(s)
	} else if (voice_tb[par_sy.top_voice].time == s.time) {
		p_v = curvoice
		curvoice = voice_tb[par_sy.top_voice]
		sym_link(s)
		curvoice = p_v
	} else if (voice_tb[par_sy.top_voice].time > s.time) {
		p_v = voice_tb[par_sy.top_voice]
		for (s2 = p_v.sym; ; s2 = s2.next) {
			if (s2.time >= s.time) {
				set_ref(s)
				s.fmt = cfmt
				s.next = s2
				s.prev = s2.prev
				if (s2.prev)
					s.prev.next = s
				else
					p_v.sym = s
				s2.prev = s
				s.v = s2.v
				s.p_v = p_v
				s.st = p_v.st
				break
			}
		}
	} else {
		set_ref(s)
		s.fmt = cfmt
		if (!parse.pq_d)
			parse.pq_d = []
		parse.pq_d.push(s)		// delayed insertion
	}
	if (!parse.pq)
		parse.pq = {}
	parse.pq[text] = s.time
} // link_pq()

// Q: tempo
function new_tempo(text) {
    var	i, c, d, nd,
	txt = text,			// (for info.Q)
	s = new_sym({
		typ: C.tempo
//		dur: 0
	})

	// get a note duration
	function get_nd(p) {
	    var	n, d,
		nd = p.match(/(\d+)\/(\d+)/)

		if (nd) {
			d = +nd[2]
			if (d && !isNaN(d) && !(d & (d - 1))) {
				n = +nd[1]
				if (!isNaN(n))
					return C.BLEN * n / d
			}
		}
		syntax(1, "Invalid note duration $1", c)
	} // get_nd()

	set_ref(s)

	if (cfmt.writefields.indexOf('Q') < 0)
		s.invis = 1 //true		// don't display

	// string before
	if (text[0] == '"') {
		c = text.match(/"([^"]*)"/)		// "
		if (!c) {
			syntax(1, "Unterminated string in Q:")
			return
		}
		s.tempo_str1 = c[1]
		text = text.slice(c[0].length).replace(/^\s+/,'')
	}

	// string after
	if (text.slice(-1) == '"') {
		i = text.indexOf('"')
		s.tempo_str2 = text.slice(i + 1, -1)
		text = text.slice(0, i).replace(/\s+$/,'')
	}

	// beat
	i = text.indexOf('=')
	if (i > 0) {
		d = text.slice(0, i).split(/\s+/)
		text = text.slice(i + 1).replace(/^\s+/,'')
		while (1) {
			c = d.shift()
			if (!c)
				break
			nd = get_nd(c)
			if (!nd)
				return
			if (!s.tempo_notes)
				s.tempo_notes = []
			s.tempo_notes.push(nd)
		}

		// tempo value
		if (text.slice(0, 4) == "ca. ") {
			s.tempo_ca = 'ca. '
			text = text.slice(4)
		}
		i = text.indexOf('/')
		if (i > 0) {
			nd = get_nd(text)
			if (!nd)
				return
			s.new_beat = nd
		} else {
			s.tempo = +text
			if (!s.tempo || isNaN(s.tempo)) {
				syntax(1, "Bad tempo value")
				return
			}
		}
	}
	tempo_build(s)

	if (parse.state < 2			// if in tune header
	 || (!curvoice.time && !glovar.tempo)) {
		info.Q = txt
		glovar.tempo = s
		return
	}
	if (!glovar.tempo)
		syntax(0, "No previous tempo")
	s.time = curvoice.time
	text = 'Q' + s.time
	if (parse.pq
	 && parse.pq[text] == s.time)
		return				// already seen
	link_pq(s, text)
} // new_tempo()

// treat the information fields which may embedded
function do_info(info_type, text) {
    var	s, d1, d2, a, vid, tim, v, p_v

	// skip this line if the current voice is ignored
	// but keep the time related definitions
	if (curvoice && curvoice.ignore) {
		switch (info_type) {
		default:
			return
		case 'P':
		case 'Q':
		case 'V':
			break
		}
	}

	switch (info_type) {

	// info fields in any state
	case 'I':
		self.do_pscom(text)
		break
	case 'L':
		a = text.match(/^1\/(\d+)(=(\d+)\/(\d+))?$/)
		if (a) {
			d1 = +a[1]
			if (!d1 || (d1 & (d1 - 1)) != 0)
				break
			d1 = C.BLEN / d1
			if (a[2]) {		// if '='
				d2 = +a[4]
				d2 = d2 ? +a[3] / d2 * C.BLEN : 0
			} else {
				d2 = d1
			}
		} else if (text == "auto") {
			d1 = d2 = -1
		}
		if (!d2) {
			syntax(1, "Bad L: value")
			break
		}
		if (parse.state <= 1) {
			glovar.ulen = d1
		} else {
			curvoice.ulen = d1
			curvoice.dur_fact = d2 / d1
		}
		break
	case 'M':
		new_meter(text)
		break
	case 'U':
		set_user(text)
		break

	// fields in tune header or tune body
	case 'P':			// (in tune body)
		s = {
			typ: C.part,
			text: text,
			time: curvoice.time
		}
		tim = parse.pq && parse.pq[text] // time of previous P: with same text
		if (tim == s.time)
			break				// already seen
		if (tim != null) {
			syntax(1, "Misplaced P:")	// different dates
			break
		}

		if (cfmt.writefields.indexOf('P') < 0)
			s.invis = 1 //true
		link_pq(s, text)
		break
	case 'Q':
		if (!parse.state)
			break
		new_tempo(text)
		break
	case 'V':
		get_voice(text)
		if (parse.state == 3)
			curvoice.ignore = !par_sy.voices[curvoice.v]
		break

	// key signature at end of tune header or in tune body
	case 'K':
		if (!parse.state)	// ignore if in file header
			break
		get_key(text)
		break

	// info in any state
	case 'r':
		if (!user.keep_remark
		 || parse.state != 3)
			break
		s = new_sym({
			typ: C.rem,
			text: text
//			dur: 0
		})
		sym_link(s)
		break
	default:
		syntax(0, "'$1:' line ignored", info_type)
		break
	}
} // do_info()

// music line parsing functions

// -- adjust the duration and time of symbols in a measure when L:auto --
function adjust_dur(s) {
    var	s2, time, auto_time, i, fac

	// search the start of the measure
	s2 = curvoice.last_sym
	if (!s2)
		return

	// the bar time is correct if there are multi-rests
	if (s2.typ == C.mrest
	 || s2.typ == C.bar)			// in second voice
		return
	while (s2.typ != C.bar && s2.prev)
		s2 = s2.prev
	time = s2.time
	auto_time = curvoice.time - time
	fac = curvoice.wmeasure / auto_time

	if (fac == 1)
		return				// already good duration

	for ( ; s2; s2 = s2.next) {
		s2.time = time
		if (!s2.dur || s2.grace)
			continue
		s2.dur *= fac
		s2.dur_orig *= fac
		time += s2.dur
		if (s2.typ != C.note && s2.typ != C.rest)
			continue
		for (i = 0; i <= s2.nhd; i++)
			s2.notes[i].dur *= fac
	}
	curvoice.time = s.time = time
} // adjust_dur()

// -- parse a bar --
function new_bar(grace) {
    var	s, s2, c, btyp

	tm.reb.lastIndex = parse.li - 1
	c = tm.reb.exec(parse.mb)	// => [1]=bar_type, [2] ="xx" or [0-9]*
	if (!c)
		return
	if (c[0] == '[' && !c[2])
		return				// chord or embedded info
	parse.li = tm.reb.lastIndex
	if (grace) {
		syntax(1, errs.bar_grace)
		return
	}

	btyp = c[1]				// bar type
	parse.iend = parse.li
	s = new_sym({
		typ: C.bar
//		dur: 0
	})

	if (vover && vover.bar)			// end of voice overlay
		get_vover('|')

	if (glovar.new_nbar) {			// %%setbarnb
		s.bar_num = glovar.new_nbar
		glovar.new_nbar = 0
	}
	if (btyp[0] == ':') {
		if (btyp == ':') {		// ":" alone
			btyp = '|'
			s.bar_dotted = 1 //true
		} else {
			s.rbstop = 2		// right repeat with bracket
		}
	}

	// if the last element is '[',
	// it may start a chord or an embedded header
	if (!c[2]) {
		if (btyp.slice(-1) == '[') {
			btyp = btyp.slice(0, -1)
			parse.li--		// let the '[' to start the next item
//			c = '['
		}
	} else {				// repeat variant
		if (c[2][0] == '"') {
			if (btyp.slice(-1) == '[') {
				s.text = c[2].slice(1, -1)
			} else {
				parse.li -= c[2].length // (chord symbol / annotation)
			}
		} else {
			s.text = c[2]
		}
	}

	// ']' as the first character indicates a repeat bar stop
	if (btyp[0] == ']') {
		s.rbstop = 2			// with bracket
		if (btyp.length != 1)
			btyp = btyp.slice(1)
		else
			s.invis = 1
	}

	// if bar + variant, remove the ending '['
	if (s.text
	 && btyp.slice(-1) == '['
	 && btyp != '[')
		btyp = btyp.slice(0, -1)

	// there cannot be variants on a left repeat bar
	if (btyp.slice(-1) == ':') {		// left repeat
		s.rbstop = 1			// end the bracket
		if (s.text) {
			syntax(1, "Variant ending on a left repeat bar")
			delete s.text
		}
		curvoice.tie_s_rep = null	// no tie anymore on new variant
	}

	// handle the accidentals (ties and repeat)
	if (s.text) {
		s.rbstart = s.rbstop = 2
		if (s.text[0] == '1') {
			curvoice.tie_s_rep = curvoice.tie_s
			if (curvoice.acc_tie)
				curvoice.acc_tie_rep = curvoice.acc_tie.slice()
			else if (curvoice.acc_tie_rep)
				curvoice.acc_tie_rep = null
		} else {
			curvoice.tie_s = curvoice.tie_s_rep
			if (curvoice.acc_tie_rep)
				curvoice.acc_tie = curvoice.acc_tie_rep.slice()
		}
		if (curvoice.norepbra
		 && !curvoice.second)
			s.norepbra = 1 //true
	}

	if (curvoice.ulen < 0)			// L:auto
		adjust_dur(s)

	// merge ":| |:" into "::" and other cases
	if ((btyp == "[" || btyp == "|:")
	 && !curvoice.eoln
	 && !a_gch && !a_dcn.length) {		// no annotation nor decoration
		s2 = curvoice.last_sym

		// if the previous symbol is also a bar
		if (s2 && s2.bar_type) {
//		 && !s2.a_gch && !s2.a_dd
//		 && !s.a_gch && !s.a_dd) {

			// remove the invisible repeat bars
			// when no shift is needed
			if ((btyp == "["
			  && !s2.text)
			 || s.norepbra) {
				if (s.text)
					s2.text = s.text
//				if (s.a_gch)
//					s2.a_gch = s.a_gch
				if (s.norepbra)
					s2.norepbra = 1 //true
				if (s.rbstart)
					s2.rbstart = s.rbstart
				if (s.rbstop)
					s2.rbstop = s.rbstop
//--fixme: pb when on next line and empty staff above
				return
			}

			// merge back-to-back repeat bars
			if (btyp == "|:" && s2.bar_type == ":|") {
				s2.bar_type = "::"	// :| + |: => ::
//				s2.rbstop = 2
				return
			}
		}
	}

	// set some flags
	switch (btyp) {
	case "[":
	case "[]":
	case "[|]":
		s.invis = 1
		btyp = s.rbstart ? "[" : "[]"
		break
	case ":|:":
	case ":||:":
		btyp = "::"
		break
	case "||":
		if (cfmt["abc-version"] >= "2.2")
			break
		// fall thru - play repeat on double bar when old ABC version
	case "[|":
	case "|]":
		s.rbstop = 2
		break
	}
	s.bar_type = btyp
	if (!curvoice.lyric_restart)
		curvoice.lyric_restart = s
	if (!curvoice.sym_restart)
		curvoice.sym_restart = s

	sym_link(s)

	s.st = curvoice.st			// original staff

	// possibly shift the volta bracket if not on the first staff
	if (s.text && s.st > 0 && !s.norepbra
	 && !(par_sy.staves[s.st - 1].flags & C.STOP_BAR)
	 && btyp != '[')
		s.xsh = 4			// volta shift

	// set the annotations and the decorations
//--fixme: what if annot/deco on removed bars ([n/|:)?
	if (a_gch)
		csan_add(s)
	if (a_dcn.length)
		deco_cnv(s)

	if (!s.bar_dotted && !s.invis)
		curvoice.acc = []		// no accidental anymore
} // new_bar()

// split an info string
function info_split(text) {
	if (text) {
	    var	a = text.match(/=|"[^"]*"|[^\s"=]+/g)
		if (a)
			return a
//fixme: bad error text
		syntax(1, "Unterminated string")
	}
	return []
} // info_split()

// parse a duration returning [numerator, denominator]
function parse_dur(c) {
    var r, n

	if (!c)
		return [1, 1]
	r = c.split('/')
	n = r[0] || 1
	if (r[1])
		return [+n, +r[1]]
	return [+n, 1 << --r.length]
} // parse_dur()

// parse the note accidental and pitch
// @ia is an array
//	1: accidental
//	2: microtone fraction
//	3: note/rest pitch/type
//	4: octave
function get_acc_pit(ia) {
    var	note, acc, pit, d, n, i, c

	// optional accidental
	c = ia[1]
	if (c[0] == '^')
		acc = c[1] == '^' ? 2 : 1
	else if (c[0] == '=')
		acc = 3
	else if (c[0] == '_')
		acc = c[1] == '_' ? -2 : -1

	// handle the microtone accidental fraction
	d = ia[2]
	if (d) {
		d = parse_dur(d)
		n = d[0]
		d = d[1]
		if (acc < 0)
			n = -n
		if (cfmt.nedo && d == 1) {
			n *= 12
			d = cfmt.nedo
		}
		acc = [n, d]
	}

	// get the pitch
	pit = C.ntb.indexOf(ia[3]) - 2		// pit 0 is E4

	// octave
	c = ia[4]
	if (c) {
		i = 0
		while (c[i] == "'") {
			pit += 7
			i++
		}
		while (c[i] == ",") {
			pit -= 7
			i++
		}
	}
	note = {
		pit: pit,
		shhd: 0,
		shac: 0
	}
	if (acc)
		note.acc = acc
	return note
} // get_acc_pit()

// -- parse note or rest with pitch and length --
// ia is an array
//	1: accidental
//	2: microtone fraction
//	3: note/rest pitch/type
//	4: octave
//	5: duration
function get_base_note(ia, ulen) {
    var	nd, c,
	note = get_acc_pit(ia)

//	if (!note)
//		return //null

	// duration
	c = ia[5]
	if (c[0] == '0') {		// compatibility
		parse.stemless = 1
		c = c.slice(1)
	}
	nd = parse_dur(c)
	note.dur = ulen * nd[0] / nd[1]
	return note
} // get_base_note()

function parse_vpos() {
    var	ty = 0,
	c = parse.mb[parse.li]

	if (a_dcn.length && a_dcn[a_dcn.length - 1] == "dot") {
		ty = C.SL_DOTTED
		a_dcn.pop()
	}
	switch (c) {
	case "'":
		parse.li++
		return ty + C.SL_ABOVE
	case ",":
		parse.li++
		return ty + C.SL_BELOW
	case '?':				// slur between staves (like ~)
		parse.li++
		return ty + C.SL_CENTER
	}
	return ty + C.SL_AUTO
} // parse_vpos()

// on end of slur, create the slur
function slur_add(s, nt) {		// nt = note if slur ending on note
    var	i, s2, sl

	// go back and find the last start of slur
	for (i = curvoice.sls.length; --i >= 0; ) {
		sl = curvoice.sls[i]

		// the slur must not start and stop on a same symbol
		if (sl.ss == s)
			continue
		curvoice.sls.splice(i, 1)
		sl.se = s			// ending symbol
		if (nt)
			sl.nte = nt
		s2 = sl.ss			// start of slur
		if (!s2.sls)
			s2.sls = []
		s2.sls.push(sl)

		// set a flag if the slur starts on a grace note
		if (sl.grace)
			sl.grace.sl1 = 1 //true
		return
	}

	// the lack of a starting slur may be due to a repeat
	for (s2 = s.prev; s2; s2 = s2.prev) {
		if (s2.bar_type
		 && s2.bar_type[0] == ':'
		 && s2.text) {
			if (!s2.sls)
				s2.sls = []
			s2.sls.push({
//fixme: should go back to the bar "|1" and find the slur type...
				ty: C.SL_AUTO,
				ss: s2,
				se: s
			})
			if (nt)
				s2.sls[s2.sls.length - 1].nte = nt
			return
		}
	}
//	syntax(1, "End of slur without start")
	if (!s.sls)
		s.sls = [];
	s.sls.push({
		ty: C.SL_AUTO,
		se: s,
		loc: 'i'			// no slur start
	})
	if (nt)
		s.sls[s.sls.length - 1].nte = nt
} // slur_add()

// convert a diatonic pitch and accidental to a MIDI pitch with cents
// @pit is the diatonic pitch from C-1
function pit2mid(pit, acc) {
    var	p = [0, 2, 4, 5, 7, 9, 11][pit % 7],	// chromatic pitch
	o = ((pit / 7) | 0) * 12,		// octave
	p0, p1, s, b40

	if (curvoice.snd_oct)
		o += curvoice.snd_oct
	if (acc == 3)				// if natural accidental
		acc = 0
	if (acc) {
		if (typeof acc == "object") {
			s = acc[0] / acc[1]	// microtonal accidental
			if (acc[1] == 100)	// in cents
				return p + o + s
		} else {
			s = acc			// simple accidental
		}
	} else {
		if (cfmt.temper)
			return cfmt.temper[tm.p_b40[pit % 7]] + o
		return p + o
	}
	if (!cfmt.nedo) {			// non equal temperament
		if (!cfmt.temper) {
			p += o + s		// standard temperament
			return p
		}
	} else {				// equal temperament
		if (typeof acc != "object") {	// if not a fraction
			b40 = tm.p_b40[pit % 7] + acc
			return cfmt.temper[b40] + o
		}

		if (acc[1] == cfmt.nedo) { // fraction of the edo divider
			b40 = tm.p_b40[pit % 7]
			return cfmt.temper[b40] + o + s
		}
	}

	p0 = cfmt.temper[tm.p_b40[pit % 7]]	// main note
	if (s > 0) {					// sharp
		p1 = cfmt.temper[(tm.p_b40[pit % 7] + 1) % 40]
		if (p1 < p0)
			p1 += 12
	} else {					// flat
		p1 = cfmt.temper[(tm.p_b40[pit % 7] + 39) % 40]
		if (p1 > p0)
			p1 -= 12
		s = -s
	}
	return p0 + o + (p1 - p0) * s
} // pit2mid()

// handle the ties
function do_ties(s,			// tie ending symbol
		 tie_s) {		// tie starting symbol
    var	i, m, not1, not2, mid, g,
	nt = 0,
	se = (tie_s.time + tie_s.dur) == curvoice.time	// 'start-end' flag

	for (m = 0; m <= s.nhd; m++) {
		not2 = s.notes[m]
		mid = not2.midi
		if (tie_s.typ != C.grace) {
			for (i = 0; i <= tie_s.nhd; i++) {
				not1 = tie_s.notes[i]
				if (!not1.tie_ty)
					continue
				if (not1.midi == mid		// same sounding pitch
				 && (!se
				  || !not1.tie_e)) {		// (if unison)
					not2.tie_s = not1
					not2.s = s
					if (se) {
						not1.tie_e = not2
						not1.s = tie_s
					}
					nt++
					break
				}
			}
		} else {
			for (g = tie_s.grasym; g; g = g.next) {
				not1 = g.notes[0]	// (fixme: only one note)
				if (!not1.tie_ty)
					continue
				if (not1.midi == mid) {
					g.ti1 = 1 //true
					not2.tie_s = not1
					not2.s = s
					not1.tie_e = not2
					not1.s = g
					nt++
					break
				}
			}
		}
	}

	if (!nt)
		error(1, tie_s, errs.bad_tie)
	else
		s.ti2 = 1 //true
} // do_ties()

//fixme: test
function get_note() {
    var	c,
	i = parse.li,
	p = parse.mb,
	ia = ["", "", "", "", "", "", ""]

	// get the accidentals
	c = p[i]
	if ("_=^".indexOf(c) >= 0) {
		while ("_=^".indexOf(c) >= 0) {
			ia[1] += c
			c = p[++i]
		}
		while ("1234567890/".indexOf(c) >= 0) {
			ia[2] += c
			c = p[++i]
		}
	}
	if ((c >= 'A' && c <= 'G')
	 || (c >= 'a' && c <= 'g')
	 || c == 'x' || c == 'y' || c == 'z'
	 || c == 'X' || c == 'Z') {
		ia[3] = c
		c = p[++i]
	} else {
		parse.li = i
// not a note
		return
	}
	while (",'".indexOf(c) >= 0) {
		ia[4] += c
		c = p[++i]
	}
	while ((c >= '0' && c <= '9') || c == '/') {
		ia[5] += c
		c = p[++i]
	}

	parse.li = i
	return ia
} // get_note()

// (possible hook)
Mus.prototype.new_note = function(grace, sls) {
    var	note, in_chord, dcn, type, tie_s, acc_tie, ia, c,
	i, n, s2, nd, res, num, dur, apit, ap, div, ty, dot,
	sl1 = [],
	a_dcn_sav = a_dcn,		// save parsed decoration names
	c = parse.mb[--parse.li],	// start of note or chord (c == '[')
	s = new_sym({
		typ: C.note,
//		stem: 0,
	})

	a_dcn = []
	parse.stemless = 0 //false

	if (curvoice.color)
		s.color = curvoice.color

	if (grace) {
		s.grace = 1 //true
	} else {
		if (curvoice.tie_s) {	// if tie from previous note / grace note
			tie_s = curvoice.tie_s
			curvoice.tie_s = null
		}
		if (a_gch)
			csan_add(s)
		if (parse.repeat_n) {
			s.repeat_n = parse.repeat_n
			s.repeat_k = parse.repeat_k
			parse.repeat_n = 0
		}
	}

	if (c == '[') {			// chord
		in_chord = 1 //true
		parse.li++
		ia = [,,c,,]
	} else {
//fixme: test
if(0){
		tm.ren.lastIndex = parse.li
		ia = tm.ren.exec(parse.mb)
		if (!ia) {
			syntax(1, "'$1' is not a note",
					parse.mb[parse.li])
			return
		}
		parse.li = tm.ren.lastIndex
} else {
		ia = get_note()
		if (!ia) {
			syntax(1, "'$1' is not a note",
					parse.mb[parse.li])
			return
		}
}
	}

	// ia = array [0..5]
	// 1: accidental
	// 2: microtone fraction
	// 3: note/rest pitch/type
	// 4: octave
	// 5: duration
	switch (ia[3]) {
	case 'X':
		s.invis = 1 //true
		// fall thru
	case 'Z':
		s.typ = C.mrest
		s.nmes = ia[5] || 1
		if (curvoice.wmeasure == 1) {
			error(1, s, "multi-measure rest, but no measure!")
			return
		}
		s.dur = curvoice.wmeasure * s.nmes

		// convert 'Z'/'Z1' to a whole measure rest
		if (s.nmes == 1) {
			s.typ = C.rest
			s.dur_orig = s.dur
			s.fmr = 1 //true	// full measure rest
			s.notes = [{
				pit: 0,
				dur: s.dur
			}]
		} else {
			glovar.mrest_p = 1 //true
			if (par_sy.voices.length == 1) {
				s.tacet = curvoice.tacet
				delete s.invis	// show the 'H' when 'Xn'
			}
		}
		break
	case 'y':
		s.typ = C.space
		s.invis = 1 //true
		s.dur = 0
		s.width = ia[5] || 10
		if (tie_s) {
			curvoice.tie_s = tie_s
			tie_s = null
		}
		break
	case 'x':
		s.invis = 1 //true
		// fall thru
	case 'z':
		s.typ = C.rest
		nd = parse_dur(ia[5])
		s.dur_orig = ((curvoice.ulen < 0) ?
					C.BLEN :
					curvoice.ulen) * nd[0] / nd[1]
		s.dur = s.dur_orig * curvoice.dur_fact
		if (s.dur == curvoice.wmeasure)
			s.fmr = 1 //true	// full measure rest
		s.notes = [{
			pit: 0,
			dur: s.dur_orig
		}]
		break
	default:			// accidental, chord, note
		if (curvoice.acc_tie) {
			acc_tie = curvoice.acc_tie
			curvoice.acc_tie = null
		}
		s.notes = []

		// loop on the chord
		while (1) {
			if (in_chord) {	// in chord, get the slurs, deco's and accidental
				while (1) {
					c = parse.mb[parse.li]
					switch (c) {
//					case undefined:
//					case '\n':
//					case '%':
//						syntax(1, "Lack of ']'")
//						return //null
					case '(':
						parse.li++
						sl1.push(parse_vpos())
						continue
					default:
						type = char_tb[c.charCodeAt(0)]
						if (type.length > 1)
							c = type	// shortcut
						if (c[0] != '!')
							break
						parse.li++
						get_deco(c)
						continue
					}
					break
				}
				
				tm.ren.lastIndex = parse.li
				ia = tm.ren.exec(parse.mb)
				if (!ia) {
					syntax(1, "'$1' is not a note",
							parse.mb[parse.li])
					return //null
				}
				parse.li = tm.ren.lastIndex
			}
			note = get_base_note(ia, s.grace ? C.BLEN / 4
							: curvoice.ulen < 0
							? C.BLEN : curvoice.ulen)

			// transpose and set the midi pitch
			set_midi(note, acc_tie)

			if (sl1.length) {
				while (1) {
					i = sl1.shift()
					if (!i)
						break
					curvoice.sls.push({
						ty: i,
						ss: s,
						nts: note	// starting note
					})
				}
			}
			if (a_dcn.length) {
				s.time = curvoice.time	// (needed for !tie)!
				dh_cnv(s, note)
			}
			s.notes.push(note)
			if (!in_chord)
				break

			// in chord: get the ending slurs and the ties
			c = parse.mb[parse.li]
			while (1) {
				switch (c) {
				case ')':
					slur_add(s, note)
					c = parse.mb[++parse.li]
					continue
				case '.':
					c = parse.mb[++parse.li]
					if (c != '-') {
//						syntax(1, "Misplaced dot")
						break
					}
					dot = 1
					parse.li++
					// fall thru
				case '-':
					parse.li++
					note.tie_ty = parse_vpos()
					if (dot) {
						dot = 0
						note.tie_ty |= C.SL_DOTTED
					}
					note.s = s
					curvoice.tie_s = s
					s.ti1 = 1 //true
					apit = note.pit + 37	// pitch from C-1
					if (curvoice.tr_dp)	// (if transposition)
						apit -= curvoice.tr_dp
					if (curvoice.acc[apit]
					 || (acc_tie && acc_tie[apit])) {
						if (!curvoice.acc_tie)
							curvoice.acc_tie = []
						n = curvoice.acc[apit]
						if (acc_tie && acc_tie[apit])
							n = acc_tie[apit]
						curvoice.acc_tie[apit] = n
					}
					c = parse.mb[parse.li]
					continue
				}
				break
			}
			if (c == ']') {			// end of chord
				s.nhd = s.notes.length - 1

				c = parse.mb[++parse.li]
				if ("123456789/".indexOf(c) < 0)
					break
				tm.rel.lastIndex = parse.li
				c = tm.rel.exec(parse.mb)
				parse.li = tm.rel.lastIndex

				// adjust the chord duration
				nd = parse_dur(c[0])
				nd = nd[0] / nd[1]
				if (nd != 1) {
					for (i = 0; i <= s.nhd ; i++) {
						note = s.notes[i]
						note.dur *= nd
					}
				}
				break
			}
		}

		// handle the starting slurs
		if (sls.length) {
			while (1) {
				i = sls.shift()
				if (!i)
					break
				curvoice.sls.push({
					ty: i,
					ss: s
					// no starting note
				})
				if (grace)
					curvoice.sls[curvoice.sls.length - 1].grace =
										grace
			}
		}

		// the duration of the chord is the duration of the 1st note
		s.dur_orig = s.notes[0].dur
		s.dur = s.notes[0].dur * curvoice.dur_fact
		break
	}
	if (s.grace && s.typ != C.note) {
		syntax(1, errs.bad_grace)
		return //null
	}

	if (s.notes) {				// if note or rest
		if (!grace) {
			switch (curvoice.pos.stm & 0x07) {
			case C.SL_ABOVE: s.stem = 1; break
			case C.SL_BELOW: s.stem = -1; break
			case C.SL_HIDDEN: s.stemless = 1; break
			}

			// adjust the symbol duration
			num = curvoice.brk_rhythm
			if (num) {
				curvoice.brk_rhythm = 0
				s2 = curvoice.last_note
				if (num > 0) {
					n = num * 2 - 1
					s.dur = s.dur * n / num
					s.dur_orig = s.dur_orig * n / num
					for (i = 0; i <= s.nhd; i++)
						s.notes[i].dur =
							s.notes[i].dur * n / num
					s2.dur /= num
					s2.dur_orig /= num
					for (i = 0; i <= s2.nhd; i++)
						s2.notes[i].dur /= num
				} else {
					num = -num
					n = num * 2 - 1
					s.dur /= num
					s.dur_orig /= num
					for (i = 0; i <= s.nhd; i++)
						s.notes[i].dur /= num
					s2.dur = s2.dur * n / num
					s2.dur_orig = s2.dur_orig * n / num
					for (i = 0; i <= s2.nhd; i++)
						s2.notes[i].dur =
							s2.notes[i].dur * n / num
				}
				curvoice.time = s2.time + s2.dur

				// adjust the time of the grace notes, bars...
				for (s2 = s2.next; s2; s2 = s2.next)
					s2.time = curvoice.time
			}
		} else {		// grace note - adjust its duration
			div = curvoice.ckey.k_bagpipe ? 8 : 4
			for (i = 0; i <= s.nhd; i++)
				s.notes[i].dur /= div
			s.dur /= div
			s.dur_orig /= div
			if (grace.stem)
				s.stem = grace.stem
		}

		curvoice.last_note = s

		// get the possible ties
		c = parse.mb[parse.li]
		while (1) {
			switch (c) {
			case '.':
				c = parse.mb[parse.li + 1]
				if (c != '-')
					break
				dot = 1
				parse.li++
				// fall thru
			case '-':
				parse.li++
				ty = parse_vpos()
				if (dot) {
					dot = 0
					ty |= C.SL_DOTTED
				}
				for (i = 0; i <= s.nhd; i++) {
					s.notes[i].tie_ty = ty
					s.notes[i].s = s
				}
				curvoice.tie_s = grace || s
				curvoice.tie_s.ti1 = 1 //true
				for (i = 0; i <= s.nhd; i++) {
					note = s.notes[i]
					apit = note.pit + 37	// pitch from C-1
					if (curvoice.tr_dp)	// (if transposition)
						apit -= curvoice.tr_dp
					if (curvoice.acc[apit]
					 || (acc_tie && acc_tie[apit])) {
						if (!curvoice.acc_tie)
							curvoice.acc_tie = []
						n = curvoice.acc[apit]
						if (acc_tie && acc_tie[apit])
							n = acc_tie[apit]
						curvoice.acc_tie[apit] = n
					}
				}
				c = parse.mb[parse.li]
				continue
			}
			break
		}

		// handle the ties ending on this chord/note
		if (tie_s		// if tie from previous note / grace note
		 && s.typ != C.rest)
			do_ties(s, tie_s)
	}

	parse.iend = parse.li
	sym_link(s)

	if (!grace) {
		if (!curvoice.lyric_restart)
			curvoice.lyric_restart = s
		if (!curvoice.sym_restart)
			curvoice.sym_restart = s
	}

	if (a_dcn_sav.length) {
		a_dcn = a_dcn_sav
		deco_cnv(s, s.prev)
	}
	if (grace && s.ottava)
		grace.ottava = s.ottava
	if (parse.stemless)
		s.stemless = 1 //true
	return s
} // new_note()

// adjust the duration of the elements in a tuplet
function tp_adj(s, fact) {
    var	d,
	tim = s.time,
	to = curvoice.time - tim,	// previous delta time
	tt = to * fact			// new delta time

	while (1) {
//fixme: tuplets in grace notes?
		s.in_tuplet = 1 //true
		if (!s.grace) {
			s.time = tim
			if (s.dur) {
				d = Math.round(s.dur * tt / to)	// new duration
				to -= s.dur		// old remaining time
				s.dur = d
				tt -= s.dur		// new remaining time
				tim += s.dur
			}
		}
		if (!s.next)
			break
		s = s.next
	}
	if (s.tpe)
		s.tpe++
	else
		s.tpe = 1
	curvoice.time = tim
} // tp_adj()

// get a decoration
function get_deco(c) {
	if (c == '!') {				// start of decoration
		tm.red.lastIndex = parse.li - 1
		c = tm.red.exec(parse.mb)
		if (!c) {
			syntax(1, "No end of decoration")
			return
		}
		c = c[0]
		parse.li = tm.red.lastIndex
	}
	c = c.slice(1, -1)			// remove the '!'s
	a_dcn.push(c)
} // get_deco()

// characters in the music line (ASCII only)
var nil = "0",
    char_tb = [
	nil, nil, nil, nil,		// 00 - ..
	nil, nil, nil, nil,
	nil, " ", "$", nil,		// . \t \n .
	nil, nil, nil, nil,
	nil, nil, nil, nil,
	nil, nil, nil, nil,
	nil, nil, nil, nil,
	nil, nil, nil, nil,		// .. - 1f
	" ", "!", '"', "#",		// (sp) ! " #
	"$", nil, "&", nil,		// $ % & '
	"(", ")", "#", nil,		// ( ) * +
	nil, "-", "!dot!", nil,		// , - . /
	nil, nil, nil, nil, 		// 0 1 2 3
	nil, nil, nil, nil, 		// 4 5 6 7
	nil, nil, "|", "#",		// 8 9 : ;
	"<", "A", "<", "#",		// < = > ?
	"#", "A", "A", "A",		// @ A B C
	"A", "A", "A", "A", 		// D E F G
	"!fermata!", "d", "d", "d",	// H I J K
	"!accent!", "!lowermordent!",
		"d", "!coda!",		// L M N O
	"!uppermordent!", "d",
		"d", "!segno!",		// P Q R S
	"!trill!", "d", "d", "d",	// T U V W
	"A", "d", "A", "[",		// X Y Z [
	"\\","|", "A", "A",		// \ ] ^ _
	"#", "A", "A", "A",	 	// ` a b c
	"A", "A", "A", "A",	 	// d e f g
	"d", "d", "d", "d",		// h i j k
	"d", "d", "d", "d",		// l m n o
	"d", "d", "d", "d",		// p q r s
	"d", "!upbow!",
		"!downbow!", "d",	// t u v w
	"A", "A", "A", "{",		// x y z {
	"|", "}", "!gmark!", nil,	// | } ~ (del)
    ] // char_tb[]

// parse a music line
// parse.mb = music line
function parse_music_line() {
    var	grace, last_note_sav, a_dcn_sav, no_eol, tps,
	tp = [],
	tpn = -1,
	sls = []

	// check if a dynamic macro matches a source sequence
	// if yes return the base note
	function check_mac(m) {
	    var	i, j, b

		for (i = 1, j = parse.li + 1; i < m.length; i++, j++) {
			if (m[i] == parse.mb[j])
				continue
			if (m[i] != 'n')		// search the base note
				return //undefined
			b = C.ntb.indexOf(parse.mb[j])
			if (b < 0)
				return //undefined
			while (parse.mb[j + 1] == "'") {
				b += 7
				j++
			}
			while (parse.mb[j + 1] == ',') {
				b -= 7
				j++
			}
		}
		parse.li = j
		return b
	} // check_mac()

	// convert a note as a number into a note as a ABC string
	function n2n(n) {
	    var	c = ''

		while (n < 0) {
			n += 7
			c += ','
		}
		while (n >= 14) {
			n -= 7
			c += "'"
		}
		return C.ntb[n] + c
	} // n2n()

	// expand a dynamic macro
	function expand(m, b) {
		if (b == undefined)		// if static macro
			return m
	    var	c, i,
		r = "",				// result
		n = m.length

		for (i = 0; i < n; i++) {
			c = m[i]
			if (c >= 'h' && c <= 'z') {
				r += n2n(b + c.charCodeAt(0) - 'n'.charCodeAt(0))
			} else {
				r += c
			}
		}
		return r
	} // expand()

	// parse a macro
	function parse_mac(k, b) {
	    var	te, ti, curv, s,
		m = mac[k],
		parse_sav = parse

		parse = Object.create(parse)
		parse.li = 0

		// if the macro is not to be displayed
		if (cfmt.writefields.indexOf('m') < 0) {

			// build the display sequence from the original sequence
			parse.mb = k.replace('n', n2n(b))
			s = curvoice.last_sym
			ti = curvoice.time		// start time
			parse_seq(1)
			if (!s)
				s = curvoice.sym
			for (s = s.next; s; s = s.next)
				s.noplay = 1 //true
			te = curvoice.time		// end time
			curv = curvoice

			// and put the macro sequence in a play specific voice
			curvoice = clone_voice(curv.id + '-p')
			if (!par_sy.voices[curvoice.v]) {
				curvoice.second = 1
				par_sy.voices[curvoice.v] = {
					st: curv.st,
					second: 1,
					range: curvoice.v
				}
			}
			curvoice.time = ti
			s = curvoice.last_sym
			parse.mb = expand(m, b)
			parse.li = 0
			parse.istart = parse_sav.istart
			parse_seq(1)
			if (curvoice.time != te)
				syntax(1, "Bad length of the macro sequence")
			if (!s)
				s = curvoice.sym
			for ( ; s; s = s.next)
				s.invis = s.play = 1 //true
			curvoice = curv
		} else {
			parse.mb = expand(m, b)
			parse_seq(1)			// same display and play
		}

		// restore the parsing variables
		parse = parse_sav
	} // parse_mac()

	// parse a music sequence
	function parse_seq(in_mac) {
	    var	c, idx, type, k, s, dcn, i, n, text, note, sav

		// scan the line item after item
		while (1) {
			parse.istart = parse.li
			c = parse.mb[parse.li]
			if (!c || c == '\n' || c == '%')
				break		// end of line

			// check if start of a macro
			if (!in_mac && maci[c[0]]) {
				n = undefined
				for (k in mac) {
					if (!mac.hasOwnProperty(k)
					 || k[0] != c[0])
						continue
					if (k.indexOf('n') < 0) {	// static
						if (parse.mb.indexOf(k, parse.li)
								!= parse.li)
							continue
						parse.li += k.length
					} else {			// dynamic
						n = check_mac(k)
						if (n == undefined)
							continue
					}
					parse_mac(k, n)
					n = 1
					break
				}
				if (n)
					continue
			}
			parse.li++

			i = c.charCodeAt(0)
			type = char_tb[i]
			if (!type) {
				syntax(1, errs.not_ascii)
				continue
			}
			if (type.length > 1)
				c = type		// shortcut
			switch (type[0]) {
			case ' ':			// beam break
				s = curvoice.last_note
				if (s) {
					s.beam_end = 1 //true
					if (grace)
						grace.gr_shift = 1 //true
				}
				break
			case '$':			// line break
				if (!cfmt.barsperstaff)
					curvoice.eoln = 1 //true
				break
			case '&':			// voice overlay
				if (grace) {
					syntax(1, errs.bad_grace)
					break
				}
				c = parse.mb[parse.li]
				if (c == ')') {
					parse.li++
					get_vover(c)	// full overlay stop
					break
				}
				get_vover('&')
				break
			case '(':			// tuplet - slur start - vover
				c = parse.mb[parse.li]
				if (c >= '1' && c <= '9') {
					tm.ret.lastIndex = parse.li
					c = tm.ret.exec(parse.mb)
					parse.li = tm.ret.lastIndex
				    var	ppl = +c[1],
					qpl = +c[2],
					rpl = +c[3]

					if (!qpl)
						qpl = C.qplet_tb[ppl]
					if (!rpl)
						rpl = ppl
					if (!qpl)
						qpl = (curvoice.wmeasure % 9) ? 2 : 3
					if (tpn < 0)
						tpn = tp.length	// new tuplet
					tp.push({
						p: ppl,
						q: qpl,
						r: rpl,
						ro: rpl,
						f: curvoice.tup || cfmt.tuplets
					})
					break
				}
				c = parse.mb[parse.li]
				if (c == '&') {		// voice overlay start
					parse.li++
					if (grace) {
						syntax(1, errs.bad_grace)
						break
					}
					get_vover('(')
					break
				}
				sls.push(parse_vpos())
				continue
			case ')':			// slur end
				s = curvoice.last_sym
				if (s) {
					switch (s.typ) {
					case C.space:
						if (!s.notes) {
							s.notes = []
							s.notes[0] = {}
						}
					case C.note:
					case C.rest:
						break
					case C.grace:

						// stop the slur on the last grace note
						for (s = s.grasym; s.next; s = s.next)
							;
						break
					default:
						s = null
						break
					}
				}
				if (!s) {
					syntax(1, errs.bad_char, c)
					break
				}
				slur_add(s)
				break
			case '!':
				get_deco(c)
				break
			case '"':
				if (grace) {
					syntax(1, errs.bad_grace)
					break
				}
				if (c.length == 1) {	// start of chord symbol
					tm.res.lastIndex = parse.li - 1
					c = tm.res.exec(parse.mb)
					if (!c) {
						syntax(1, "No end of annotation")
						break
					}
					c = c[0]
					parse.li = tm.res.lastIndex
				}
				parse_gchord(c)
				break
			case '[':
				if (c.length == 1) {
					c = parse.mb[parse.li]		// bar?
					if ('[]|:"0123456789'.indexOf(c) >= 0) {
						new_bar(grace)		// yes
						break
					}
					c = parse.mb[parse.li + 1]
					if (c == ':') {
						i = parse.mb.indexOf(']', parse.li)
						if (i < 0 || i > parse.eol) {
							syntax(1, "Lack of ']'")
							break
						}
						c = parse.mb.slice(parse.li - 1, ++i)
						parse.li = i
					}
				}
				if (c[2] == ':') {	// if [X:xxx]
					if(c[1] == 'I') {
						self.do_pscom(c.slice(3, -1))
					} else {

						// change the start/stop indexes
						//	for V: multivoice
						sav = {
							is: parse.istart,
							ie: parse.eol
						}
						parse.istart = parse.li - c.length
						parse.eol = parse.li
						do_info(c[1], c.slice(3, -1))
						parse.istart = sav.is
						parse.eol = sav.ie
					}
					break
				}
				// must be the start of a chord
				// fall thru
			case 'A':				// note/rest
				s = self.new_note(grace, sls)
				if (!s)
					continue

				// handle the tuplets
				if (grace || !s.notes)
					continue

				if (tpn >= 0) {		// new tuplet
					s.tp = tp.slice(tpn)
					tpn = -1
					if (tps)
						s.tp[0].s = tps	// if nested
					tps = s
				} else if (!tps) {
					continue	// no tuplet active
				}

				k = tp[tp.length - 1]
				if (--k.r > 0)
					continue	// not end of tuplet yet

				while (1) {
					tp_adj(tps, k.q / k.p)
					i = k.ro	// number of notes of this tuplet
					if (k.s)
						tps = k.s  // start of upper tuplet

					tp.pop()		// previous level
					if (!tp.length) {
						tps = null	// done
						break
					}
					k = tp[tp.length - 1]
					k.r -= i
					if (k.r > 0)
						break
				}
				break
			case '<':				// '<' and '>'
				if (!curvoice.last_note) {
					syntax(1, "No note before '<'")
					break
				}
				if (grace) {
					syntax(1, "Cannot have a broken rhythm in grace notes")
					break
				}
				n = c == '<' ? 1 : -1
				while (1) {
					n *= 2
					c = parse.mb[parse.li + 1]
					if (c != '<' && c != '>')
						break
					parse.li++
				}
				curvoice.brk_rhythm = n
				break
			case '#':				// ignore
				break
			case '{':
				if (grace) {
					syntax(1, "'{' in grace note")
					break
				}
				last_note_sav = curvoice.last_note
				curvoice.last_note = null
				a_dcn_sav = a_dcn
				a_dcn = []
				parse.iend = parse.li
				grace = new_sym({
					typ: C.grace
//					dur: 0
				})
				if (curvoice.color)
					grace.color = curvoice.color
				switch (curvoice.pos.gst & 0x07) {
				case C.SL_ABOVE: grace.stem = 1; break
				case C.SL_BELOW: grace.stem = -1; break
				case C.SL_OPPOS: grace.stem = 2; break	// opposite
				}
				sym_link(grace)
				c = parse.mb[parse.li]
				if (c == '/') {
					parse.li++
					grace.sappo = 1	// acciaccatura
					break
				}
				break
			case '|':
				new_bar(grace)
				break
			case '}':
				if (curvoice.ignore) {
					grace = null
					break
				}
				s = curvoice.last_note
				if (!grace || !s) {
					syntax(1, errs.bad_char, c)
					break
				}
				if (a_dcn.length)
					syntax(1, "Decoration ignored")
				grace.grasym = grace.next
				grace.grasym.prev = null
				grace.next = null
				curvoice.last_sym = grace
				grace = null
				if (!s.prev			// if one grace note
				 && !curvoice.ckey.k_bagpipe) {
					for (i = 0; i <= s.nhd; i++)
						s.notes[i].dur *= 2
					s.dur *= 2
					s.dur_orig *= 2
				}
				curvoice.last_note = last_note_sav
				a_dcn = a_dcn_sav
				break
			case '\\':
				c = parse.mb[parse.li]
				if (c) {
					tm.ree.lastIndex = parse.li - 1
					c = tm.ree.exec(parse.mb)	// end of line?
					if (!c) {
						syntax(1, errs.bad_char,
							parse.mb[parse.li])
						break
					}
					parse.li = tm.ree.lastIndex - 1
				}
				no_eol = 1 //true
				break
			default:
				syntax(1, errs.bad_char, c)
				break
			}
		}
	} // parse_seq()

	// ---- parse_music_line ----
	if (parse.state != 3)		// if not in tune body
		return

	if (parse.tp) {
		tp = parse.tp
		tpn = parse.tpn
		tps = parse.tps
		parse.tp = null
	}

	parse_seq()

	if (tp.length) {
		parse.tp = tp
		parse.tps = tps
		parse.tpn = tpn
	}
	if (sls.length)
		syntax(1, "Start of slur without note")
	if (grace) {
		syntax(1, "No end of grace note sequence")
		curvoice.last_sym = grace.prev
		curvoice.last_note = last_note_sav
		if (grace.prev)
			grace.prev.next = null
	}
	if (!no_eol && !cfmt.barsperstaff && !vover
	 && char_tb['\n'.charCodeAt(0)] == '$')
		curvoice.eoln = 1 //true
//fixme
//	if (curvoice.in_bm && 
	if (curvoice.last_note
	 && curvoice.eoln && cfmt.breakoneoln)
		curvoice.last_note.beam_end = 1 //true
} // parse_music_line()
