// txtmus - tm.js
//
// Copyright (C) 2014-2024 Jean-Francois Moine
//
// This file is part of txtmus-core.
//
// txtmus-core is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// txtmus-core is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with txtmus-core.  If not, see <http://www.gnu.org/licenses/>.

// define the txtmus object if not done yet
if (typeof tm == "undefined")
    var	tm = {}

// constants
tm.C = {
	BLEN: 1536,

	// note heads
	FULL: 0,
	EMPTY: 1,
	OVAL: 2,
	OVALBARS: 3,
	SQUARE: 4,

	// position types
	SL_ABOVE: 0x01,		// position (3 bits)
	SL_BELOW: 0x02,
	SL_AUTO: 0x03,
	SL_HIDDEN: 0x04,
	SL_OPPOS: 0x05,
	SL_DOTTED: 0x08,	// modifiers
	SL_ALI_MSK: 0x70,	// align
		SL_ALIGN: 0x10,
		SL_CENTER: 0x20,
		SL_CLOSE: 0x40,

	// staff system flags
	O_BRACE: 0x01,
	C_BRACE: 0x02,
	O_BRACKET: 0x04,
	C_BRACKET: 0x08,
	O_PARENTH: 0x10,
	C_PARENTH: 0x20,
	STOP_BAR: 0x40,
	FL_VOICE: 0x80,
	O2_BRACE: 0x0100,
	C2_BRACE: 0x0200,
	O2_BRACKET: 0x0400,
	C2_BRACKET: 0x0800,
	MASTER_VOICE: 0x1000,

	// music symbols
	bar:	{nm: 'bar',	w: 6},
	block:	{nm: 'block',	w: 0},
	clef:	{nm: 'clef',	w: 2},
	custos: {nm: 'custos',	w: 8},
	grace:	{nm: 'grace',	w: 0},	// must be null
	key:	{nm: 'key',	w: 3},
	meter:	{nm: 'meter',	w: 4},
	mrest:	{nm: 'mrest',	w: 9},
	note:	{nm: 'note',	w: 9},
	part:	{nm: 'part',	w: 0},
	rem:	{nm: 'rem',	w: 0},
	rest:	{nm: 'rest',	w: 9},
	space:	{nm: 'space',	w: 5},	// before bar
	staves:	{nm: 'staves',	w: 0},
	stbrk:	{nm: 'stbrk',	w: 5},	// same as space
	tempo:	{nm: 'tempo',	w: 0},

	// parse constants
	qplet_tb: new Int8Array([ 0, 1, 3, 2, 3, 0, 2, 0, 3, 0 ]),
	ntb: "CDEFGABcdefgab"
    } // C

	// RE to extract accidental and pitch
tm.reap = /([_=^]*)(x?)([A-Ga-g])([,']*)/g	//'

	// RE to get a measure bar + variant
tm.reb = /([\][|:]+)("[^"]*"|[\d.,-]*)/g	//"

	// RE to get a command and its parameters
tm.recp = /([\w-]+)\s*([^]*)/;

	// RE to get a decoration
tm.red = /![^!\s]+!/g

	// RE to search if end of music line
tm.ree = /\\\s*[\n%]/g

	// RE to get an embedded information field
tm.rei = /\[[A-Za-z]:.+?\]/g

	// RE to get an integer
tm.rel = /[\d/]+/g

	// RE to extract the note/rest elements
	// 1: accidental
	// 2: microtone fraction
	// 3: note/rest pitch/type
	// 4: octave
	// 5: duration
tm.ren = /([_=^]*)(\d*\/?\d*)([A-Ga-gxyzXZ])([,']*)([\d/]*)/g	// '

	// RE to get a double quoted string
tm.res = /".+?"/g

	// RE to get a tuplet
tm.ret = /(\d+):?(\d*):?(\d*)/g

	// key table - index = number of accidentals + 7
tm.keys = [
	new Int8Array([-1,-1,-1,-1,-1,-1,-1]),	// 7 flat signs
	new Int8Array([-1,-1,-1, 0,-1,-1,-1]),	// 6
	new Int8Array([ 0,-1,-1, 0,-1,-1,-1]),	// 5
	new Int8Array([ 0,-1,-1, 0, 0,-1,-1]),	// 4
	new Int8Array([ 0, 0,-1, 0, 0,-1,-1]),	// 3
	new Int8Array([ 0, 0,-1, 0, 0, 0,-1]),	// 2
	new Int8Array([ 0, 0, 0, 0, 0, 0,-1]),	// 1
	new Int8Array(7),			// no accidental
	new Int8Array([ 0, 0, 0, 1, 0, 0, 0]),	// 1 sharp signs
	new Int8Array([ 1, 0, 0, 1, 0, 0, 0]),	// 2
	new Int8Array([ 1, 0, 0, 1, 1, 0, 0]),	// 3
	new Int8Array([ 1, 1, 0, 1, 1, 0, 0]),	// 4
	new Int8Array([ 1, 1, 0, 1, 1, 1, 0]),	// 5
	new Int8Array([ 1, 1, 1, 1, 1, 1, 0]),	// 6
	new Int8Array([ 1, 1, 1, 1, 1, 1, 1])	// 7
]

// base-40 representation of musical pitch
// (http://www.ccarh.org/publications/reprints/base40/)
tm.p_b40 = new Int8Array(			// staff pitch to base-40
//		 C  D   E   F   G   A   B
		[2, 8, 14, 19, 25, 31, 37])
tm.b40_p = new Int8Array(			// base-40 to staff pitch
//		       C		 D
		[0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1,
//	      E		     F		       G
	2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4,
//	      A			B
	5, 5, 5, 5, 5, 5, 6, 6, 6, 6, 6])
tm.b40_a = new Int8Array(			// base-40 to accidental
//		         C		      D
		[-2, -1, 0, 1, 2, -3, -2, -1, 0, 1, 2, -3,
//		E		 F		      G
	-2, -1, 0, 1, 2, -2, -1, 0, 1, 2, -3, -2, -1, 0, 1, 2, -3,
//		A		     B
	-2, -1, 0, 1, 2, -3, -2, -1, 0, 1, 2])
tm.b40_m = new Int8Array(			// base-40 to midi
//			 C		   D
		[-2, -1, 0, 1, 2, 0, 0, 1, 2, 3, 4, 0,
//	      E		     F		       G
	2, 3, 4, 5, 6, 3, 4, 5, 6, 7, 0, 5, 6, 7, 8, 9, 0,
//	      A			    B
	7, 8, 9, 10, 11, 0, 9, 10, 11, 12, 13])
tm.b40l5 = new Int8Array([			// base-40 to line of fifth
//		  C			  D
	-14, -7,  0,  7, 14,  0,-12, -5,  2,  9, 16,  0,
//		  E		      F			      G	
	-10, -3,  4, 11, 18,-15, -8, -1,  6, 13,  0,-13, -6,  1,  8, 15,  0,
//		  A			  B
	-11, -4,  3, 10, 17,  0, -9, -2,  5, 12, 19 ])

tm.ib40 = new Int8Array(		// interval with sharp to base-40 interval
	[0, 1, 6,7,12,17,18,23,24,29,30,35])

tm.pab40 = function(p, a) {
	p += 37				// staff pitch from C-1
   var	b40 = ((p / 7) | 0) * 40 + tm.p_b40[p % 7]
	if (a && a != 3)		// if some accidental, but not natural
		b40 += a
	return b40
} // pit2b40()
tm.b40p = function(b) {
	return ((b / 40) | 0) * 7 + tm.b40_p[b % 40] - 37
} // b40p()
tm.b40a = function(b) {
	return tm.b40_a[b % 40]
} // b40a()
tm.b40m = function(b) {
	return ((b / 40) | 0) * 12 + tm.b40_m[b % 40]
} // b40m()

// chord table
// This table is used in various modules
// to convert the types of chord symbols to a minimum set.
// More chord types may be added by the command %%chordalias.
tm.ch_alias = {
	"maj": "",
	"min": "m",
	"-": "m",
	"°": "dim",
	"+": "aug",
	"+5": "aug",
	"maj7": "M7",
	"Δ7": "M7",
	"Δ": "M7",
	"min7": "m7",
	"-7": "m7",
	"ø7": "m7b5",
	"°7": "dim7",
	"min+7": "m+7",
	"aug7": "+7",
	"7+5": "+7",
	"7#5": "+7",
	"sus": "sus4",
	"7sus": "7sus4"
} // ch_alias

// global fonts
tm.styfx = {}		// font style => font.fx
//tm.fxus = []		// font.fx => used
tm.frfx = 0		// free font.fx

// font weight
// reference:
//	https://developer.mozilla.org/en-US/docs/Web/CSS/font-weight
tm.ft_w = {
	thin: 100,
	extralight: 200,
	light: 300,
	regular: 400,
	medium:  500,
	semi: 600,
	demi: 600,
	semibold: 600,
	demibold: 600,
	bold: 700,
	extrabold: 800,
	ultrabold: 800,
	black: 900,
	heavy: 900
} // ft_w

tm.ft_re = new RegExp('\
-?Thin|-?Extra Light|-?Light|-?Regular|-?Medium|\
-?[DS]emi|-?[DS]emi[ -]?Bold|\
-?Bold|-?Extra[ -]?Bold|-?Ultra[ -]?Bold|-?Black|-?Heavy/',
	"i")

// position of a voice element
tm.posval = {
	above: tm.C.SL_ABOVE,
	auto: 0,		// !! not C.SL_AUTO !!
	below: tm.C.SL_BELOW,
	down: tm.C.SL_BELOW,
	hidden: tm.C.SL_HIDDEN,
	opposite: tm.C.SL_OPPOS,
	under: tm.C.SL_BELOW,
	up: tm.C.SL_ABOVE
} // posval

// simplify a rational number n/d
tm.rat = function(n, d) {
    var	a, t,
	n0 = 0,
	d1 = 0,
	n1 = 1,
	d0 = 1

	while (1) {
		if (d == 0)
			break
		t = d
		a = (n / d) | 0
		d = n % d
		n = t
		t = n0 + a * n1
		n0 = n1
		n1 = t
		t = d0 + a * d1
		d0 = d1
		d1 = t
	}
	return [n1, d1]
} // rat()

// text options
tm.textopt = {
	align: 'j',
	center: 'c',
	fill: 'f',
	justify: 'j',
	obeylines: 'l',
	ragged: 'f',
	right: 'r',
	skip: 's'
}
// compare pitches
// This function is used to sort the note pitches
tm.pitcmp = function(n1, n2) { return n1.pit - n2.pit }

// start of the Mus object
tm.Mus = function(user) {
	"use strict"

    // constants
    var	C = tm.C,
	Mus = tm.Mus,			// pointer to the function for .prototype

	// mask some unsafe functions
	require = empty_function,
	system = empty_function,
	write = empty_function,
	XMLHttpRequest = empty_function,
	std = null,
	os = null

// -- constants --

// staff system
    var	IN = 96,		// resolution 96 PPI
	CM = 37.8,		// 1 inch = 2.54 centimeter
	YSTEP			// number of steps for y offsets

// error texts
var errs = {
	bad_char: "Bad character '$1'",
	bad_grace: "Bad character in grace note sequence",
	bad_tie: "Bad tie",
	bad_transp: "Bad transpose value",
	bad_val: "Bad value in $1",
	bar_grace: "Cannot have a bar in grace notes",
	ignored: "$1: inside tune - ignored",
	misplaced: "Misplaced '$1' in %%score",
	must_note: "!$1! must be on a note",
	must_note_rest: "!$1! must be on a note or a rest",
	nonote_vo: "No note in voice overlay",
	not_ascii: "Not an ASCII character",
	not_enough_n: 'Not enough notes/rests for %%repeat',
	not_enough_m: 'Not enough measures for %%repeat',
	not_enough_p: "Not enough parameters in %%map",
	not_in_tune: "Cannot have '$1' inside a tune",
	notransp: "Cannot transpose with a temperament"
}

    var	self = this,				// needed for modules
	glovar = {
		meter: {
			typ: tm.C.meter,	// meter in tune header
			wmeasure: 1,		// no M:
			a_meter: []		// default: none
		},
	},
	info = {},			// information fields
	parse = {
		ctx: {},
		prefix: '%',
		state: 0,
		ottava: [],
	},
	fxus = [],		// font.fx => used
	tunes = [],		// first time symbol and voice array per tune for playing
	psvg			// PostScript

// utilities
function clone(obj, lvl) {
	if (!obj)
		return obj
    var	tmp = new obj.constructor
	for (var k in obj)
	    if (obj.hasOwnProperty(k)) {
		if (lvl && typeof obj[k] == "object")
			tmp[k] = clone(obj[k], lvl - 1)
		else
			tmp[k] = obj[k]
	    }
	return tmp
} // clone()

function errbld(sev, txt, fn, idx) {
    var	i, j, l, c, h

	if (user.errbld) {
		switch (sev) {
		case 0: sev = "warn"; break
		case 1: sev = "error"; break
		default: sev= "fatal"; break
		}
		user.errbld(sev, txt, fn, idx)
		return
	}

	// convert the buffer offset into line number, column number
	if (idx != undefined && idx >= 0) {
		i = l = 0
		while (1) {
			j = parse.mb.indexOf('\n', i)
			if (j < 0 || j > idx)
				break
			l++
			i = j + 1
		}
		c = idx - i
	}
	h = ""
	if (fn) {
		h = fn
		if (l)
			h += ":" + (l + 1) + ":" + (c + 1)
		h += " "
	}
	switch (sev) {
	case 0: h += "Warning: "; break
	case 1: h += "Error: "; break
	default: h += "Internal bug: "; break
	}
	user.errmsg(h + txt, l, c)
} // errbld()

function error(sev, s, msg, a1, a2, a3, a4) {
    var	i, j, regex, tmp

	if (!sev && cfmt.quiet)
		return
	if (s) {
		if (s.err)		// only one error message per symbol
			return
		s.err = true
	}
	if (user.textrans) {
		tmp = user.textrans[msg]
		if (tmp)
			msg = tmp
	}
	if (arguments.length > 3)
		msg = msg.replace(/\$./g, function(a) {
			switch (a) {
			case '$1': return a1
			case '$2': return a2
			case '$3': return a3
			default  : return a4
			}
		})
	if (s && s.fname)
		errbld(sev, msg, s.fname, s.istart)
	else
		errbld(sev, msg)
} // error()

function syntax(sev, msg, a1, a2, a3, a4) {
    var	s = {
		fname: parse.fname,
		istart: parse.istart
	}

	error(sev, s, msg, a1, a2, a3, a4)
} // syntax()

// inject javascript code
function js_inject(js) {
	eval('"use strict";\n' + js)
} // js_inject()
