// txtmus - tune.js - tune generation
//
// Copyright (C) 2014-2024 Jean-Francois Moine
//
// This file is part of txtmus-core.
//
// txtmus-core is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// txtmus-core is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with txtmus-core.  If not, see <http://www.gnu.org/licenses/>.

    var	par_sy,		// current staff system for parse
	cur_sy,		// current staff system for generation
	voice_tb,
	curvoice,
	staves_found,
	vover,		// voice overlay
	tsfirst,
	multicol,	// multi column object
	sav = {}	// save of global variables

// this function is called at program start and on end of tune
function init_tune() {
	nstaff = -1
	voice_tb = []
	par_sy =
		curvoice = null
	new_syst()
	staves_found = -1
	gene = {}
	a_de = []			// remove old decorations
	cross = {}			// new cross voice decorations
} // init_tune()

// save the global variables when entering in a tune
function start_tune() {
	sav.cfmt = clone(cfmt)
	sav.info = clone(info, 2)	// (level 2 for info.V[])
	sav.char_tb = clone(char_tb)
	sav.glovar = clone(glovar)
	sav.maps = clone(maps, 1)
	sav.mac = clone(mac)
	sav.maci = clone(maci)
	parse.state = 1			// in tune header
} // start_tune()

// generate the tune and restore the global variables
function end_tune() {
	generate()
	cfmt = sav.cfmt
	info = sav.info
	char_tb = sav.char_tb
	glovar = sav.glovar
	maps = sav.maps
	mac = sav.mac
	maci = sav.maci
	parse.tune_v_opts = null
	parse.scores = null
	parse.ufmt = 0 //false
	delete parse.pq
	init_tune()
	img.chg = 1 //true
	set_page()
	parse.state = 0			// out of tune
} // end_tune()

// apply the %%voice options of the current voice
function voice_filter() {
    var	opt

	function vfilt(opts, opt) {
	    var	i,
		sel = new RegExp(opt)

		if (sel.test(curvoice.id)
		 || sel.test(curvoice.nm)) {
			for (i = 0; i < opts.length; i++)
				self.do_pscom(opts[i])
		}
	} // vfilt()

	// global
	if (parse.voice_opts)
	    for (opt in parse.voice_opts) {
		if (parse.voice_opts.hasOwnProperty(opt))
			vfilt(parse.voice_opts[opt], opt)
	}

	// tune
	if (parse.tune_v_opts)
	    for (opt in parse.tune_v_opts) {
		if (parse.tune_v_opts.hasOwnProperty(opt))
			vfilt(parse.tune_v_opts[opt], opt)
	}
}

// create a symbol in the current voice
// @s is an object containing the specific attributes of the symbol
//	It is also the returned value.
function new_sym(s) {
//	s.v = curvoice.v
//	s.p_v = curvoice
//	s.time = curvoice.time
//	s.st = curvoice.cst
	s.dur = 0
	switch (s.typ) {
	case C.note:
	case C.rest:
	case C.mrest:
	case C.space:
		s.nhd = 0
//		s.notes = []
		s.xmx = 0
	}
	return s
} // new_sym()

// -- link a music symbol into the current voice --
// if a voice is ignored (not in %%staves) don't link the symbol
//	but update the time for P: and Q:
function sym_link(s) {
    var	tim = curvoice.time

	if (!s.fname)
		set_ref(s)
	if (!curvoice.ignore) {
		s.prev = curvoice.last_sym
		if (curvoice.last_sym)
			curvoice.last_sym.next = s
		else
			curvoice.sym = s
	} else if (s.bar_type) {
		curvoice.last_bar = s
	}
	curvoice.last_sym = s

	s.v = curvoice.v
	s.p_v = curvoice
	s.st = curvoice.cst
	s.time = tim
	if (s.dur && !s.grace)
		curvoice.time += s.dur
	parse.ufmt = 1
	s.fmt = cfmt				// global parameters
	s.pos = curvoice.pos
	if (curvoice.second)
		s.second = 1 //true
	if (curvoice.floating)
		s.floating = 1 //true
	if (curvoice.eoln) {
		s.soln = 1 //true
		curvoice.eoln = 0 //false
	}
} // sym_link()

// -- add a new symbol in a voice --
function sym_add(p_v, typ) {
    var	s2,
	s = {
		typ: typ,
		dur: 0
	},
	p_v2 = curvoice

	curvoice = p_v
	sym_link(s)
	curvoice = p_v2
	s2 = s.prev
	if (!s2)
		s2 = s.next
	if (s2) {
		s.fname = s2.fname
		s.istart = s2.istart
		s.iend = s2.iend
	}
	return s
} // sym_add()

// set the time linkage
function sort_all() {
    var	s, s2, time, w, wmin, ir, fmt,
	fl, new_sy,
	nv = voice_tb.length,
	vtb = [],
	vn = [],			// voice indexed by range
	sy = cur_sy,			// first staff system
	v = sy.top_voice,
	p_v = voice_tb[v],
	prev = new_sym({		// symbol defining the first staff system
		typ: C.staves,
		fname: parse.fname,
		v: v,
		p_v: p_v,
		time: 0,
		st: 0,
		sy: sy,
		next: p_v.sym,
		seqst: 1 //true
	})

	if (!p_v.sym)			// no symbol in the top voice!
		return

	// check if different bars at the same time
	function b_chk() {
	    var	bt, s, s2, v, t,
		ir = 0

		while (1) {
			v = vn[ir++]
			if (v == undefined)
				break
			s = vtb[v]
			if (!s || !s.bar_type || s.invis
			 || s.time != time)
				continue
			if (!bt) {
				bt = s.bar_type
				if (s.text && bt == '|')
					t = s.text
				continue
			}
			if (s.bar_type != bt)
				break
			if (s.text && !t && bt == '|') {
				t = s.text
				break
			}
		}

		// if the previous symbol is a grace note at the same offset as the bar
		// remove the grace notes from the previous time sequence
		if (!fl) {
			while (prev.typ == C.grace
			    && vtb[prev.v] && !vtb[prev.v].bar_type) {
				vtb[prev.v] = prev
				prev = prev.ts_prev
				fl = 1 //true
			}
		}

		if (v == undefined)
			return			// no bar problem

		// change "::" to ":| |:"
		// and    "|1" to "| [1"
		if (bt == "::" || bt == ":|"
		 || t) {
			ir = 0
			bt = t ? '|' : "::"
			while (1) {
				v = vn[ir++]
				if (v == undefined)
					break
				s = vtb[v]
				if (!s || s.invis
				 || s.bar_type != bt
				 || (bt == '|' && !s.text))
					continue
				s2 = clone(s)
				if (bt == "::") {
					s.bar_type = ":|"
					s2.bar_type = "|:"
				} else {
//					s.bar_type = '|'
					delete s.text
					delete s.rbstart
					s2.bar_type = '['
					s2.invis = 1 //true
					s2.xsh = 0
				}
				s2.next = s.next
				if (s2.next)
					s2.next.prev = s2
				s2.prev = s
				s.next = s2
			}
		} else {
			error(1, s, "Different bars $1 and $2",
				(bt + (t || '')), (s.bar_type + (s.text || '')))
		}
	} // b_chk()

	// set the first symbol of each voice
	for (v = 0; v < nv; v++) {
		vtb[v] = s = voice_tb[v].sym
		if (sy.voices[v]) {
			vn[sy.voices[v].range] = v
			if (!fmt && s)
				prev.fmt = fmt = s.fmt	// starting format
		}
	}
	if (!fmt)
		return				// no symbol yet

	// insert the first staff system in the top voice
	v = p_v.v
	p_v.sym = tsfirst = s = prev
	if (s.next)
		s.next.prev = s
	else
		p_v.last_sym = s

	// if Q: from tune header, put it at start of the music
	// after the staff system
	s = glovar.tempo
	if (s) {
		s.v = v
		s.p_v = p_v
		s.st = 0
		s.time = 0
		s.prev = prev
		s.next = prev.next
		if (s.next)
			s.next.prev = s
		else
			p_v.last_sym = s
		s.prev.next = s
		s.fmt = fmt
//		glovar.tempo = null
		vtb[v] = s
	}

	// if only one voice, quickly create the time links
	if (nv == 1) {
		s = tsfirst
		s.ts_next = s.next
		while (1) {
			s = s.next
			if (!s)
				return
			if (s.time != s.prev.time
			 || s.prev.typ.w
			 || s.typ == C.grace && s.prev.typ == C.grace)
				s.seqst = 1 //true
			s.ts_prev = s.prev
			s.ts_next = s.next
		}
		// not reached
	}

	// loop on the symbols of all voices
	while (1) {
		if (new_sy) {
			sy = new_sy
			new_sy = null
			vn.length = 0
			for (v = 0; v < nv; v++) {
				if (!sy.voices[v])
					continue
				vn[sy.voices[v].range] = v
			}
		}

		// search the min time and symbol weight
		wmin = time = 10000000		// big int
		ir = 0
		while (1) {
			v = vn[ir++]
			if (v == undefined)
				break
			s = vtb[v]
			if (!s || s.time > time)
				continue
			w = s.typ.w
			if (s.typ == C.grace && s.next && s.next.typ == C.grace)
				w--
			if (s.time < time) {
				time = s.time
				wmin = w
			} else if (w < wmin) {
				wmin = w
			}
		}

		if (wmin > 127)
			break			// done

		// check the type of the measure bars
		if (wmin == 6)			// !! weight of bars
			b_chk()

		// link the vertical sequence
		ir = 0
		while (1) {
			v = vn[ir++]
			if (v == undefined)
				break
			s = vtb[v]
			if (!s
			 || s.time != time)
				continue
			w = s.typ.w
			if (!w
			 && s.typ == C.grace && s.next && s.next.typ == C.grace)
				w--
			if (w != wmin)
				continue
			if (s.typ == C.staves)
				new_sy = s.sy
			if (fl) {
				fl = 0
				s.seqst = 1 //true
			}
			s.ts_prev = prev
			prev.ts_next = s
			prev = s

			vtb[v] = s.next
		}
		if (wmin)			// if some width
			fl = 1 //true		// start a new sequence
	}
} // sort_all()

// adjust some voice elements before sorting by time
// (possible hook)
Mus.prototype.voice_adj = function(sys_chg) {
    var	p_voice, s, s2, v, sl

	// do the note mapping stuff
	function do_map(s) {
	    var	m, nt

		for (m = 0; m <= s.nhd; m++) {
			nt = s.notes[m]
			set_map(s.p_v, nt, nt.acc, 1)	// transpose done
		}
	} // do_map()

	// insert the delayed P: and Q: in the top_voice
	function ins_pq() {
	    var	s, s2,
		p_v = voice_tb[par_sy.top_voice]

		while (1) {
			s = parse.pq_d.shift()
			if (!s)
				break
			for (s2 = p_v.sym; ; s2 = s2.next) {
				if (s2.time >= s.time
				 && s2.dur) {
					s.next = s2
					s.prev = s2.prev
					s.prev.next =
						s2.prev = s
					s.v = s2.v
					s.p_v = p_v
					s.st = s2.st
					break
				}
			}
		}
	} // ins_pq()

	// set the duration of the notes under a feathered beam
	function set_feathered_beam(s1) {
	    var	s, s2, t, d, b, i, a,
		d = s1.dur,
		n = 1

		// search the end of the beam
		for (s = s1; s; s = s.next) {
			if (s.beam_end || !s.next)
				break
			n++
		}
		if (n <= 1) {
			delete s1.feathered_beam
			return
		}
		s2 = s
		b = d / 2		// smallest note duration
		a = d / (n - 1)		// delta duration
		t = s1.time
		if (s1.feathered_beam > 0) {		// !beam-accel!
			for (s = s1, i = n - 1;
			     s != s2;
			     s = s.next, i--) {
				d = ((a * i) | 0) + b
				s.dur = d
				s.time = t
				t += d
			}
		} else {				// !beam-rall!
			for (s = s1, i = 0;
			     s != s2;
			     s = s.next, i++) {
				d = ((a * i) | 0) + b
				s.dur = d
				s.time = t
				t += d
			}
		}
		s.dur = s.time + s.dur - t
		s.time = t
	} // end set_feathered_beam()

	// ----- voice_adj -----

	// terminate voice cloning
	if (curvoice && curvoice.clone) {
		parse.istart = parse.eol
		do_cloning()
	}

	// if only one voice and a time skip,
	// fill the voice with the sequence "Z |" (multi-rest and bar)
	if (par_sy.one_v)			// if one voice
		fill_mr_ba(voice_tb[par_sy.top_voice])

	if (parse.pq_d)
		ins_pq()			// insert delayed P: and Q:

	for (v = 0; v < voice_tb.length; v++) {
		p_voice = voice_tb[v]
		if (!sys_chg) {			// if not %%score
			delete p_voice.eoln
			while (1) {		// set the end of slurs
				sl = p_voice.sls.shift()
				if (!sl)
					break
				s = sl.ss
//				error(1, s, "Lack of ending slur(s)")
				if (!s.sls)
					s.sls = []
				sl.loc = 'o'		// no slur end
				s.sls.push(sl)
			}
		} // not %%score
		for (s = p_voice.sym; s; s = s.next) {
			if (s.time >= staves_found)
				break
		}
		for ( ; s; s = s.next) {

			// if the symbol has a sequence weight smaller
			// than the one of the bar
			// and if there is a time skip,
			// then, add an invisible bar before it
			if (s.typ.w < 5
			 && s.typ != C.staves
			 && s.typ != C.clef
			 && s.time			// not at start of tune
			 && (!s.prev || s.time > s.prev.time + s.prev.dur)) {
				s2 = new_sym({
					typ: C.bar,
					bar_type: "[]",
					v: s.v,
					p_v: s.p_v,
					st: s.st,
					time: s.time,
					next: s,
					prev: s.prev,
					fmt: s.fmt,
					invis: 1
				})
				if (s.prev)
					s.prev.next = s2
				else
					voice_tb[s.v].sym = s2
				s.prev = s2
			}

			switch (s.typ) {
			case C.grace:
				if (!cfmt.graceword)
					continue
				for (s2 = s.next; s2; s2 = s2.next) {
					switch (s2.typ) {
					case C.space:
						continue
					case C.note:
						if (!s2.a_ly)
							break
						s.a_ly = s2.a_ly
						s2.a_ly = null
						break
					}
					break
				}
				break
			case C.note:
				if (s.feathered_beam)
					set_feathered_beam(s)
				if (p_voice.map
				 && maps[p_voice.map])
					do_map(s)
				break
			}

		}
	}
} // voice_adj()

// -- create a new staff system --
function new_syst() {
    var	st, v, sy_staff, p_voice,
	sy_new = {
		voices: [],
		staves: [],
		top_voice: 0,
		mxlb: []
	}

	if (!par_sy) {				// first staff system
		cur_sy = par_sy = sy_new
		return
	}

	// update the previous system
	for (v = 0; v < voice_tb.length; v++) {
		if (par_sy.voices[v]) {
			st = par_sy.voices[v].st
			sy_staff = par_sy.staves[st]
			p_voice = voice_tb[v]

			sy_staff.staffnonote = p_voice.staffnonote
			if (p_voice.staffscale)
				sy_staff.staffscale = p_voice.staffscale
		}
	}
	for (st = 0; st < par_sy.staves.length; st++) {
		sy_new.staves[st] = clone(par_sy.staves[st])
		sy_new.staves[st].flags = 0
	}
	par_sy.next = sy_new
	par_sy = sy_new
} // new_syst()

// -- set the bar numbers --
// (possible hook)
Mus.prototype.set_bar_num = function() {
    var	s, s2, rep_tim, k, n, nu, txt,
	tim = 0,			// time of the previous bar
	bar_num = gene.nbar,		// number of the first measure
	bn = 1,				// bar number (sequential)
	bar_tim = 0,			// time of previous repeat variant
	ptim = 0,			// time of previous bar
	wmeasure = voice_tb[cur_sy.top_voice].meter.wmeasure

	// check the measure duration
	function check_meas() {
	    var	s3

		if (tim > ptim + wmeasure
		 && s.prev.typ != C.mrest)
			return 1 //true

		// the measure is too short,
		// check if there is a bar a bit further
		for (s3 = s.next; s3 && s3.time == s.time; s3 = s3.next)
			;
		for ( ; s3 && !s3.bar_type; s3 = s3.next)
			;
		return s3 && (s3.time - bar_tim) % wmeasure
	}

	// don't count a bar at start of tune
	for (s = tsfirst; ; s = s.ts_next) {
		if (!s)
			return
		switch (s.typ) {
		case C.meter:
			wmeasure = s.wmeasure
		default:
			continue
		case C.bar:
			if (s.bar_num)
				bar_num = s.bar_num	// (%%setbarnb)
			if (!s.time || s.invis || s.bar_dotted)
				continue
			break
		}
		break
	}

	if (s.time && s.time < wmeasure)		// if anacrusis
		bar_tim = s.time

	// set the measure numbers
	for ( ; s; s = s.ts_next) {
		switch (s.typ) {
		case C.bar:
			if (s.time <= tim)
				break			// already seen
			tim = s.time

			nu = 1 //true			// no num update
			txt = ""
			for (s2 = s; s2; s2 = s2.next) {
				if (s2.time > tim)
					break
				if (!s2.bar_type)
					continue
				if (s2.bar_type != '[')
					nu = 0 //false	// do update
				if (s2.text)
					txt = s2.text
			}

			if (s.bar_num) {
				bar_num = s.bar_num	// (%%setbarnb)
				ptim = bar_tim = tim
				s.bar_n = ++bn		// continuous bar number
				break
			}
			if (wmeasure == 1) {		// if M:none
				if (s.bar_dotted)
					break
				if (txt) {
					if (!cfmt.contbarnb) {
						if (txt[0] == '1')
							rep_tim = bar_num
						else
							bar_num = rep_tim
					}
				}
				if (!nu) {
					s.bar_num = ++bar_num
					s.bar_n = ++bn	// continuous bar number
				}
				break
			}

			n = bar_num + (tim - bar_tim) / wmeasure
			k = n - (n | 0)
			if (cfmt.checkbars
			 && k
			 && check_meas())
				error(0, s, "Bad measure duration")
			if (tim > ptim + wmeasure) {	// if more than one measure
				n |= 0
				k = 0
				bar_tim = tim		// re-synchronize
				bar_num = n
			}
			if (txt) {
				if (txt[0] == '1') {
					if (!cfmt.contbarnb)
						rep_tim = tim - bar_tim
					if (!nu)
						s.bar_num = n
				} else {
					if (!cfmt.contbarnb)
						bar_tim = tim - rep_tim
					n = bar_num + (tim - bar_tim) / wmeasure
					if (n == (n | 0))
						s.bar_num = n
				}
			} else if (n == (n | 0)) {
				s.bar_num = n
			}
			if (!k)
				ptim = tim
			s.bar_n = ++bn		// continuous bar number
			break
		case C.meter:
			if (wmeasure != 1)		// if not M:none
				bar_num += (s.time - bar_tim) / wmeasure
			bar_tim = s.time
			wmeasure = s.wmeasure
			while (s.ts_next && s.ts_next.wmeasure)
				s = s.ts_next
			break
		}
	}
} // set_bar_num()

// convert a note to ABC
function not2abc(pit, acc) {
    var	i,
	nn = ''

	if (acc && acc != 3) {
		if (typeof acc != "object") {
			nn = ['__', '_', '', '^', '^^'][acc + 2]
		} else {
			i = acc[0]
			if (i > 0) {
				nn += '^'
			} else {
				nn += '_'
				i = -i
			}
			nn += i + '/' + acc[1]
		}
	}
	nn += C.ntb[(pit + 93) % 7]
	for (i = pit; i >= 5; i -= 7)
		nn += "'"
	for (i = pit; i < -2; i += 7)
		nn += ","
	return nn
} // not2abc()

// note mapping
// %%map map_name note [print [note_head]] [param]*
function get_map(text) {
	if (!text)
		return

    var	i, nt, notes, map, ia, ns,
	ty = '',
	a = text.split(/\s+/)

	if (a.length < 3) {
		syntax(1, errs.not_enough_p)
		return
	}
	ns = a[1]
	if (ns[0] == '*' || ns.indexOf("all") == 0) {
		ns = 'all'
	} else {
		if (ns.indexOf("octave,") == 0	// remove the octave part
		 || ns.indexOf("key,") == 0) {
			ty = ns[0]
			ns = ns.split(',')[1]
			ns = ns.replace(/[,']+/, '').toUpperCase() //'
			if (ns.indexOf("key,") == 0)
				ns = ns.replace(/[=^_]+/, '')
		}
		tm.reap.lastIndex = 0
		ia = tm.reap.exec(ns)
		nt = get_acc_pit(ia)
//		if (!nt) {
//			syntax(1, "Bad note in %%map")
//			return
//		}
		ns = ty + not2abc(nt.pit, nt.acc)
	}

	notes = maps[a[0]]
	if (!notes)
		maps[a[0]] = notes = {}
	map = notes[ns]
	if (!map)
		notes[ns] = map = []

	// try the optional 'print' and 'heads' parameters
	a.shift()
	a.shift()
	if (!a.length)
		return
	a = info_split(a.join(' '))
	i = 0
	if (a[1] != '=') {
		if (a[0][0] != '*') {
			tm.reap.lastIndex = 0
			ia = tm.reap.exec(a[0])
			map[1] = get_acc_pit(ia)
		}
		if (!a[1])
			return
		i++
		if (a[2] != '=') {
			map[0] = a[1].split(',')	// heads
			i++
		}
	}

	for (; i < a.length; i++) {
		if (a[i + 1] != '='
		 || !a[i + 2]) {
			syntax(1, errs.not_enough_p)
			break
		}
		switch (a[i]) {
		case "heads":
			map[0] = a[i + 2].split(',')
			break
		case "print":
		case "play":
		case "print_notrp":
			tm.reap.lastIndex = 0
			ia = tm.reap.exec(a[i + 2])
			nt = get_acc_pit(ia)
			if (a[i][5] == '_')		// if print no transpose
				nt.notrp = 1 //true
			if (a[i][1] == 'r')
				map[1] = nt
			else
				map[3] = nt
			break
		case "color":
			map[2] = a[i + 2]
			break
		}
		i += 2
	}
} // get_map()

// set the pitch/accidental/MIDI of a note according to the voice mapping
//
// The global 'maps' object is indexed by the map name.
// Its content is an object ('map') indexed from the map type:
// - normal = ABC note
// - octave = 'o' + ABC note in C..B interval
// - key    = 'k' + scale index
// - all    = 'all'
// The 'map' is stored in the note. It is an array of
//	[0] array of heads (glyph names)
//	[1] print (note)
//	[2] color
//	[3] play (note)
function set_map(p_v, note, acc,
		trp) {			// transpose done
    var	nn = not2abc(note.pit, acc),
	map = maps[p_v.map]		// never null

	if (!map[nn]) {
		nn = 'o' + nn.replace(/[',]+/, '')	// ' octave
		if (!map[nn]) {
			nn = 'k' + C.ntb[(note.pit + 93
					- p_v.ckey.k_sf * 11) % 7]
			if (!map[nn]) {
				nn = 'all'		// 'all'
				if (!map[nn])
					return
			}
		}
	}
	map = map[nn]
	if (!trp) {				// transpose not done yet
		if (map[1]) {			// if note shift
			note.pit = map[1].pit
			note.acc = map[1].acc
			if (map[1].notrp) {
				note.notrp = 1 //true	// no transpose
				note.noplay = 1 //true	// no play
			}
		}
		return
	}
	note.map = map

	if (map[2])				// if color
		note.color = map[2]
	nn = map[3]
	if (nn)					// if play map
		note.midi = pit2mid(nn.pit + 37, nn.acc)
} // set_map()

// transpose and set the MIDI pitch of a note
function set_midi(nt, acc_tie) {
    var	s, apit, ap,
	a = nt.acc				// keep the original accidental

	if (curvoice.octave)
		nt.pit += curvoice.octave * 7

	apit = nt.pit + 37			// pitch from C-1
	ap = cfmt["propagate-accidentals"][0] == 'p' ? apit % 7 : apit

	if (!a) {
		a = curvoice.acc[ap]		// accidental from the measure
			|| curvoice.ckey.k_map[apit % 7] // or from the key
			|| 0
//		if (!a)
//			nt.acc = 0
	}
	if (a && cfmt["propagate-accidentals"][0] != 'n')
		curvoice.acc[ap] = a			// set for the measure

	if (acc_tie && acc_tie[ap])
		a = acc_tie[ap]				// accidental from a tie

	// set the MIDI pitch
	if (!nt.midi)		// if not map play
		nt.midi = pit2mid(apit, a)

	// transpose
	if (curvoice.map
	 && maps[curvoice.map])
		set_map(curvoice, nt, a)	// transpose not done
	if (curvoice.tr_sco) {
		if (!nt.notrp) {
			a = nt_trans(nt, a)
			if (a == -3) {		// if triple sharp/flat
				error(1, s, "triple sharp/flat")
				a = nt.acc > 0 ? 1 : -1
				nt.pit += a
				nt.acc = a
			}
		}
	}
	if (curvoice.tr_snd)
		nt.midi += curvoice.tr_snd

	if (a) {
		switch (cfmt["writeout-accidentals"][1]) {
		case 'd':			// added
			s = curvoice.ckey
			if (!s.k_a_acc)
				break
			for (n = 0; n < s.k_a_acc.length; n++) {
				if ((s.k_a_acc[n].pit - nt.pit) % 7 == 0) {
					nt.acc = a
					break
				}
			}
			break
		case 'l':			// all
			nt.acc = a
			break
		}
	}
} // set_midi()

// get a abcMIDI compatible transposition value as a base-40 interval
// The value may be only
//	[+|-]<number of semitones>[s|f]
function get_transp(parm) {
	if (parm[0] == '0')
		return 0
	if ("123456789-+".indexOf(parm[0]) >= 0) {	// by semi-tone
	    var	val = parseInt(parm)
		if (isNaN(val) || val < -36 || val > 36) {
//fixme: no source reference...
			syntax(1, errs.bad_transp)
			return
		}
		val += 36
		val = ((val / 12 | 0) - 3) * 40 + tm.ib40[val % 12]
		if (parm.slice(-1) == 'b')
			val += 4
		return val
	}
	// return undefined
} // get_transp()

// -- process a pseudo-comment (%% or I:) --
// (possible hook)
Mus.prototype.do_pscom = function(text) {
    var	h1, val, s, cmd, parm, n, k, b

	cmd = text.match(tm.recp)	// => [ match, cmd, param ]
	if (!cmd)
		return
	parm = cmd[2]
	cmd = cmd[1]

	// ignore the command if the voice is ignored,
	// but not if %%score/%%staves!
	if (curvoice && curvoice.ignore) {
		switch (cmd) {
		case "staves":
		case "score":
			break
		default:
			return
		}
	}

	if (parm.slice(-5) == ' lock') {
		fmt_lock[cmd] = 1
		parm = parm.slice(0, -5).trim()
	} else if (fmt_lock[cmd]) {
		return
	}

	switch (cmd) {
	case "clef":
		if (parse.state >= 2) {
			s = new_clef(parm)
			if (s)
				get_clef(s)
		}
		return
	case "deco":
		deco_add(parm)
		return
	case "linebreak":
		set_linebreak(parm)
		return
	case "map":
		get_map(parm)
		return
	case "maxsysstaffsep":
	case "sysstaffsep":
		if (parse.state == 3) {
			val = get_unit(parm)
			if (isNaN(val)) {
				syntax(1, errs.bad_val, "%%" + cmd)
				return
			}
			par_sy.voices[curvoice.v][cmd[0] == 'm' ? "maxsep" : "sep"] =
					val
			return
		}
		break
	case "multicol":
		switch (parm) {
		case "start":
		case "new":
		case "end":
			break
		default:
			syntax(1, "Unknown keyword '$1' in %%multicol", parm)
			return
		}
		s = new_sym({
			typ: C.block,
			subtype: "mc_" + parm
		})
		if (parse.state >= 2) {
			curvoice = voice_tb[0]
			curvoice.eoln = 1 //true
			sym_link(s)
			return
		}
		set_ref(s)
		self.block_gen(s)
		return
	case "ottava":
		if (parse.state != 3)
			return
		n = parseInt(parm)
		if (isNaN(n) || n < -2 || n > 2
		 || (!n && !curvoice.ottava)) {
			syntax(1, errs.bad_val, "%%ottava")
			return
		}
		k = n
		if (n) {
			curvoice.ottava = n
		} else {
			n = curvoice.ottava
			curvoice.ottava = 0
		}
		a_dcn.push(["15mb", "8vb", "", "8va", "15ma"][n + 2]
			+ (k ? '(' : ')'))
		return
	case "repbra":
		if (curvoice)
			curvoice.norepbra = !get_bool(parm)
		return
	case "repeat":
		if (parse.state != 3)
			return
		if (!curvoice.last_sym) {
			syntax(1, "%%repeat cannot start a tune")
			return
		}
		if (!parm.length) {
			n = 1
			k = 1
		} else {
			b = parm.split(/\s+/)
			n = +b[0]
			k = +b[1]
			if (isNaN(n) || n < 1
			 || (curvoice.last_sym.bar_type
			  && n > 2)) {
				syntax(1, "Incorrect 1st value in %%repeat")
				return
			}
			if (isNaN(k)) {
				k = 1
			} else {
				if (k < 1) {
					syntax(1, "Incorrect 2nd value in %%repeat")
					return
				}
			}
		}
		parse.repeat_n = curvoice.last_sym.bar_type ? n : -n
		parse.repeat_k = k
		return
	case "sep":
	    var	h2, len, values, lwidth

		set_page()
		lwidth = img.width - img.lm - img.rm
		h1 = h2 = len = 0
		if (parm) {
			values = parm.split(/\s+/)
			h1 = get_unit(values[0])
			if (values[1]) {
				h2 = get_unit(values[1])
				if (values[2])
					len = get_unit(values[2])
			}
			if (isNaN(h1) || isNaN(h2) || isNaN(len)) {
				syntax(1, errs.bad_val, "%%sep")
				return
			}
		}
		if (h1 < 1)
			h1 = 14
		if (h2 < 1)
			h2 = h1
		if (len < 1)
			len = 90
		if (parse.state >= 2) {
			s = new_block(cmd)
			s.x = (lwidth - len) / 2 / cfmt.scale
			s.l = len / cfmt.scale
			s.sk1 = h1
			s.sk2 = h2
			return
		}
		vskip(h1)
		output += '<path class="stroke"\n\td="M'
		out_sxsy((lwidth - len) / 2 / cfmt.scale, ' ', 0)
		output += 'h' + (len / cfmt.scale).toFixed(1) + '"/>\n'
		vskip(h2)
		blk_flush()
		return
	case "setbarnb":
		val = +parm
		if (isNaN(val) || val < 1) {
			syntax(1, "Bad %%setbarnb value")
			break
		}
		glovar.new_nbar = val		// number of the next measure
		return
	case "staff":
		if (parse.state != 3)
			return
		val = +parm
		if (isNaN(val)) {
			syntax(1, "Bad %%staff value '$1'", parm)
			return
		}
		var st
		if (parm[0] == '+' || parm[0] == '-')
			st = curvoice.cst + val
		else
			st = val - 1
		if (st < 0 || st > nstaff) {
			syntax(1, "Bad %%staff number $1 (cur $2, max $3)",
					st, curvoice.cst, nstaff)
			return
		}
		delete curvoice.floating
		curvoice.cst = st
		return
	case "staffbreak":
		if (parse.state != 3)
			return
		s = new_sym({
			typ: C.stbrk
		})
		if (parm.slice(-1) == 'f') {
			s.stbrk_forced = 1
			parm = parm.replace(/\sf$/, '')
		}
		if (parm) {
			val = get_unit(parm)
			if (isNaN(val)) {
				syntax(1, errs.bad_val, "%%staffbreak")
				return
			}
			s.xmx = val
		} else {
			s.xmx = 14
		}
		sym_link(s)
		return
	case "tacet":
		if (parm[0] == '"')
			parm = parm.slice(1, -1)
		// fall thru
	case "stafflines":
	case "staffscale":
	case "staffnonote":
		set_v_param(cmd, parm)
		return
	case "staves":
	case "score":
		if (!parse.state)
			return
		if (parse.scores && parse.scores.length > 0) {
			text = parse.scores.shift()
			cmd = text.match(tm.recp)	// => [ match, cmd, param ]
			parm = cmd[2]
			cmd = cmd[1]
		}
		get_staves(cmd, parm)
		return
	case "center":
	case "text":
		k = cmd[0] == 'c' ? 'c' : cfmt.textoption
		set_font("text")
		if (parse.state >= 2) {
			s = new_block("text")
			s.text = parm
			s.opt = k
			s.font = cfmt.textfont
			return
		}
		write_text(parm, k)
		return
	case "tune":
//fixme: to do
		return
	case "user":
		set_user(parm)
		return
	case "voicecolor":
		if (curvoice)
			curvoice.color = parm
		return
	case "vskip":
		val = get_unit(parm)
		if (isNaN(val)) {
			syntax(1, errs.bad_val, "%%vskip")
			return
		}
		if (val < 0) {
			syntax(1, "%%vskip cannot be negative")
			return
		}
		if (parse.state >= 2) {
			s = new_block(cmd)
			s.sk = val
			return
		}
		vskip(val)
		return
	case "newpage":
	case "leftmargin":
	case "rightmargin":
	case "pagescale":
	case "pagewidth":
	case "scale":
	case "staffwidth":
		if (parse.state >= 2) {
			s = new_block(cmd)
			s.parm = parm
			return
		}
		if (cmd == "newpage") {
			blk_flush()
			if (user.page_format)
				blkdiv = 2	// start the next SVG in a new page
			return
		}
		break
	}
	self.set_format(cmd, parm)
} // do_pscom()

// treat the %%beginxxx / %%endxxx sequences
// (possible hook)
Mus.prototype.do_begin_end = function(type,
			opt,
			text) {
	var i, j, action, s

	switch (type) {
	case "js":
		js_inject(text)
		break
	case "ml":
		if (cfmt.pageheight) {
			syntax(1, "Cannot have %%beginml with %%pageheight")
			break
		}
		if (parse.state >= 2) {
			s = new_block(type)
			s.text = text
		} else {
			blk_flush()
			if (user.img_out)
				user.img_out(text)
		}
		break
	case "svg":
		j = 0
		while (1) {
			i = text.indexOf('<style', j)
			if (i < 0)
				break
			i = text.indexOf('>', i)
			j = text.indexOf('</style>', i)
			if (j < 0) {
				syntax(1, "No </style> in %%beginsvg sequence")
				break
			}
			style += text.slice(i + 1, j).replace(/\s+$/, '')
		}
		j = 0
		while (1) {
			i = text.indexOf('<defs>\n', j)
			if (i < 0)
				break
			j = text.indexOf('</defs>', i)
			if (j < 0) {
				syntax(1, "No </defs> in %%beginsvg sequence")
				break
			}
			defs_add(text.slice(i + 6, j))
		}
		break
	case "text":
		action = get_textopt(opt)
		if (!action)
			action = cfmt.textoption
		set_font("text")
		if (text.indexOf('\\') >= 0)
			text = cnv_escape(text)
		if (parse.state > 1) {
			s = new_block(type)
			s.text = text
			s.opt = action
			s.font = cfmt.textfont
			break
		}
		write_text(text, action)
		break
//	default:
//		if (typeof tm["be" + type] == "function") { // if in a module
//			tm["be" + type](self)		// switch
//			self.do_begin_end(type, opt, text)
//		}
//		break
	}
} // do_begin_end()

// -- generate a tune --
function generate() {
    var s, v

	if (a_dcn.length) {
		syntax(1, "Decoration(s) without symbol: $1", a_dcn)
		a_dcn.length = 0
	}

	if (parse.tp) {
		syntax(1, "No end of tuplet")
		s = parse.tps
		if (s)
			delete s.tp
		parse.tp = null
	}

	if (vover) {
		syntax(1, "No end of voice overlay")
		get_vover(vover.bar ? '|' : ')')
	}

	self.voice_adj()
	sort_all()			// define the time / vertical sequences

	write_heading()			// output the tune header
//fixme: removed because problem with tune header at the bottom of a page?
//fixme: 24-02-17 back again
	blk_flush()			// in a specific SVG

	if (tsfirst) {
		for (v = 0; v < voice_tb.length; v++) {
			if (!voice_tb[v].key)
				voice_tb[v].key = parse.key	// set the starting key
		}
		if (user.annotate)
			put_ref = anno_out
		self.set_bar_num()

		if (info.P)
			tsfirst.parts = info.P[0].t	// for play

		if (user.img_out)		// if SVG generation
			self.output_music()

		tunes.push([tsfirst, voice_tb, info, cfmt])
	}

	// finish the generation
	set_page()			// the page layout may have changed
	if (info.W)
		put_words(info.W)
	put_history()

	parse.state = 0			// file header (must be before blk_flush)
	blk_flush()			// (force end of block)
	tsfirst = null
} // generate()

// transpose a key
function key_trans() {
    var	i, n, a_acc, b40, d,
	s = curvoice.ckey,			// current key
	ti = s.time || 0

	if (s.k_bagpipe || s.k_drum)
		return				// no transposition

	// set the score transposition
	n = (curvoice.score | 0)		// new transposition
		+ (curvoice.shift | 0)
		+ (cfmt.transp | 0)
	if ((curvoice.tr_sco | 0) == n) {	// if same transposition
		s.k_sf = curvoice.ckey.k_sf
		return
	}

	// get the current key or create a new one
	if (is_voice_sig()) {			// if no symbol yet
		curvoice.key = s		// new root key of the voice
	} else if (curvoice.time != ti) {	// if no K: at this time
		s = clone(s.orig || s)		// new key
		if (!curvoice.new)
			s.k_old_sf = curvoice.ckey.k_sf
		sym_link(s)
	}
	curvoice.ckey = s			// current key

	if (cfmt.transp && curvoice.shift)	// if %%transpose and shift=
		syntax(0, "Mix of old and new transposition syntaxes")

	// define the new key
	curvoice.tr_sco = n			// b40 interval

	n = tm.b40l5[(n + 202) % 40]		// transpose in the line of fifth
		+ s.orig.k_sf			// + old = new sf
	if (n < -7) {
		n += 12
		curvoice.tr_sco -= 4
	} else if (n > 7) {
		n -= 12
		curvoice.tr_sco += 4
	}
	if (!s.k_none)
		s.k_sf = n
	for (b40 = 0; b40 < 40; b40++) {
		if (tm.b40l5[b40] == n)
			break
	}
	s.k_b40 = b40

	// transpose the accidental list
	if (!s.k_a_acc)
		return
	d = b40 - s.orig.k_b40
	a_acc = []
	for (i = 0; i < s.k_a_acc.length; i++) {
		b40 = tm.pab40(s.k_a_acc[i].pit, s.k_a_acc[i].acc) + d
		a_acc[i] = {
			pit: tm.b40p(b40),
			acc: tm.b40a(b40) || 3
		}
	}
	s.k_a_acc = a_acc
} // key_transp()

// fill a voice with a multi-rest and a bar
function fill_mr_ba(p_v) {
    var	v, p_v2,
	mxt = 0

	for (v = 0; v < voice_tb.length; v++) {
		if (voice_tb[v].time > mxt) {
			p_v2 = voice_tb[v]
			mxt = p_v2.time
		}
	}
	if (p_v.time >= mxt)
		return

    var	p_v_sav = curvoice,
	dur = mxt - p_v.time,
	s = new_sym({
		typ: C.mrest,
//		stem: 0,
//		multi: 0,
		frm: 1, //true			// full measure rest
		dur: dur,
		dur_orig: dur,
		nmes: dur / p_v.wmeasure,
		notes: [{
			pit: 0,
			dur: dur
		}],
		tacet: p_v.tacet
	}),
	s2 = new_sym({
		typ: C.bar,
		bar_type: '|',
//		multi: 0
	})

	if (p_v2.last_sym.bar_type)
		s2.bar_type = p_v2.last_sym.bar_type
//	s2.soln = p_v2.last_sym.soln

	glovar.mrest_p = 1 //true

	curvoice = p_v
	sym_link(s)
	sym_link(s2)

	curvoice = p_v_sav
} // fill_mr_ba()

// -- get staves definition (%%staves / %%score) --
function get_staves(cmd, parm) {
    var	s, p_voice, p_voice2, i, flags, v, vid, a_sy, a,
	nv = voice_tb.length,
	maxtime = 0

	// return the greatest label
	function elen(e) {
	    var	i = 0,
		l = ""

		if (e.indexOf("\\n") < 0)
			return e
		e = e.split("\\n")
		while (1) {
			if (!e[i])
				return l
			if (e[i].length > l.length)
				l = e[i]
			i++
		}
		// not reached
	} // elen()

	// if sequence with many voices, load the other voices
	if (curvoice && curvoice.clone) {
		i = parse.eol
		parse.eol = parse.bol		// remove the %%staves line
		do_cloning()
		parse.eol = i
	}

	if (parm) {
		a = parm.match(/[^[\]|{}()*+"'\s]+|"[^"]*"|[^\s]/g)
		if (!a) {
			syntax(1, errs.bad_val, "%%score")
			return
		}
	}

	// create a new staff system
	for (v = 0; v < nv; v++) {
		p_voice = voice_tb[v]
		if (p_voice.time > maxtime)
			maxtime = p_voice.time
	}
	if (!maxtime) {				// if first %%staves
		par_sy.staves = []
		par_sy.voices = []
	} else {

		self.voice_adj(1)

		// create a new staff system
		// and link the 'staves' symbol in a voice which is seen from
		// the previous system
		for (v = 0; v < par_sy.voices.length; v++) {
			if (par_sy.voices[v]) {
				curvoice = voice_tb[v]
				break
			}
		}

		curvoice.time = maxtime
		s = new_sym({
			typ: C.staves
		})

		sym_link(s)		// link the staves in this voice
//		par_sy.nstaff = nstaff	// (already set)

		// if no parameter, duplicate the current staff system
		// and synchronize the voices
		if (!parm) {
//			s.sy = clone(par_sy, 1)	// clone the staves and voices?
			s.sy = clone(par_sy)
			par_sy.next = s.sy
			par_sy = s.sy
			par_sy.sync = 1 //true			// (for toabc)
			staves_found = maxtime
			for (v = 0; v < nv; v++)
				voice_tb[v].time = maxtime
			if (cmd != "sync")			// if not V: sync
				curvoice = voice_tb[par_sy.top_voice]
			return
		}

		new_syst()
		s.sy = par_sy
	}

	staves_found = maxtime

	// reset the old voices
	for (v = 0; v < nv; v++) {
		p_voice = voice_tb[v]
		delete p_voice.second
		delete p_voice.floating
		if (p_voice.ignore) {
			p_voice.ignore = 0 //false
			s = p_voice.sym
			if (s) {
				while (s.next)
					s = s.next
			}
			p_voice.last_sym = s		// set back the last symbol
		}
	}

	// check the groups and set the voices
    var	e, in_par, gr1, gr2, l,
	st = -1,
	range = 0,
	lb = 0,
	mxlb = ["", ""]			// greatest labels

	for (i = 0; i < a.length; i++) {
		e = a[i]
		switch (e[0]) {
		case "'":
		case '{':
			if (gr1 == e || gr2 == e) {
				syntax(1, errs.misplaced, e)
				err = 1
				break
			}
			// fall thru
		case '[':
			if (gr1 && gr2) {
				syntax(1, errs.misplaced, e)
				err = 1
				break
			}
			if (gr1)
				gr2 = e
			else
				gr1 = e
			lb = 0
			break
		case '"':
			l = elen(e)
			if (!lb) {
				if (l.length > mxlb[0].length)
					mxlb[0] = l
				lb++
			} else {
				if (l.length > mxlb[1].length)
					mxlb[1] = l
			}
			break
		case '(':
			if (in_par) {
				syntax(1, errs.misplaced, e)
				err = 1
				break
			}
			in_par = 1
			lb = 0
			break
		case '}':
			par_sy.staves[st].brace = 2	// end of brace
			// fall thru
		case ']':
			if (gr2) {
				if ((gr2 == '[' && e == ']')
				 || (gr2 == '{' && e == '}')) {
					gr2 = null
					break
				}
			} else {
				if ((gr1 == '[' && e == ']')
				 || (gr1 == '{' && e == '}')) {
					gr1 = null
					break
				}
			}
			syntax(1, errs.misplaced, e)
			err = 1
			break
		case ')':
			if (!in_par) {
				syntax(1, errs.misplaced, e)
				err = 1
				break
			}
			in_par = 0
			break
		case '|':
			flags |= C.STOP_BAR
			break
		case '+':
			if (in_par)
				flags |= C.MASTER_VOICE
			break
		case '*':
			flags |= C.FL_VOICE
			break
		default:
			vid = a[i]
			p_voice = new_voice(vid)
			p_voice.time = maxtime
			v = p_voice.v

//??			a_sy[i].p_v = p_voice
			if (!par_sy.top_voice)
				par_sy.top_voice = v

			if (flags & C.FL_VOICE) {
				p_voice.floating = 
					p_voice.second = 1 //true
			} else if (!in_par || in_par == 1) {
				if (in_par)			// if first voice in (
					in_par = p_voice	// (master voice)
				par_sy.staves[++st] = {
					stafflines: p_voice.stafflines || "|||||",
					staffscale: 1,
					flags: flags & C.STOP_BAR
				}
				if (gr1 == '{' || gr2 == '{')
					par_sy.staves[st].brace = 1 // start of brace
			} else {
				if (flags & C.MASTER_VOICE)
					in_par.second = 1 //true
				else
					p_voice.second = 1 //true
			}

			flags &= ~(C.FL_VOICE | C.MASTER_VOICE | C.STOP_BAR)

			// set the staff, range and add the overlay voices
			while (1) {
				p_voice.st = p_voice.cst = st
				par_sy.voices[v] = {
					st: st,
					range: range++
				}
				p_voice = p_voice.voice_down
				if (!p_voice)
					break
				p_voice.second = 1 //true
				v = p_voice.v
			}
			lb = 0
			break
		}
	}
	if (range == 1)
		par_sy.one_v = 1 //true			// one voice
	if (gr2 != undefined)
		par_sy.grlvl = 2			// level of [ and {
	else if (gr1 != undefined)
		par_sy.grlvl = 1
	par_sy.mxlb = mxlb

	if (st < 0)
		st = 0
	par_sy.nstaff = nstaff = st

	// change the behaviour of '|' in %%score
	if (cmd[1] == 'c') {				// if %%score
		for (st = 0; st < nstaff; st++)
			par_sy.staves[st].flags ^= C.STOP_BAR
	}

	nv = voice_tb.length
	st = 0
	for (v = 0; v < nv; v++) {
		p_voice = voice_tb[v]
		if (par_sy.voices[v])
			st = p_voice.st
		else
			p_voice.st = st	// (this avoids later crashes)

		// if first %%staves
		// update the staff of the symbols with no time
		if (!maxtime) {
			for (s = p_voice.sym; s; s = s.next)
				s.st = st
		}

		if (!par_sy.voices[v])
			continue

		// set the staff of the overlay voices
		p_voice2 = p_voice.voice_down
		while (p_voice2) {
			i = p_voice2.v
			p_voice2.st = p_voice2.cst =
					par_sy.voices[i].st = st
			p_voice2 = p_voice2.voice_down
		}

		par_sy.voices[v].second = p_voice.second
		st = p_voice.st
		if (st > 0 && p_voice.norepbra == undefined
		 && !(par_sy.staves[st - 1].flags & C.STOP_BAR))
			p_voice.norepbra = 1 //true
	}

	par_sy.a = a		// (used for drawing the staff system)

	curvoice = parse.state >= 2 ? voice_tb[par_sy.top_voice] : null
} // get_staves()

// get a voice or create a clone of the current voice
function clone_voice(id) {
    var	p_v = new_voice(id)

	if (!p_v.new)
		return p_v
	delete p_v.new
	p_v.clef = clone(curvoice.clef)
	p_v.key = clone(curvoice.key)
	p_v.ckey = clone(curvoice.ckey)
	p_v.meter = curvoice.meter
	p_v.wmeasure = curvoice.wmeasure
	p_v.pos = clone(curvoice.pos)
	p_v.ulen = curvoice.ulen
	p_v.scale = curvoice.scale
	p_v.st = curvoice.st
	p_v.cst = curvoice.cst
	if (curvoice.ignore)
		p_v.ignore = curvoice.ignore
	return p_v
} // clone_voice()

// -- get a voice overlay --
function get_vover(type) {
    var	p_voice2, p_voice3, range, s, time, v, v2, v3, s2

	// treat the end of overlay
	if (type == '|'
	 || type == ')')  {
		if (!curvoice.last_note) {
			syntax(1, errs.nonote_vo)
			if (vover) {
				curvoice = vover.p_v
				vover = null
			}
			return
		}
		curvoice.last_note.beam_end = 1 //true
		if (!vover) {
			syntax(1, "Erroneous end of voice overlay")
			return
		}
		p_voice2 = vover.p_v			// main voice
		if (curvoice.time != p_voice2.time) {
//fixme: removed 2024-02-04 - voice option?
//			if (!curvoice.ignore)
//				syntax(1, "Wrong duration in voice overlay")
			if (curvoice.time > p_voice2.time)
				p_voice2.time = curvoice.time
		}
		curvoice.acc = []		// no accidental anymore

		// if the last symbols are spaces, move them to the main voice
		s = curvoice.last_sym
		if (s.typ == C.space && p_voice2.last_sym.typ != C.space) {
			s.p_v = p_voice2
			s.v = s.p_v.v
			while (s.prev.typ == C.space) {
				s = s.prev
				s.p_v = p_voice2
				s.v = s.p_v.v
			}
			s2 = s.prev
			s2.next = null
			s.prev = p_voice2.last_sym
			s.prev.next = s
			p_voice2.last_sym = curvoice.last_sym
			curvoice.last_sym = s2
		}

		curvoice = p_voice2
		vover = null
		return
	}

	// treat the full overlay start
	if (type == '(') {
		if (vover) {
			syntax(1, "Voice overlay already started")
			return
		}
		vover = {
			p_v: curvoice,
			time: curvoice.time
		}
		return
	}

	// (here is treated a new overlay - '&')
	// create the extra voice if not done yet
	if (!curvoice.last_note) {
		syntax(1, errs.nonote_vo)
		return
	}
	curvoice.last_note.beam_end = 1 //true
	p_voice2 = curvoice.voice_down
	if (!p_voice2) {
		p_voice2 = clone_voice(curvoice.id + '+')
		curvoice.voice_down = p_voice2
		p_voice2.time = 0
		p_voice2.second = 1 //true
		p_voice2.last_note = null
		v2 = p_voice2.v
		if (par_sy.voices[curvoice.v]) {	// if voice in the staff system
			par_sy.voices[v2] = {
				st: curvoice.st,
				second: 1 //true
			}
			range = par_sy.voices[curvoice.v].range
			for (v = 0; v < par_sy.voices.length; v++) {
				if (par_sy.voices[v]
				 && par_sy.voices[v].range > range)
					par_sy.voices[v].range++
			}
			par_sy.voices[v2].range = range + 1
		}
	}
	p_voice2.ulen = curvoice.ulen
	p_voice2.dur_fact = curvoice.dur_fact
	p_voice2.acc = []			// no accidental

	if (!vover) {				// first '&' in a measure
		time = p_voice2.time
		if (curvoice.ignore) {
			s = curvoice.last_bar
		} else {
			for (s = curvoice.last_sym; /*s*/; s = s.prev) {
				if (s.bar_type
				 || s.time <= time)	// (if start of tune)
					break
			}
		}
		vover = {
			bar: (s && s.bar_type) ? s.bar_type : '|',
			p_v: curvoice,
			time: s ? s.time : curvoice.time
		}
	} else {
		if (curvoice != vover.p_v
		 && curvoice.time != vover.p_v.time) {
			syntax(1, "Wrong duration in voice overlay")
			if (curvoice.time > vover.p_v.time)
				vover.p_v.time = curvoice.time
		}
	}
	p_voice2.time = vover.time
	curvoice = p_voice2
} // get_vover()

// check if a clef, key or time signature may go at start of the current voice
function is_voice_sig() {
	if (curvoice.time)
		return 0 //false
//	if (!curvoice.last_sym)
//		return 1 //true
	for (var s = curvoice.last_sym; s; s = s.prev)
		if (s.typ.w)
			return 0 //false
	return 1 //true
} // is_voice_sig()

// treat a clef found in the tune body
function get_clef(s) {
    var	s2, s3

	// special case for percussion
	if (s.clef_type == 'p') {		// if percussion clef
		s2 = curvoice.ckey
		s2.k_drum = 1 //true
		s2.k_sf = 0
		s2.k_b40 = 2
		s2.k_map = tm.keys[7]
		if (!curvoice.key)
			curvoice.key = s2	// new root key
	}

	if (!curvoice.time		// (force a clef when new voice)
	 && is_voice_sig()) {
		curvoice.clef = s
		s.fmt = cfmt
		return
	}

	// move the clef before a key and/or a (not right repeat) bar
	if (s.clef_none)
		s2 = null
	else
		for (s2 = curvoice.last_sym;
		     s2 && s2.time == curvoice.time;
		     s2 = s2.prev) {
			if (s2.typ.w)
				break
		}
	if (s2
	 && s2.time == curvoice.time		// if no time skip
	 && s2.k_sf != undefined) {
		s3 = s2				// move before a key signature
		s2 = s2.prev
	}
	if (s2
	 && s2.time == curvoice.time
	 && s2.bar_type && s2.bar_type[0] != ':')
		s3 = s2				// move before a measure bar
	if (s3) {
		s2 = curvoice.last_sym
		curvoice.last_sym = s3.prev
		sym_link(s)
		s.next = s3
		s3.prev = s
		curvoice.last_sym = s2
		if (s.soln) {
			delete s.soln
			curvoice.eoln = true
		}
	} else {
		sym_link(s)
	}

	if (s.prev)				// if not the first clef of the voice
		s.clef_small = 1 //true		// have a small clef
} // get_clef()

// treat K: (key signature)
function get_key(parm) {
    var	v, clo,
//	[s, a] = new_key(parm)	// KO with nodejs
	a = new_key(parm),
	s = a[0],
	empty = s.k_sf == undefined && !s.k_a_acc

	a = a[1]

	if (empty)
		s.invis = 1 //true		// don't display empty K:
	else
		s.orig = s			// new transposition base

	if (parse.state == 1) {			// in tune header (first K:)
		parse.key = s			// root key
		if (empty) {
			s.k_sf = 0
			s.k_none = 1 //true
			s.k_map = tm.keys[7]
		}
		for (v = 0; v < voice_tb.length; v++)
			voice_tb[v].ckey = clone(s)

		if (a.length) {
			memo_kv_parm('*', a)
			a = []
		}
		if (!glovar.ulen)
			glovar.ulen = C.BLEN / 8
		goto_tune()
	} else if (!empty) {
		if (curvoice.tr_sco)
			curvoice.tr_sco = undefined
		s.k_old_sf = curvoice.ckey.k_sf	// memorize the previous key
		curvoice.ckey = s
		sym_link(s)
	}

	// set the voice parameters
	if (!curvoice) {			// if first K:
		if (!voice_tb.length) {
			curvoice = new_voice("1")
		    var	def = 1 //true
		} else {
			curvoice = voice_tb[staves_found < 0 ? 0 : par_sy.top_voice]
		}
	}

	clo = curvoice.clone
	if (clo)
		curvoice.clone = null		// don't stop the multi-voice sequence
	get_voice(curvoice.id + ' ' + a.join(' '))
	if (clo)
		curvoice.clone = clo

	if (def)
		curvoice.default = 1 //true
} // get_key()

// get / create a new voice
function new_voice(id) {
    var	v,
	p_voice = voice_tb[0],
	n = voice_tb.length

	// if first explicit voice and no music, replace the default V:1
	if (n == 1
	 && p_voice.default) {
		p_voice.default = 0 //false
		if (id == "1")
			return p_voice
		if (!p_voice.time) {		// if no symbol yet
			voice_tb['v' + p_voice.id] = null
			p_voice.init = 0	// set back the global voice parameters
			p_voice.id = id
			voice_tb['v' + id] = p_voice
			return p_voice		// default voice
		}
	}
	if (voice_tb['v' + id])
		return voice_tb['v' + id]	// old voice

	p_voice = {
		v: voice_tb.length,
		id: id,
		time: 0,
		new: 1, //true
		pos: {
//			dyn: 0,
//			gch: 0,
//			gst: 0,
//			orn: 0,
//			stm: 0,
//			tup: 0,
//			voc: 0,
//			vol: 0
		},
		scale: 1,
//		st: 0,
//		cst: 0,
		ulen: glovar.ulen,
		dur_fact: 1,
//		key: clone(parse.key),		// key at start of tune (parse / gene)
//		ckey: clone(parse.key),		// current key (parse / gene)
		meter: clone(glovar.meter),
		wmeasure: glovar.meter.wmeasure,
		staffnonote: 1,
		clef: new_sym({
			typ: C.clef,
			clef_auto: 1, //true
			clef_type: "a",		// auto
			time: 0
		}),
		acc: [],		// accidentals of the measure (parse)
		sls: [],		// slurs - used in parsing and in generation
		hy_st: 0
	}
	voice_tb.push(p_voice)
	voice_tb['v' + id] = p_voice	// double access (index and 'v'+id)

	if (parse.state == 3) {
//		p_voice.key = parse.key		// (done later in music.js)
		p_voice.ckey = clone(parse.key)
//		if (parse.key.k_bagpipe
//		 && !p_voice.pos.stm) {
//			p_voice.pos = clone(p_voice.pos)
//			p_voice.pos.stm &= ~0x07
//			p_voice.pos.stm |= C.SL_BELOW
//		}
		if (parse.key.k_bagpipe)
			p_voice.pos.stm == C.SL_BELOW
	}
	
	return p_voice
} // new_voice()

// treat V: with many voices
function do_cloning() {
    var	i,
	clo = curvoice.clone,
	vs = clo.vs,			// voices to clone
	a = clo.a,			// V: parameters
	parse1 = clo.parse,
	parse_sav = parse,
	bol = parse1.eol,		// start of the 1st voice sequence
	eol = parse.eol			// end

	delete curvoice.clone

	if (parse.mb[bol] == '\n')	// if stop on V:xx
		bol++

	// insert the music sequence in each voice
	include++
	for (i = 0; i < vs.length; i++) {
		parse = Object.create(parse1)	// create a new parse context
		get_voice(vs[i] + ' ' + a.join(' '))
		self.tosvg(parse.fname, parse.mb, bol, eol)
	}
	include--
	parse = parse_sav	// restore the parse context
} // do_cloning()

// treat a 'V:' info
function get_voice(parm) {
    var	v, vs, s,
	a = info_split(parm),
	vid = a.shift()

	if (!vid)
		return				// empty V:

	// if end of sequence with many voices, load the other voices
	if (curvoice && curvoice.clone)
		do_cloning()

	vs = vid.indexOf(',') > 0		// if many voices
		? vid.split(',')
		: vs = [vid]

	if (parse.state < 2) {			// memorize the voice parameters
		while (1) {
			vid = vs.shift()
			if (!vid)
				break
			if (a.length)
				memo_kv_parm(vid, a)
			if (vid != '*' && parse.state == 1)
				curvoice = new_voice(vid)
		}
		return
	}

	if (vid == '*') {
		syntax(1, "Cannot have V:* in tune body")
		return
	}

	curvoice = new_voice(vs[0])
	v = curvoice.v

	// if many voices, memorize the start of sequence
	if (vs.length > 1) {
		vs.shift()
		curvoice.clone = {
			vs: vs,
			a: a.slice(0),		// copy of the parameters
			parse: parse
		}
		parse = Object.create(parse)
//--fixme: useful?
//		if (parse.mb[curvoice.clone.bol - 1] != ']')
//			curvoice.clone.bol++	// start of new line
	}

	set_kv_parm(a)

	key_trans()

	if (curvoice.new) {			// if new voice
		delete curvoice.new
		if (staves_found < 0) {		// if no %%score/%%staves
			curvoice.st = curvoice.cst = ++nstaff
			par_sy.nstaff = nstaff
			par_sy.voices[v] = {
				st: nstaff,
				range: v
			}
			par_sy.staves[nstaff] = {
				stafflines: curvoice.stafflines || "|||||",
				staffscale: 1
			}
		} else if (!par_sy.voices[v]) {
			curvoice.ignore = 1	// voice not declared in %%staves
			return
		}
	}

	if (!curvoice.filtered
	 && par_sy.voices[v]
	 && (parse.voice_opts
	  || parse.tune_v_opts)) {
		curvoice.filtered = 1 //true
		voice_filter()
	}
} // get_voice()

// change state from 'tune header' to 'in tune body'
// curvoice is defined when called from get_voice()
function goto_tune() {
    var	v, p_voice

	set_page()
	if (glovar.new_nbar) {
		gene.nbar = glovar.new_nbar	// number of the first measure
		glovar.new_nbar = 0
	} else {
		gene.nbar = 1
	}

	parse.state = 3				// in tune body

	// update some voice parameters
	for (v = 0; v < voice_tb.length; v++) {
		p_voice = voice_tb[v]
		p_voice.ulen = glovar.ulen
		if (parse.key.k_bagpipe
		 && !p_voice.pos.stm) {
			p_voice.pos = clone(p_voice.pos)
			p_voice.pos.stm &= ~0x07
			p_voice.pos.stm |= C.SL_BELOW
		}
	}

	// initialize the voices when no %%staves/score	
	if (staves_found < 0) {
		v = voice_tb.length
		par_sy.nstaff =
			nstaff = v - 1
		while (--v >= 0) {
			p_voice = voice_tb[v]
			delete p_voice.new		// old voice
			p_voice.st = p_voice.cst = v
			par_sy.voices[v] = {
				st: v,
				range: v
			}
			par_sy.staves[v] = {
				stafflines: p_voice.stafflines || "|||||",
				staffscale: 1
			}
		}
	}
} // goto_tune()
