// alda.js - module to handle ALDA tunes
//	(https://alda.io/)
//
// Copyright (C) 2022-2024 Jean-Francois Moine
//
// This file is part of txtmus.
//
// txtmus is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// txtmus is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with txtmus.  If not, see <http://www.gnu.org/licenses/>.
//
// This module is loaded when "%%alda" appears in a text source.
//

if (typeof tm == "undefined")
	tm = {}

tm.alda = function(mus) {

    var	instr = {			// reduced names of the MIDI instruments
		"m-a-g-p": 0, "m-pia": 0, "pia": 0,	// acoustic-grand-piano
		"m-b-a-p": 1,				// bright-acoustic-piano
		"m-e-g-p": 2,				// electric-grand-piano
		"m-h-t-p": 3,				// honky-tonk-piano
		"m-e-p-1": 4,				// electric-piano-1
		"m-e-p-2": 5,				// electric-piano-2
		"m-harps": 6, "harps": 7,		// harpsichord
		"m-clav": 7, "clav": 7,			// clavi
		"m-cele": 8, "cele": 8,			// celesta
		"m-gl": 9, "gl": 9,			// glockenspiel
		"m-m-b": 10, "m-box": 10,		// music-box
		"m-vibra": 11, "vibra": 11,		// vibraphone
		"m-m": 12, "m": 12,			// marimba
		"m-x": 13, "x": 13,			// xylophone
		"m-tu-b": 14, "t-b": 14,		// tubular-bells
		"m-d": 15, "d": 15,			// dulcimer
		"m-d-o": 16,				// drawbar-organ
		"m-p-o": 17,				// percussive-organ
		"m-ro-o": 18,				// rock-organ
		"m-c-o": 19, "or": 19,			// church-organ
		"m-re-o": 20,				// reed-organ
		"m-ac": 21, "a": 21,			// accordion
		"m-harmo": 22, "harmo": 22,		// harmonica
		"m-t-a": 23,				// tango-accordion
		"m-a-g-n": 24, "m-a-g": 24, "a-g": 24,
					"gu": 24,	// acoustic-guitar
		"m-a-g-s": 25,				// acoustic-guitar-steel
		"m-e-g-j": 26,				// electric-guitar-jazz
		"m-e-g-c": 27, "e-c-g": 27,		// electric-guitar-clean
		"m-e-g-p-m": 28,			// electric-guitar-palm-muted
		"m-e-g-o": 29, "e-g-o": 29,		// electric-guitar-overdrive
		"m-e-g-d": 30, "e-g-d": 30,		// electric-guitar-distorted
		"m-e-g-h": 31, "e-g-h": 31,		// electric-guitar-harmonics
		"m-ac-b": 32, "ac-b": 32, "u-b": 32,	// acoustic-bass
		"m-e-b-f": 33, "e-b-f": 33, "e-b": 33,	// electric-bass
		"m-e-b-p": 34, "e-b-p": 34,		// electric-bass-pick
		"m-f-ba": 35, "f-b": 35,		// fretless-bass
		"m-b-sl": 36,				// bass-slap
		"m-b-po":  37,				// bass-pop
		"m-s-ba-1": 38,				// synth-bass-1
		"m-s-ba-2": 39,				// synth-bass-2
		"m-violi": 40, "violi": 40,		// violin
		"m-viola": 41, "viola": 41,		// viola
		"m-cell": 42, "cell": 42,		// cello
		"m-cont": 43, "s-b": 43, "ar-b": 43,
			"d-b": 43, "cont": 43, "m-s-b": 43,
			"m-ar-b": 43, "m-d-b": 43,	// contrabass
		"m-t-st": 44,				// tremolo-strings
		"m-p-st": 45,				// pizzicato-strings
		"m-o-ha": 46, "harp": 46, "o-h": 46,
					"m-harp": 46,	// harp
		"m-t": 47, "tim": 47,			// timpani
		"m-s-e-1": 48,				// string-ensemble-1
		"m-s-e-2": 49,				// string-ensemble-2
		"m-s-s-1": 50,				// synth-strings-1
		"m-s-s-2": 51,				// synth-strings-2
		"m-c-a": 52,				// choir-aahs
		"m-v-o": 53,				// voice-oohs
		"m-s-voi": 54,				// synth-voice
		"m-o-hi": 55,				// orchestra-hit
		"m-tru": 56, "tru": 56,			// trumpet
		"m-tro": 57, "tro": 57,			// trombone
		"m-tub": 58, "tub": 58,			// tuba
		"m-m-tr": 59,				// muted-trumpet
		"m-f-h": 60, "f-h": 60,			// french-horn
		"m-b-se": 61,				// brass-section
		"m-s-br-1": 62,				// synth-brass-1
		"m-s-br-2": 63,				// synth-brass-2
		"m-s-s": 64, "s-s": 64,			// soprano-saxophone
		"m-a-s": 65, "a-s": 65,			// alto-saxophone
		"m-t-sa": 66, "t-s": 66,		// tenor-saxophone
		"m-b-sa": 67, "b-s": 67,		// baritone-saxophone
		"m-ob": 68, "ob": 68,			// oboe
		"m-e-h": 69, "e-h": 69,			// english-horn
		"m-bas": 70, "bas": 70,			// bassoon
		"m-clar": 71, "clar": 71,		// clarinet
		"m-pic": 72, "pic": 72,			// piccolo
		"m-flu": 73, "f": 73,			// flute
		"m-re": 74, "r": 74,			// recorder
		"m-p-f": 75, "p-f": 75,			// pan-flute
		"m-bot": 76, "bot": 76,			// bottle
		"m-shak": 77, "shak": 77,		// shakuhachi
		"m-wh": 78, "w": 78,			// whistle
		"m-oc": 79, "oc": 79,			// ocarina
		"m-s-l": 80, "squa": 80, "sq-w": 80,
			"sq-l": 80, "m-squa": 80,
					 "m-sq-w": 80, 	//square-lead
		"m-sa-w": 81, "saw": 81, "sa-w": 81,
			"sa-l": 81, "m-saw": 81,
					 "m-sa-l": 81,	// saw-wave
		"m-ca-l": 82, "ca-l": 82, "m-call": 82,	// calliope-lead
		"m-ch-l": 83, "ch-l": 83, "chi": 83,
					"m-chif": 83,	// chiffer-lead
		"m-char": 84, "char": 84,		// charang
		"m-s-vox": 85,				// solo-vox
		"m-fif": 86, "m-sa-f": 86,		// fifths
		"m-b-a-l": 87, "m-bal": 87,		// bass-and-lead
		"m-s-p-n-a": 88, "m-p-n-a": 88,
					"m-n-a-p": 88,	// synth-pad-new-age
		"m-s-p-w": 89, "m-p-w": 89,
					"m-w-p": 89,	// synth-pad-warm
		"m-s-p-p": 90, "m-p-p": 90,		// synth-pad-polysynth
		"m-s-p-c": 91, "m-p-c": 91,
					"m-c-p": 91,	// synth-pad-choir
		"m-s-p-b": 92, "m-p-b": 92,
			"m-b-pa": 92, "m-p-b-g": 92,
					"m-b-g-p": 92,	// synth-pad-bowed
		"m-s-p-m": 93, "m-p-m": 93,
					"m-m-p": 93,	// synth-pad-metallic
		"m-s-p-h": 94, "m-p-h": 94,
					"m-h-p": 94,	// synth-pad-halo
		"m-s-p-s": 95, "m-p-sw": 95,
					"m-s-p": 95,	// synth-pad-sweep
		"m-f-r": 96, "m-f-i-r": 96, "m-ra": 96,
					"m-i-r": 96,	// fx-rain
		"m-f-s": 97, "m-soun": 97,		// fx-soundtrack
		"m-f-c": 98, "m-crys": 98,		// fx-crystal
		"m-f-a": 99, "m-at": 99,		// fx-atmosphere
		"m-f-br": 100, "m-bri": 100,		// fx-brightness
		"m-f-g": 101, "m-go": 101,		// fx-goblins
		"m-f-e": 102, "m-f-e-d": 102,
			"m-e": 102, "m-e-d": 102,	// fx-echoes
		"m-f-s-f": 103, "m-sc-f": 103,		// fx-sci-fi
		"m-sita": 104, "sita": 104,		// sitar
		"m-ban": 105, "ban": 105,		// banjo
		"m-sham": 106, "sham": 106,		// shamisen
		"m-ko": 107, "ko": 107,			// koto
		"m-ka": 108, "ka": 108,			// kalimba
		"m-bag": 109, "bag": 109,		// bagpipes
		"m-fid": 110,				// fiddle
		"m-sheh": 111, "sheh": 111, "shan": 111,
			"m-shah": 111, "m-shen": 111,
					"m-shan": 111,	// shehnai
		"m-ti-b": 112,				// tinkle-bell
		"m-ag": 113,				// agogo
		"m-st-d": 114, "s-d": 114,		// steel-drums
		"m-wo": 115,				// woodblock
		"m-t-d": 116,				// taiko-drum
		"m-m-to": 117,				// melodic-tom
		"m-sy-d": 118,				// synth-drum
		"m-r-c": 119,				// reverse-cymbal
		"m-g-f-n": 120,				// guitar-fret-noise
		"m-b-n": 121,				// breath-noise
		"m-seas": 122,				// seashore
		"m-b-t": 123,				// bird-tweet
		"m-t-r": 124,				// telephone-ring
		"m-helic": 125,				// helicopter
		"m-ap": 126,				// applause
		"m-gu": 127				// gunshot
	} // instr

// ambiguous cases
    var	amb = {
		ab: "21",
		b: "3",
		c: "4",
		cl: "21",
		h: "5",
		g: "2",
		k: "2",
		ma: "12",
		mab: "121",
		mb: "13",
		mbp: "112",
		mbs: "112",
		mc: "14",
		mcl: "121",
		mf: "13",
		mfb: "112",
		mg: "12",
		mh: "15",
		mk: "12",
		mmt: "112",
		mo: "12",
		moh: "112",
		mp: "13",
		mps: "112",
		mr: "12",
		mro: "121",
		ms: "14",
		msd: "121",
		msf: "121",
		msb1: "1121",
		msb2: "1121",
		msl: "121",
		msv: "113",
		msw: "121",
		mt: "13",
		mtb: "121",
		mts: "112",
		mv: "15",
		mw: "12",
		o: "2",
		p: "3",
		s: "4",
		sl: "21",
		sw: "21",
		t: "3",
		v: "5"
	}

    var	x_ = 0,				// current tune index
	otosvg				// original mus.tosvg()

    function parse_alda() {
    var	bol, eol, in_tune, curv,
	i, j, r, vid, in_chord, volta,
	C = tm.C,
	ds = C.BLEN / 2,		// internal duration of one second (M:1/4=120)
	cfmt = mus.cfmt(),
	parse = mus.parse,
	mb = parse.mb,			// music buffer
	v_tb = mus.get_v_tb(),		// voice table
	v = "1",
	markers = {},
	cram = [],
	vars = {},
	vari = {},			// 2 letters of the variables
	voices = ['1'],			// list of the voices of the next sequence
	names = {}

	// create a tuplet
	// curv.nst is the time of last standard duration
	function do_tp() {
	    var	bd, f, s2, m, r,
		d = 2000,		// big value
		tim = curv.time,
		s = curv.last_sym

		while (!s.dur)
			s = s.prev
		s2 = s				// end of the tuplet

		while (s.time != curv.nst) {	// search the beginning of the tuplet
			if (s.dur && s.dur < d)
				d = s.dur	// and the smallest duration
			s = s.prev
		}
		for (bd = 24; ; bd *= 2) {
			if (bd > d)		// base note type
				break
		}
		r = tm.rat(d, bd)
		s.tp = [{
			p: r[1],
			q: r[0],
			f: curv.tup || cfmt.tuplets
		}]
		f = bd / d			// factor giving the normal duration
		while (1) {
			if (s.dur) {
				d = s.dur * f + 4
				s.dur_orig = ((d / 12) | 0) * 12
				for (m = 0; m <= s.nhd; m++) {
					d = s.notes[m].dur * f + 4
					s.notes[m].dur = ((d / 12) | 0) * 12
				}
				s.in_tuplet = 1 //true
				if (s == s2)
					break
			}
			s = s.next
		}
		s2.tpe = 1
	} // do_tp()

	// get a duration
	// curv.ulen = current/default duration
	function get_dur() {
	    var	c, i, err, s, dot,
		u = curv.ulen,
		d = ""

		while (1) {
			c = mb[bol]
			if (c < '0' || c > '9')
				break
			d += c
			bol++
		}
		if (!d)
			return u
		if (c == '.') {
			c = mb[++bol]
			if (c == '.') {
				dot = ".."
				bol++
			} else if (c < '0' || c > '9') {
				dot = "."
			} else {
				d += '.'
				while (c >= '0' && c <= '9') {
					d += c
					c = mb[++bol]
				}
			}
		}
		if (c == 's') {
			u = ds * d
		} else if (c == 'm') {
			if (mb[bol + 1] != 's') {
				err = 1
			} else {
				bol++
				u = ds * d / 1000
			}
		} else {
			u = C.BLEN / d
			if (dot)
				u *= dot.length == 1 ? 1.5 : 2.25
		}
		if (err) {
//			parse.istart = bol
			parse.iend = bol
			mus.syntax(1, "Bad duration")
		}
		curv.ulen = u
		return u
	} // get_dur()

	// search the MIDI number of an instrument
	function get_instr(p) {
	    var	i, j, s, t, v

		t = p.split('-')		// compress
		s = ''
		for (i = 0; i < t.length; i++)
			s += t[i][0]

		v = instr[s]
		if (v != undefined)
			return v
		a = amb[s]
		if (!a)
			return 0		// unknown instrument
		s = ''
		for (i = 0; i < t.length; i++) {
			for (j = 0; j < +a[i]; j++) {
				if (t[i][j])
					s += t[i][j]
			}
		}
		return instr[s] || 0
	} // get_instr()

	// get the part (instrument/voice) at start of line
	// on entry, eol points to the string after the ':'
	// the part can be:
	//	instrument [ "/" instrument ]* [ '"' name '"' ]
	//	name "/" name [ "/" name ]* [ '"' name '"' ]
	//			(the last name defines a named part or a group of parts)
	//	name "." instrument
	//	"V" voice_number ("0", "1", "2", "3" - "V0" does a voice synchronization)
	function get_part(p) {
	    var	a, t, b, u, p_v, p_v2, v1

		// add a voice to the list of the voices
		// and resynchronize the voices with many instrument parts
		function vadd(p) {
		    var	i

			voices.push(p)
			p_v = v_tb['v' + p]
			if (p_v && p_v.mit) {		// if many parts
				t = p_v.time
				for (i = 1; i < 5; i++) {
					p_v2 = v_tb['v' + p + '.V' + i]
					if (!p_v2)
						break
					if (p_v2.time > t)
						t = p_v2.time
				}
//				while (--i > 1)		// resynchronize
//					v_tb['v' + p + '.' + i].time = t
				p_v.time =
					p_v.mit = t
			}
		} // vadd()

		v1 = voices[0]			// keep 1st voice if Vx
		voices.length = 0		// reset the voice list

		// try name "." instrument
		t = p.split('.')
		if (t[1]) {		// yes
			a = names[t[0]]
			if (a && a.indexOf(t[1]) < 0)
				a = ''
			if (!a) {
				parse.istart = bol
				mus.syntax(1, "Unknown part " + p)
			} else {
				voices.push(p)
			}
			return
		}

		t = p.split(/\s+/)
		p = t[0]			// instrument/name list
		a = t[1]			// name

		// handle the instrument voices
		if (/^V\d$/.test(p)) {			// if instrument voice
//			b = curv.id.replace(/\.V\d$/, '')
			b = v1.replace(/\.V\d$/, '')
			p_v = v_tb['v' + b]
			if (!p_v) {			// no main instrument?
				mus.get_voice(b)
				p_v = mus.get_curvoice()
			}
			if (p != 'V0') {
				if (p_v.mit == undefined)
					p_v.mit = p_v.time	// main instrument time
				b = b + '.' + p		// voice ID
				voices.push(b)
				if (!v_tb['v' + b]) {	// if new voice
					mus.get_voice(b)
					p_v2 = mus.get_curvoice()
					p_v2.key = p_v2.ckey = p_v.ckey
					p_v2.ulen = p_v.ulen
					p_v2.oct = p_v.oct
				}
				return
			}
			p = b			// 'V0' goes to the main instrument
		}

		// check if there is a definition of a named part or a group of parts
		if (a) {			// if a named part
			if (a[0] != '"' || a[a.length - 1] != '"') {
				parse.istart = bol
				mus.syntax(1, "Bad named part definition " + a)
				return
			}
			a = a.slice(1, -1)
			names[a] = p
			p = a
		}

		// get the instruments/parts and resolve the names
		t = p.split('/')
		while (t.length) {
			p = t.shift()
			b = names[p]
			if (b) {
				u = b.split('/')
				if (u.length > 1) {
					while (u.length)
						t.push(u.shift)
					continue
				}
				vadd(p)
			} else {
				vadd(p)
			}
		}
	} // get_part()

	// definition of a variable
	function variable() {
	    var	k, v,
		i = bol,
		j = mb.indexOf('=', i)

//		if (j < 0 || j > eol)
//			return
		k = mb.slice(bol, j).trim()	// key = variable name

		// get the value
		i = j + 1
		while (mb[i] == ' ')
			i++
		if (mb[i] != '[') {
			v = mb.slice(i, eol)
		} else {
			j = mb.indexOf(']', i)
			if (j < 0) {
				parse.istart = i
				mus.syntax(1, "No end of variable")
				return
			}
			v = mb.slice(i, j).replace('\n', ' ').trim()
//			if (j > eol)
////				parse.eol = 
//				eol = mb.indexOf('\n', j)
		}
		vars[k] = v
		vari[k[0] + k[1]] = 1 //true	// 2 first letters in variable
	} // variable()

	// generate the music contained in the variable 'v'
	function do_seq(v) {
	    var	sav = {
			mb: mb,
			bol: bol,
			eol: eol,
			eof: parse.eof,
		}

		// parse the music
		mb = v
		bol = 0
		parse.eol = -1
		parse_seq()

		mb = sav.mb		// restore the scanning parameters
		bol = parse.bol = sav.bol
		eol = parse.eol = sav.eol
		parse.eof = sav.eof
	} // do_seq()

	// check if start of a variable
	function var_p() {
	    var	v,
		n = 1,
		i = bol

		while (i < eol) {
			if (!/\w/.test(mb[i]))
				break
			i++
		}
		v = vars[mb.slice(bol - 1, i)]
		if (!v)
			return
		while (mb[i] == ' ')
			i++
		if (mb[i] == '*') {		// repeat
			i++
			while (mb[i] == ' ')
				i++
			n = +mb[i++]
		}
		bol = i
		while (--n >= 0)
			do_seq(v)
		return 1
	} // var_p()

	// extract the name of a mark
	function marknam() {
	    var	i = bol,
		j = i + 1

		while (j < eol && mb[j] != ' ' && mb[j] != '\t')
			j++
		bol = j
		return mb.slice(i, j)
	} // marknam()

	function bar(ty) {
		return {
			typ: C.bar,
			bar_type: ty,
			dur: 0
		}
	} // bar()

	// insert a measure bar
	function add_bar() {
	    var	tim = 0,
		s = curv.last_sym,
		s2 = curv.last_bar || curv.sym

		if (s2) {
			while (s2 != s) {
				if (s2.bar_type)
					tim = s2.time
				s2 = s2.next
			}
		}
		do {
			tim += curv.wmeasure	// time of the next bar
		} while (tim < curv.time)
		if (tim == curv.time && s2 && !s2.bar_type) {
			s2 = bar("|")		// insert a bar
			mus.sym_link(s2)
			curv.last_bar = s2
		}
	} // add_bar()

	// return an empty note symbol
	function new_sym_note() {
		if (curv.wmeasure != 1		// if M: != none
		 && curv.time >= curv.wmeasure)	// insert a measure bar
			add_bar()
		return mus.new_sym({		// note template
			typ: tm.C.note,
			dur: 0,
			stem: curv.pos.stm == C.SL_ABOVE ? 1
				: curv.pos.stm == C.SL_BELOW ? -1 : 0,
			notes: []
		})
	} // new_sym_note()

	// -- music elements --
	// sequence '[' ..music.. ']' [ '*' nb | '\'' repeat ]
	function seq() {
	    var	seq, s,
		i = bol,
		j = i,
		n = 1

		while (1) {
			if (mb[j] == '[') {
				n++
			} else if (mb[j] == ']') {
				if (--n == 0) {
					break
				}
			}
			j++
		}
		seq = mb.slice(i, j)
		j++
		while (mb[j] == ' ' || mb[j] == '\t')
			j++
		switch (mb[j]) {
		case "'":
			j++
			i = j
			while (mb[j] > ' ')
				j++
			s = mus.get_curvoice().last_sym
			if (s & s.bar_type) {
				s.rbstart = 2
				s.rbstop = 2
				s.text = mb.slice(i, j)
			} else {
				s = bar("[")
				s.rbstart = 2
				s.rbstop = 2
				s.invis = 1 //true
				s.text = mb.slice(i, j)
				parse.istart = bol
				parse.iend = j
				mus.sym_link(s)
			}
			bol = j + 1
			do_seq(seq)
			volta = 1	// stop the volta on the next measure bar
			return
		default:
			bol = j
			do_seq(seq)
			return
		case '*':
			break
		}
		j++
		while (mb[j] == ' ' || mb[j] == '\t')
			j++
		n = +mb[j]
		bol = j + 1
		if (isNaN(n)) {
			parse.istart = j
			mus.syntax(1, "Bad sequence repeat number")
			return
		}
		s = mus.get_curvoice().last_sym
		if (s && s.bar_type) {
			s.bar_type = s.bar_type[0] == ':' ? "::" : "|:"
			s.rbstop = 2
		} else {
			s = bar("|:")
			s.rbstop = 2
			parse.istart = bol
			parse.iend = j
			mus.sym_link(s)
		}
		do_seq(seq)
		s = mus.get_curvoice().last_sym
		if (s && s.bar_type) {
			s.bar_type = ":|"
			s.rbstop = 2
		} else {
			s = bar(":|")
			s.rbstop = 2
			parse.istart = bol
			parse.iend = j
			mus.sym_link(s)
		}
		while (--n > 2)
			s.bar_type += ':'
	} // seq()

	function note() {
	    var	c, s, tim,
		pit = "cdefgab".indexOf(mb[bol - 1]) + curv.oct * 7 - 30,
		nt = {
			pit: pit,
			shhd: 0,
			shac: 0
		},
		a = 0

		parse.istart = bol - 1

		c = mb[bol]
		while (c == '+' || c == '-') {
			if (c == '+')
				a++
			else
				a--
			c = mb[++bol]
		}
		if (c == '_') {
			a = 3			// natural
//			c = mb[++bol]
			bol++
		}
		if (a)
			nt.acc = a
		mus.set_midi(nt)

		nt.dur = get_dur()
		if (in_chord) {
			in_chord = 0 //false
			s = curv.last_sym
			if (nt.dur != s.dur	// if different chord duration
			 || s.ti2) {		// or after a sequence of tied notes
				tim = curv.time
				if (s.ti2) {
					while (s.prev && (s.ti2 || !s.dur))
						s = s.prev
					while (!s.dur)
						s = s.next
				} else {
					if (nt.dur > s.dur) {	// longest note => 1st symbol
						tim += nt.dur - s.dur
						c = s.notes[s.nhd]
						s.notes[s.nhd] = nt
						s.dur = s.dur_orig = nt.dur
						nt = c
//fixme: may not work in all cases...
						curv.nst = -1
					}
				}
				curv.time = s.time
				mus.get_vover('(')	// start an overlay sequence
				curv.time = tim
				mus.get_vover('&')	// new overlay voice
				curv = mus.get_curvoice()
				curv.ovt = tim		// time of end of overlay
				curv.oct = s.p_v.oct
				curv.nst = -1
				if (nt.midi > s.notes[0].midi)
					curv.pos.stm = (curv.pos.stm & 0x07)
						| C.SL_ABOVE
				s = new_sym_note()
				s.notes = []
				s.dur = s.dur_orig = nt.dur
				parse.iend = bol
				mus.sym_link(s)
				curv.last_note = s
//			} else {		// else, just add the note in the chord
			}
		} else {
			s = new_sym_note()
			s.notes = []
			s.dur = s.dur_orig = nt.dur
			parse.iend = bol
			mus.sym_link(s)
			curv.last_note = s
			mus.deco_cnv(s)
		}
		s.nhd = s.notes.length
		s.notes.push(nt)
	} // note()

	function rest() {
		if (curv.wmeasure != 1
		 && curv.time >= curv.wmeasure)	// insert a measure bar
			add_bar()
	    var	s2,
		s = {
			typ: C.rest,
			istart: bol - 1,
			stem: 0,
			nhd: 0,
			xmx: 0,
			notes: [ {pit: 0} ]
		}

		parse.istart = bol - 1
		s.dur = s.dur_orig = s.notes[0].dur = get_dur()
		if (s.dur == curv.wmeasure)
			s.fmr = 1 //true		// full measure rest
		parse.iend = bol
		if (in_chord) {
			s2 = curv.last_sym
			while (!s2.dur)
				s2 = s2.prev
			if (s2.dur > s.dur) {
				s2.dur = s.dur
				curv.time = s2.time + s2.dur
			}
			in_chord = 0
		} else {
			mus.sym_link(s)
		}
	} // rest()

	function tie() {
	    var	c, s2, not1, not2, m, d,
		s = curv.last_sym

		parse.istart = bol - 1
		parse.iend = bol
		while (s && !s.dur)		// search the start of the tie
			s = s.prev
		if (!s) {
			mus.syntax(1, "Bad tie")
			return
		}
		if (mb[bol] == '|') {		// there may be a bar as "c~|2"
			elts['|']()
			bol++
		}
		c = mb[bol]
		if (c < '0' || c > '9')		// if no duration, this is a slur
			return

		// if a tie from a chord, move the start note to an overlay voice
		if (s.nhd) {
			tim = curv.time
			curv.time = s.time		// time of the chord
			mus.get_vover('(')		// open an overlay voice
			curv.time = tim
			mus.get_vover('&')		// new overlay voice
			curv = mus.get_curvoice()
			curv.ovt = tim			// end of overlay
			curv.oct = s.p_v.oct
			curv.nst = -1
//			curv.mit = s.p_v.mit
			if (s.notes[1].midi > s.notes[0].midi)
				curv.pos.stm = (curv.pos.stm & 0x07)
					| C.SL_ABOVE
			s2 = new_sym_note()		// symbol for the start note
			s2.notes = [s.notes[s.nhd--]]
			s2.dur = s2.dur_orig = s2.notes[0].dur
			mus.sym_link(s2)		// put the note in the ov. voice
			curv.last_note = s2
//			d = s.notes[m].dur		// recompute the chord duration
//			for (m = 0; m < s.nhd; m++) {
//				if (d > s.notes[m].dur)
//					d = s.notes[m].dur
//			}
//			s.dur = s.dur_orig = d
			s = s2				// new start note
		}

		if (curv.wmeasure != 1
		 && curv.time >= curv.wmeasure)		// insert a measure bar
			add_bar()
		s2 = mus.clone(s)			// tied symbol
		s2.next = null
		s2.notes = mus.clone(s.notes)
		s2.nhd = 0

		not1 = s.notes[0]			// start of the tie
		not2 = mus.clone(not1)			// end of the tie
		s2.notes[0] = not2
		not2.s = s2
		s2.dur = s2.dur_orig = not2.dur = get_dur()
		mus.sym_link(s2)		// new note
		not1.tie_ty = C.SL_AUTO
		not1.tie_e = not2
		not1.s = s
		s.ti1 = 1 //true			// start of tie
//		not2.tie_ty = C.SL_AUTO
		not2.tie_s = not1
		not2.acc = 0
		s2.ti2 = 1				// end of tie
	} // tie()

	function attr() {
	    var	a, glo, v,
		n = 1,
		i = bol

		while (n) {
			if (mb[++i] == '(')
				n++
			else if (mb[i] == ')')
				n--
		}
		if (i > eol) {
			parse.istart = bol
			mus.syntax(1, "Bad attribute")
			return
		}
		a = mb.slice(bol, i).match(/(\S+)\s*(.*)/)
		if (!a) {
			parse.istart = bol
			mus.syntax(1, "Bad attribute")
			return
		}
		if (a[1].slice(-1) == '!') {
			glo = 1
			a[1] = a[1].slice(0, -1)
		}
		switch (a[1]) {
		case "f":
		case "ff":
		case "fff":
		case "ffff":
		case "mp":
		case "mf":
		case "p":
		case "pp":
		case "ppp":
		case "pppp":
			mus.dcn_add(a[1])
			break
		case "key-sig":
		case "key-signature":
			switch (a[2][0]) {
			case '"':		// "f+ c+ g+"
				v = a[2].slice(1, -1)
					.replace(/\s/g,'')
				if (!v.length)
					v = "C"
				else if ("f+c+g+d+a+e+b+".indexOf(v) == 0)
					v = "GDAEBFC"[v.length / 2 - 1]
				else if ("b-e-a-d-g-c-f-".indexOf(v) == 0)
					v = "FBEADGC"[v.length / 2 - 1]
				else
//					v = 'none exp ' + v
					v = 'C ' + v
							.replace(/(.)(\+)/g, "^$1")
							.replace(/(.)(\-)/g, "_$1")
							.toUpperCase()
				break
			case "'":		// '(f (sharp) c (sharp) g (sharp))
						// '(a major) with modes ionian dorian ...
				if (a[2][1] != '('
				 || a[2].slice(-1) != ')')
					break
				v = a[2].slice(2, -1)
				if (v.indexOf('(') > 0) {
					v = v.match(/[a-g]|\([^)]+\)/g)
					if (!v || (v.length & 1))
						break
					v = 'none exp '
						+ v.join('')
							.replace(/\(sharp\)/g, '+')
							.replace(/\(flat\)/g, '-')
							.replace(/(.)(\+)/g, "^$1")
							.replace(/(.)(\-)/g, "_$1")
							.toUpperCase()
					break
				}
				v = v.split(/\s+/)
				v[0] = v[0].toUpperCase()
				switch (v[1]) {
				case 'sharp':
					v[1] = '#'
					break
				case 'flat':
					v[1] = 'b'
					break
				}
				v = v.join('')
				break
//			default:
//				if (!a[2])
//					v = "C"
//				break
			}
			parse.istart = bol
			if (!v) {
				mus.syntax(1, "Bad key signature")
				break
			}
			mus.get_key(v)
			if (glo)
				parse.key = curv.ckey
			break
		case "metric-modulation":
//			break
		case "octave":
//			break
		case "pan":
		case "panning":
//			break
		case "quant":
		case "quantize":
		case "quantization":
//			break
		case "set-duration":
//			break
		case "set-note-length":
//			break
		case "set-duration-ms":
			break
		case "tempo":
// [ note_length ] BPM
// assume a beat on a quarter note by default
// note_length = normal note length, but as "4." if dotted
			mus.new_tempo('1/4=' + a[2])
			ds = C.BLEN / 4 * 60 / a[2]
			break
		case "track-vol":
		case "track-volume":
//			break
		case "transpose":
		case "transposition":
//			break
		case "vol":
		case "volume":
			break
		default:
			parse.istart = bol
			mus.syntax(1, "Unknown attribute " + a)
			break
		}
		bol = i + 1
	} // attr()

	function do_cram() {
	    var	m, fo,
		d = get_dur(),			// cram duration
		s = curv.last_sym,
		tim = cram.pop(),
		ntim = s.time + s.dur,
		f = (ntim - tim) / d,
		r = tm.rat(d, ntim - tim)

//console.log('dt:'+(ntim-tim)+' d:'+d+' f:'+f+' r:'+r)
		if (tim == undefined) {
			parse.istart = bol
			mus.syntax(1, "No start of cram")
			return
		}
		if (!s.tpe)
			s.tpe = 1
		else
			s.tpe++
		curv.time = ntim = tim + d
		if (f > 2)
			fo = 2
		while (s.time >= tim) {
			s.in_tuplet = 1 //true
			if (s.dur) {
				s.dur /= f
				if (fo)
					s.dur_orig /= fo
				if (s.time == tim)
					break
				ntim -= s.dur
			}
			s.time = ntim
			if (!s.prev)
				break
			s = s.prev
		}
		if (!s.tp)
			s.tp = []
		s.tp.push({
			p: r[1],
			q: r[0],
			f: curv.tup || cfmt.tuplets
		})
	} // do_cram()

	function empty() {}

	// language elements
    var	elts = {
		'\t': empty,
		'\n': empty,
		' ': empty,
		'#': function() {
			bol = mb.indexOf('\n', bol)
			if (bol < 0)
				bol = mb.length
		},
		'%': function() {		// mark time (or ABC pseudo-comment)
			if (mb[bol - 2] == '\n' && mb[bol] == '%') {
				mus.do_pscom(mb.slice(bol + 1, eol))
				bol = cfmt.alda
					? eol
					: mb.length	// stop by %%alda 0
//fixme: stop does not work
				return
			}

		    var mn = marknam()

			if (mn)
				markers[mn] = curv.time
		},
		'(': attr,
		'*': function() {		// repeat last symbol
		    var	n, s

			parse.istart = bol - 1
			while (mb[bol] == ' ')
				bol++
			parse.iend = bol
			n = +mb[bol++]
			if (isNaN(n)) {
				mus.syntax(1, "Bad repeat number")
				return
			}
			n--
			s = curv.last_sym
			while (--n >= 0) {
				s = mus.clone(s)
				s.next = null
				mus.sym_link(s)
			}
		},
		'/': function() {		// chord
			in_chord = 1 //true
		},
		'<': function() {
			curv.oct--
		},
		'>': function() {
			curv.oct++
		},
		'@': function() {		// set time
		    var	mn = marknam()

			if (mn) {
				if (markers[mn] == undefined) {
					parse.istart = bol - 1
					mus.syntax(1, "No such marker " + mn)
				} else {
					curv.time =
						curv.mit = markers[mn]
				}
			}
		},
		'[': seq,
		a: note,
		b: note,
		c: note,
		d: note,
		e: note,
		f: note,
		g: note,
		o: function() {		// octave
		    var	oct = +mb[bol++]

			if (isNaN(oct)) {
				parse.istart = bol - 1
				mus.syntax(1, "Bad octave")
			} else {
				mus.get_curvoice().oct = oct
			}
		},
		r: rest,
		'{': function() {
			cram.push(mus.get_curvoice().time)
		},
		'|': function() {
		    var	s = bar("|")

			if (volta) {
				volta = 0
				s.rbstop = 2
			}
			parse.istart = bol - 1
			parse.iend = bol
			mus.sym_link(s)
			curv.last_bar = s
		},
		'}': do_cram,
		'~': tie
	} // elts

	// parse the music sequence
	// on entry, the sequence is contained in 'mb' between bol and eol
	// the concerned voices are in 'voices'
	function parse_seq() {
	    var	i, c, v, iv, s, p_v,
		sav = {
			bol: bol,
			eol: eol
		}

		// loop on the voices
		for (iv = 0; iv < voices.length; iv++) {
			v = voices[iv]			// voice ID
			mus.get_voice(v)
			curv = mus.get_curvoice()
			if (curv.nst == undefined) {	// if a new voice
				i = names[v]		// check if named part
				if (!i)
					i = v
				curv.instr = get_instr(i)
				if (curv.id != "1") {
					curv.nm = v
					curv.snm = v.slice(0, 4)
					if (/\.V\d$/.test(v))
						curv.snm += '.' + v.slice(-1)
					curv.new_name = 1
				}
				curv.nst = -1
//			} else {
//				nst = curv.nst
			}
			if (/\.V\d$/.test(v)) {		// if instrument part
				p_v = v_tb['v' + v.slice(0, -3)] // main instrument
				if (curv.time < p_v.mit)
					curv.time = p_v.mit
			}
//fixme: octave per voice or global?
			if (!curv.oct)
				curv.oct = 4
//fixme: u global? no: per voice
//			u = C.BLEN / 4

			// loop on music sequence
			bol = sav.bol
			eol = sav.eol
			while (1) {
				if (bol >= eol)
					break
				c = mb[bol++]
				if (!c)
					break
				if (vari[c + mb[bol]]	// if start of a variable
				 && var_p())
					continue
				if (typeof elts[c] != "function") {
					parse.istart = bol - 1
					parse.iend = bol
					mus.syntax(1, "Unknown element " + c)
					break
				}
				if (elts[c]())
					break		// error

				// handle the non standard durations
				if (curv.time / 3 != (curv.time / 3 | 0)) {
					if (curv.nst < 0)	// start a tuplet
						curv.nst = curv.last_sym
								? curv.last_sym.time
								: 0
				} else if (curv.nst >= 0) {	// end of tuplet?
					do_tp()			// yes, build it
					curv.nst = -1
				}

				// handle the end of voice overlay
				if (curv.time >= curv.ovt) {
//					curv.ovt = 10000000
					i = curv.ulen
					c = curv.oct
					mus.get_vover(')')
					curv = mus.get_curvoice()
					curv.ulen = i
					curv.oct = c
				}
			}
		}
		bol = sav.eol + 1
	} // parse_seq()

	// -- code of parse_alda --

	in_tune = parse.state
	switch (in_tune) {
	case 0:
		mus.start_tune()
		mus.info().X = (++x_).toString()	// new tune
		// fall thru
	case 1:
		mus.get_key('C')		// start the tune body
		break
	}

	bol = parse.eol + 1
loop1:
	while (1) {
		while (mb[bol] == '\n')
			bol++
		eol = mb.indexOf('\n', bol)
		if (eol < 0) {
			eol = mb.length
			break
		}
		while (1) {
			switch (mb[bol]) {
			case ' ':
			case '\t':
			case '\n':
				bol++
				continue
			case '#':
				bol = eol + 1
				continue loop1
			case '%':
				if (mb[bol + 1] == '%') {	// if ABC pseudo-comment
					mus.do_pscom(mb.slice(bol + 2, eol))
					if (!cfmt.alda)
						break loop1
					bol = eol + 1
					continue loop1
				}
			}
			break
		}
		i = mb.indexOf(':', bol)
		if (i > 0 && i < eol) {
			get_part(mb.slice(bol, i).trim())
			bol = i + 1
		} else {
			i = mb.indexOf('=', bol)
			if (i > 0 && i < eol) {
				variable()
				bol = eol + 1
				continue
			}
		}
		if (bol < eol)
			parse_seq()
	}

	// generate the tune if pure alda
	if (!in_tune)
		mus.end_tune()

	parse.eol = eol				// continue ABC scan
    } // parse_alda()

    // parse a alda source file
    function tosvg(of, fn, mb, bol, eof) {
    var	parse = mus.parse

	parse.mb = mb
	parse.fname = fn
	if (bol == undefined)
		bol = 0
	parse.eol = bol - 1
	parse_alda()
	set_fmt(null, "alda", "0")
    } // tosvg()

    function set_fmt(of, cmd, parm) {
	if (cmd != "alda") {
		of(cmd, parm)
		return
	}

    var	cfmt = mus.cfmt()

	cfmt.alda = mus.get_bool(parm)
	if (!cfmt.alda) {
		if (otosvg) {		// restore the tosvg function
			mus.tosvg = otosvg
			otosvg = null
		}
		return
	}
	if (!otosvg) {
		otosvg = mus.tosvg
		mus.tosvg = tosvg.bind(mus, mus.tosvg)
	}
	parse_alda()
    } // set_fmt()

	mus.set_format = set_fmt.bind(mus, mus.set_format)
} // alda

// the module is loaded
if (!tm.mhooks)
	tm.mhooks = {}
tm.mhooks.alda = tm.alda
