// clip.js - module to handle the %%clip command
//
// Copyright (C) 2018-2024 Jean-Francois Moine
//
// This file is part of txtmus.
//
// txtmus is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// txtmus is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with txtmus.  If not, see <http://www.gnu.org/licenses/>.
//
// This module is loaded when "%%clip" appears in a ABC source.
//
// Parameters
//	%%clip start_measure_nb [":" num "/" den] "-" end_measure_nb [":" num "/" den]

if (typeof tm == "undefined")
	tm = {}

tm.clip = function(mus) {

    function get_clip(parm) {
    var	C = tm.C

	// get the start/stop points
	function get_symsel(a) {
	    var	j, d, sq,
		b = a.match(/(\d+)([a-z]?)(:\d+\/\d+)?/)

		if (!b)
			return
		if (b[2])
			sq = b[2].charCodeAt(0) - 0x61
		if (!b[3])
			return {m: b[1], t: 0, sq: sq}	// on measure bar
		a = b[3].match(/:(\d+)\/(\d+)/)
		if (!a || a[2] < 1)
			return
		return {m: b[1], t: a[1] * C.BLEN / a[2], sq: sq}
	} // get_symsel()

	    var	b, c,
		a = parm.split(/[ -]/)

		if (a.length != 2) {
			mus.syntax(1, mus.errs.bad_val, "%%clip")
			return
		}
		if (!a[0])
			b = {m: 0, t: 0}
		else
			b = get_symsel(a[0])
		c = get_symsel(a[1])
		if (!b || !c) {
			mus.syntax(1, mus.errs.bad_val, "%%clip")
			return
		}
		mus.cfmt().clip = [b, c]
    } // get_clip()

    // cut the tune
    function do_clip() {
    var	C = tm.C
	v_tb = mus.get_v_tb(),
	cfmt = mus.cfmt()

	// go to a global (measure + time)
	function go_global_time(s, sel) {
	    var	s2, bar_time, seq

		if (sel.m <= 1) {	// special case: there is no measure 0/1
			if (sel.m == 1) {
				for (s2 = s; s2; s2 = s2.ts_next) {
					if (s2.typ == C.bar
					 && s2.time != 0)
						break
				}
				if (s2.time < v_tb[mus.get_cur_sy().top_voice].
								meter.wmeasure)
					s = s2
			}
		} else {
			for ( ; s; s = s.ts_next) {
				if (s.typ == C.bar
				 && s.bar_n >= sel.m)
					break
			}
			if (!s)
				return // null

			if (sel.sq) {
				seq = sel.sq
				for (s = s.ts_next; s; s = s.ts_next) {
					if (s.typ == C.bar
					 && s.bar_n == sel.m) {
						if (--seq == 0)
							break
					}
				}
				if (!s)
					return // null
			}
		}

		if (sel.t == 0)
			return s
		bar_time = s.time + sel.t
		while (s.time < bar_time) {
			s = s.ts_next
			if (!s)
				return s
		}
		do {
			s = s.ts_prev	// go back to the previous sequence
		} while (!s.seqst)
		return s
	} // go_global_time()

	    var	s, s2, sy, p_voice, v

		// remove the beginning of the tune
		s = mus.get_tsfirst()
		if (cfmt.clip[0].m > 0
		 || cfmt.clip[0].t > 0) {
			s = go_global_time(s, cfmt.clip[0])
			if (!s) {
				mus.set_tsfirst(null)
				return
			}

			// update the start of voices
			sy = mus.get_cur_sy()
			for (s2 = mus.get_tsfirst(); s2 != s; s2 = s2.ts_next) {
				switch (s2.typ) {
				case C.clef:
					s2.p_v.clef = s2
					break
				case C.key:
					s2.p_v.key = mus.clone(s2.as.u.key)
					break
				case C.meter:
					s2.p_v.meter = mus.clone(s2.as.u.meter)
					break
				case C.staves:
					sy = s2.sy
					mus.set_cur_sy(sy)
					break
				}
			}
			for (v = 0; v < v_tb.length; v++) {
				p_voice = v_tb[v]
				for (s2 = s; s2; s2 = s2.ts_next) {
					if (s2.v == v) {
						delete s2.prev
						break
					}
				}
				p_voice.sym = s2
			}
			mus.set_tsfirst(s)
			delete s.ts_prev
		}

		/* remove the end of the tune */
		s = go_global_time(s, cfmt.clip[1])
		if (!s)
			return

		// keep the current sequence
		do {
			s = s.ts_next
			if (!s)
				return
		} while (!s.seqst)

		/* cut the voices */
		for (v = 0; v < v_tb.length; v++) {
			p_voice = v_tb[v]
			for (s2 = s.ts_prev; s2; s2 = s2.ts_prev) {
				if (s2.v == v) {
					delete s2.next
					break
				}
			}
			if (!s2)
				p_voice.sym = null
		}
		delete s.ts_prev.ts_next
	} // do_clip()

	function set_fmt(of, cmd, parm) {
		if (cmd == "clip")
			get_clip(parm)
		else
			of(cmd, parm)
	} // set_fmt()

    function set_bar_num(of) {
	of()
	if (mus.cfmt().clip)
		do_clip()
    } // set_bar_num()

	// set the hooks
	mus.set_format = set_fmt.bind(mus, mus.set_format)
	mus.set_bar_num = set_bar_num.bind(mus, mus.set_bar_num)
} // clip

if (!tm.mhooks)
	tm.mhooks = {}
tm.mhooks.clip = tm.clip
