// txtmus - cmni.js - convert carnatic music notation to SVG
//
// Copyright (C) 2022-2024 Jean-Francois Moine
//
// This file is part of txtmus.
//
// txtmus is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// txtmus is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with txtmus.  If not, see <http://www.gnu.org/licenses/>.

// This file is loaded when the music file contains
//	%%cmni [options]
// options:
//	case=		0: both letter cases are normal notes
//			1: uppercase letters are notes in the upper octave
//			2: lowercase letters are grace notes
//	nlines=		0: the lyrics are enclosed by [ .. ]
//			1: first line = music, next line = lyrics
//	raga=		0: scale by accidental
//			1: scale by key signature
//	syntax=		0: other
//			1: Gaayaka (guessed from start of file - 999999)
//			2: raagaa (guessed from '{' = start of comment and no '}'
//				and '#' = directive)

//doc Gaayaka files
//	http://carnatic.1apps.com/rasikasw/gkformat.htm
// file extension .gka
// first line = numbers
//	1: version = 999999
//	2: (dummy) = 0
//	3: scale = 7 digits
//		0 (S)
//		1,2,3 (R)
//		0,1,2 (G)
//		1,2 (M)
//		0 (P)
//		1,2,3 (D)
//		0,1,2 (N)
//	4: duration, unit 1/100th of second
//	5: akscaram = number of notes the thaalam' one akshalam
//	6: number of akshalams in the Laghu
//	7: thaalam number (0..9)
//	8: quoted string = 1: beat(clap), 2: wave or count, 3: silence
//			other digits: number of aksharams in one thaalam cycle
// music:	sA, rI, gA, mA, pA, dA, nI,	// lower octave
//		sa, ri, ga, ma, pa, da, ni,
//		Sa, Ri, Ga, Ma, Pa, Da, Ni	// upper octave
//	double voyelle => double duration
//	dot over / below => octave up / down
//	( and ) half and double duration
//	, double duration or silence if not after note or after '-'
//	; triple duration or silence if not after note or after '-'
//	. half duration or silence if not after note or after '-'
//	< and > increment / decrement the pitch by (81/80) / (80/81)
//	{M n} = mode change
//	{T n} = tempo change (n in 1/100th second)
//	... other symbols...

// doc raagaa
// https://carnaticmusicguru.sourceforge.net/raagaa_help.htm
// file extension .cmn ?
// #D #I #M #N = tempo, instrument, mode, temperament
//	^ or ' upper octave
//	. lower octave
//	- emphasis or start of silence if followed by , or ;
//	, double duration or silence if not after note
//	; triple duration or silence if not after note
//	: half duration or silence if not after note
//	| or || end of thaala
//	/ or \ slide
//	>n and <n increase/decrease the note frequency

if (typeof tm == "undefined")
	tm = {}

tm.cmni = function(mus) {

	// temperaments
    var	sc_cmn12 =
	//	 S   R1    R2   G2   G3   M1    M2     P   D1    D2    N2    N3   S^
	//	"1, 16/15, 9/8, 6/5, 5/4, 4/3, 45/32, 3/2, 8/5, 27/16, 9/5, 15/8 ,2",
		"0   12     4   16  -14   -2   -10     2   14     6    18   -12",
//	//	    S   R1     R2     R3    G1    G2   G3   M1    M2
//	sc_cmn16 = "1, 32/31, 16/15, 10/9, 32/27, 6/5, 5/4, 4/3, 17/12,
//			 P     D1    D2   D3    N1   N2    N3   S^
//			3/2, 128/81, 8/5, 5/3, 16/9, 9/5, 15/8, 2",
//	//          S   R1     R2     R3   R4    G1    G2   G3    G4
//	sc_cmn22 = "1, 32/31, 16/15, 10/9, 9/8, 32/27, 6/5, 5/4, 81/64,
//	cents:	    0   90    112    182  204   294   316  386   408
//			 M1    M2     M3     M4     P     D1    D2   D3    D4
//			 4/3, 27/20, 45/32, 64/45, 3/2, 128/81, 8/5, 5/3, 27/16,
//			498   520    590    610   702    792   814  884   906
//			 N1     N2    N3    N4     S^
//			 16/9,  9/5, 15/8, 31/16,  2",
//			 996  1018  1088  1110   1200
// from http://www.carnaticcorner.com/articles/sruthis.html
// pitch cents (cmn22) from http://www.carnaticcorner.com/articles/sruthis.html
// fractions discussed in http://www.carnaticcorner.com/articles/22_srutis.htm

	// modes (mELakarthaa ragas)
	// (same scale when going up or down)
	// see http://carnaticmusicguru.sourceforge.net/raaga_list.htm
	mode_tb = [
		[],				// (0)
		[0, -1, -2, 0, 0, -1, -2],	// 1
		[0, -1, -2, 0, 0, -1, -1],	// 2
		[0, -1, -2, 0, 0, -1, 0],	// 3
		[0, -1, -2, 0, 0, 0, -1],	// 4
		[0, -1, -2, 0, 0, 0, 0],	// 5
		[0, -1, -2, 0, 0, 1, 0],	// 6

		[0, -1, -1, 0, 0, -1, -2],	// 7
		[0, -1, -1, 0, 0, -1, -1],	// 8
		[0, -1, -1, 0, 0, -1, 0],	// 9
		[0, -1, -1, 0, 0, 0, -1],	// 10
		[0, -1, -1, 0, 0, 0, 0],	// 11
		[0, -1, -1, 0, 0, 1, 0],	// 12

		[0, -1, 0, 0, 0, -1, -2],	// 13
		[0, -1, 0, 0, 0, -1, -1],	// 14
		[0, -1, 0, 0, 0, -1, 0],	// 15
		[0, -1, 0, 0, 0, 0, -1],	// 16
		[0, -1, 0, 0, 0, 0, 0],		// 17
		[0, -1, 0, 0, 0, 1, 0],		// 18

		[0, 0, -1, 0, 0, -1, -2],	// 19
		[0, 0, -1, 0, 0, -1, -1],	// 20
		[0, 0, -1, 0, 0, -1, 0],	// 21
		[0, 0, -1, 0, 0, 0, -1],	// 22
		[0, 0, -1, 0, 0, 0, 0],		// 23
		[0, 0, -1, 0, 0, 1, 0],		// 24

		[0, 0, 0, 0, 0, -1, -2],	// 25
		[0, 0, 0, 0, 0, -1, -1],	// 26
		[0, 0, 0, 0, 0, -1, 0],		// 27
		[0, 0, 0, 0, 0, 0, -1],		// 28
		[0, 0, 0, 0, 0, 0, 0],		// 29
		[0, 0, 0, 0, 0, 1, 0],		// 30

		[0, 1, 0, 0, 0, -1, -2],	// 31
		[0, 1, 0, 0, 0, -1, -1],	// 32
		[0, 1, 0, 0, 0, -1, 0],		// 33
		[0, 1, 0, 0, 0, 0, -1],		// 34
		[0, 1, 0, 0, 0, 0, 0],		// 35
		[0, 1, 0, 0, 0, 1, 0],		// 36

		[0, -1, -2, 1, 0, -1, -2],	// 37
		[0, -1, -2, 1, 0, -1, -1],	// 38
		[0, -1, -2, 1, 0, -1, 0],	// 39
		[0, -1, -2, 1, 0, 0, -1],	// 40
		[0, -1, -2, 1, 0, 0, 0],	// 41
		[0, -1, -2, 1, 0, 1, 0],	// 42

		[0, -1, -1, 1, 0, -1, -2],	// 43
		[0, -1, -1, 1, 0, -1, -1],	// 44
		[0, -1, -1, 1, 0, -1, 0],	// 45
		[0, -1, -1, 1, 0, 0, -1],	// 46
		[0, -1, -1, 1, 0, 0, 0],	// 47
		[0, -1, -1, 1, 0, 1, 0],	// 48

		[0, -1, 0, 1, 0, -1, -2],	// 49
		[0, -1, 0, 1, 0, -1, -1],	// 50
		[0, -1, 0, 1, 0, -1, 0],	// 51
		[0, -1, 0, 1, 0, 0, -1],	// 52
		[0, -1, 0, 1, 0, 0, 0],		// 53
		[0, -1, 0, 1, 0, 1, 0],		// 54

		[0, 0, -1, 1, 0, -1, -2],	// 55
		[0, 0, -1, 1, 0, -1, -1],	// 56
		[0, 0, -1, 1, 0, -1, 0],	// 57
		[0, 0, -1, 1, 0, 0, -1],	// 58
		[0, 0, -1, 1, 0, 0, 0],		// 59
		[0, 0, -1, 1, 0, 1, 0],		// 60

		[0, 0, 0, 1, 0, -1, -2],	// 61
		[0, 0, 0, 1, 0, -1, -1],	// 62
		[0, 0, 0, 1, 0, -1, 0],		// 63
		[0, 0, 0, 1, 0, 0, -1],		// 64
		[0, 0, 0, 1, 0, 0, 0],		// 65
		[0, 0, 0, 1, 0, 1, 0],		// 66

		[0, 1, 0, 1, 0, -1, -2],	// 67
		[0, 1, 0, 1, 0, -1, -1],	// 68
		[0, 1, 0, 1, 0, -1, 0],		// 69
		[0, 1, 0, 1, 0, 0, -1],		// 70
		[0, 1, 0, 1, 0, 0, 0],		// 71
		[0, 1, 0, 1, 0, 1, 0]		// 72
	], // mode_tb

	a_tb = ['__', '_', '', '^', '^^'],

	x_ = 0,				// current tune index
	otosvg				// original mus.tosvg()

    function parse_cmn() {
    var	li, c, c1, i, j, dur, s, a, wd, p_v, eol, str, info, ly1, ppit, last_nt,
	mode = mode_tb[29],		// default mode C major
	C = tm.C,
	cfmt = mus.cfmt(),
	parse = mus.parse,
	bol = parse.eol,
	mb = parse.mb		// music source
	instr = 15,		// instrument (veena)
	wl = [],		// lyrics (w: line)
	nw = 1,			// put new word
	ly = [],
	bdur = 1		// duration base

	// build the explicit list of accidentals of K:
	function build_exp() {
		if (!cfmt.cmn_raga)
//			return " treble"
			return "none"
	    var	a,
		t = a_tb,
		e = ''

		a = mode[6]
		if (a)
			e += t[a + 2] + 'B'
		a = mode[2]
		if (a)
			e += t[a + 2] + 'e'
		a = mode[5]
		if (a)
			e += t[a + 2] + 'A'
		a = mode[1]
		if (a)
			e += t[a + 2] + 'd'
		a = mode[3]
		if (a)
			e += t[a + 2] + 'f'
		if (!e)
			return 'C'
//		return "none exp " + e + " clef=treble"
		return "none exp " + e
	} // build_exp()

	// output a note or rest
	// - adjust the duration,
	// - for a note, adjust the octaveand set the MIDI pitch
	// - link the symbol
	// - no duration nor symbol anymore
	function sym_out() {
	    var	m, j

		if (!s)
			return
		dur *= C.BLEN / 4
		s.dur = s.dur_orig = dur
		for (m = 0; m <= s.nhd; m++)
			s.notes[m].dur = dur

		if (s.typ == C.note) {
			if (!cfmt.cmn_case		// is same letters
			 && cfmt.cmn_syntax != 1) {
				j = s.notes[0].pit
				if (ppit - j >= 4 && j < 3)	// A
					s.notes[0].pit += 7
				else if (ppit - j <= -4 && j > 2) // G
					s.notes[0].pit -= 7
				ppit = s.notes[0].pit
			}
			s.notes[0].midi = mus.pit2mid(s.notes[0].pit + 37,
							s.notes[0].acc)
			last_nt = s
		}
		mus.sym_link(s)
		dur = 0
		s = null
	} // sym_out()

	// build a mode from a list of notes
	// list of notes:	S R[1,2,3] G[1,2,3] M[1,2] P D[1,2,3] N[1,2,3]
	// shift from 12edo	0   -1..1   -2..0     0,1      -1..1    -2..0
	// example: S R2 G2 M1 D2 N2 S => [0, 0, -1, 0, 0, 0, -1] = mode 22
	var mod_sh = [0, -1, -2, 0, 0, -1, -2]
	function get_mode(a) {
	    var	i, v,
		mod = [0, 0, 0, 0, 0, 0, 0],
		a = a.split(/\s+/)

		while (1) {
			v = a.shift()
			if (!v)
				break
			i = 'SRGMPDN'.indexOf(v[0])
			if (i < 0)
				break
			v = +v[1]
			if (v)
				mod[i] = v + mod_sh[i]
		}
		return mod
	} // get_mode()

	// get a word or a double quoted string
	function get_word() {
	    var	r, i

		spsk()
		if (mb[bol] == '"') {
			i = bol
			while (1) {
				i = mb.indexOf('"', i + 1)
				if (i < 0)
					break
				if (mb[i - 1] != '\\')
					break
			}
			if (i < 0)
				return ""
			r = bol + 1
			bol = i + 1
			return mb.slice(r, i)
		}
		i = bol
		while (1) {
			switch (mb[bol]) {
			case ' ':
			case '\t':
			case '\n':
				break
			default:
				bol++
				continue
			}
			break
		}
		return mb.slice(i, bol)
	}

	// skip spaces
	function spsk() {
		while (1) {
			switch (mb[bol]) {
			case ' ':
			case '\t':
			case '\n':
				bol++
				continue
			}
			break
		}
	} // spsk()

	// ----- parse_cmn -----

	// parse the carnatic source
	while (1) {
		if (bol) {
			bol = mb.indexOf('\n', bol)	// go to end of previous line
			if (bol < 0) {
				bol = mb.length
				break
			}
			bol++
		}
		spsk()
		if (mb[bol] == '\n') {
			bol++				// empty line
			continue
		}
		if (bol == mb.length)			// source end
			break

		// handle ABC parameters
		if (mb[bol] == '%') {
			if (mb[++bol] != '%')
				continue		// comment
			i = bol + 1
			bol = mb.indexOf('\n', i)
			if (bol < 0) {
				mus.do_pscom(mb.slice(i))
				parse.eol = parse.bol = mb.length
				return		// no carnatic music anymore
			}
			mus.do_pscom(mb.slice(i, bol))
			if (!cfmt.cmni) {
				parse.eol = bol
				return
			}
			continue
		}

		// the beginning of file may give the syntax
		if (mb[bol] == '9'
			&& mb.substr(bol, 7) == '999999 ') {
			cfmt.cmn_syntax = 1		// Gaayaka

			bol += 7
			get_word()			// '0'

			// get the mode
			wd = get_word()			// mode
			mode = [0,
				(+wd[1] - 2),
				(+wd[2] - 2),
				(+wd[3] - 1),
				0,
				(+wd[5] - 2),
				(+wd[6] - 2)]
//console.log('mode: '+mode)

			wd = get_word()
			mus.new_tempo('1/4=' + ((6000 / wd) | 0))

			get_word()			// notes per aksharam
			get_word()			// number of aksharams in thaalam
			get_word()			// thaalam number
			get_word()			// thalaam sound structure

			spsk()
		} else if (!cfmt.cmn_syntax) {
			if (mb.indexOf('{', bol) >= 0
			 && mb.indexOf('}', bol) < 0)
				cfmt.cmn_syntax = 2	// raagaa
		}

		// handle the comments
		if (mb[bol] == '[') {
			j = mb.indexOf(']', bol)
			mus.do_begin_end("text", "fill", mb.slice(bol + 1, j))
			j = mb.indexOf('\n', j)
			bol = j
			continue
		}
		if (mb[bol] == '{' && cfmt.cmn_syntax == 2) {
			j = mb.indexOf('\n', bol)
			mus.do_begin_end("text", "fill", mb.slice(bol + 1, j))
			bol = j
			continue
		}

		// directives
		if (mb[bol] == '#') {			// raagaa
			if (!cfmt.cmn_syntax)
				cfmt.cmn_syntax = 2
			switch (mb[bol + 1]) {
			case 'D':			//tempo (in beats/mn)
				i = parseInt(mb.substr(bol + 2, 8))
				if (isNaN(i) || i <= 0) {
					mus.error(0, null,
						'bad tempo ' + mb.substr(bol + 2, 8))
					break
				}
				parse.istart = bol
				parse.iend = bol + 3
				mus.new_tempo('1/4=' + i)
				break
			case 'I':			// instrument
//--fixme: may be "#I3;4" for 2 instruments at the same time (violon, piano?)
				i = parseInt(mb.substr(bol + 2, 8))
				if (isNaN(i) || i < 0) {
					mus.error(0, null,
						'bad instrument ' + mb.substr(bol + 2, 8))
					break
				}
				switch (i) {
				default:
//				case 0:			// Veena
//				case 1:
					instr = 15	// Dulcimer
					break
				case 2:			// Flute
					instr = 74	// Recorder
					break
//				case 3:			// Sarod
				case 4:			// Violin	
					instr = 41	// Viola
					break
				case 5:			// Sitar
					instr = 104	// Sitar
					break
				case 6:			// Shanai
					instr = 111	// Shanai
					break
				case 7:			// Guitar
					instr = 24
					break
				case 8:			// Piano
					instr = 0
					break
				}
				break
			case 'M':		// mode (mELakarthaa raga)
				i = parseInt(mb.substr(bol + 2, 8))
				if (isNaN(i) || i <= 0 || i >= 72) {
					mus.error(0, null,
						'bad mode ' + mb.substr(bol + 2, 8))
					break
				}
				mode = mode_tb[i]
				break
			case 'N':			// temperament
				i = parseInt(mb.substr(bol + 2, 8))
				switch (i) {
//				case '0':		// 12-EDO
				case '1':
					do_pscom('temperament ' + sc_cmn12)
//console.log('temp 1')
					break
//				case '2':
//					scale = sc_cmn12
//					break
//				case '3':
//					scale = sc_cmn12
//					break
				}
				break
			}
			continue
		}
		
		if (mb[bol] == '{') {
			j = bol + 1
			bol = mb.indexOf('}', j)
			switch (mb[j]) {
			case 'm':		// mode
			case 'M':
				i = parseInt(mb.substr(j + 1, bol))
				if (isNaN(i) || i <= 0 || i >= 72) {
					mus.error(0, null,
						'bad mode ' + mb.substr(j, bol))
					break
				}
				mode = mode_tb[i]
				break
			case 't':
			case 'T':		// tempo
				i = parseInt(mb.substr(j + 1, bol))
				if (isNaN(i) || i <= 0) {
					mus.error(0, null,
						'bad tempo ' + mb.substr(j + 1, bol))
					break
				}
				parse.istart = j
				parse.iend = i
				mus.new_tempo('1/4=' + ((6000 / i) | 0))
				break
			}
			continue
		}

		// directives
		spsk()
		i = mb.indexOf(':', bol)
		if (i > bol && i < bol + 32) {
			j = mb.indexOf('\n', i)
			wd = mb.slice(i + 1, j).trim()
			info = mus.info()

			// https://gk.palem.in/music.html
			switch (mb.slice(bol, i)) {
			case 'Song':
				str = {}
				mus.strnew(str, wd, "title")
				if (!info.T)
					info.T = []
				info.T.push(str)
				break
			case 'Movie Name':
				str = {}
				mus.strnew(str, wd, "history")
				if (!info.B)
					info.B = []
				info.B.push(str)
				break
			case 'Composer':
				str = {}
				mus.strnew(str, wd, "composer")
				if (!info.C)
					info.C = []
				info.C.push(str)
				break
			case 'Base notes:':
			case 'ARO:':
				mode = get_mode(wd)
				break
			case 'Pallavi':
			case 'Charanam':
			case 'Charanam 1':
			case 'Charanam 2':
				parse.istart = bol
				parse.iend = j
				if (parse.state != 3) {
					if (!mus.parse.state) {
						mus.parse.state = 1
						mus.info().X = (++x_).toString()
					}
					mus.get_key(build_exp())
					mus.get_voice("1")
				}
				mus.do_info('P', mb.slice(bol, i))
				break
			}
			bol = j
			continue
		}

		// line starting with ':' and ending with '\\n:' or '\\t:'
		//	https://github.com/shivanan/cmc (.notes)
		if (i == bol) {
			j = mb.indexOf(':', i + 1)
			if (j > 0) {
				if (mb[j - 2] == '\\')
					j -= 2
				mus.do_pscom("text " + mb.slice(i + 1, j))
			} else {
				j = i
			}
			bol = j + 2
			continue
		}

		// enter the tune body if not done yet
		parse.istart = bol
		parse.iend = bol + 4
		if (parse.state != 3) {
			if (!mus.parse.state) {
				mus.parse.state = 1
				mus.info().X = (++x_).toString()
			}
			mus.get_key(build_exp())
			mus.get_voice("1")
			mus.do_pscom("MIDI program " + instr)
		}

		// look if there is a lyric line after the music line
		i = mb.indexOf('\n', bol)
		wd = ""
		if (i > 0) {
			j = mb.indexOf('\n', ++i)
			if (j > 0)
				wd = mb.slice(i, j)
		}
		if (wd					// line after music
		 && (cfmt.cmn_nlines
		  || wd.indexOf('[') >= 0)) {
			ly = wd.replace(/^[\s[]*/, '')
				.replace(/[-\s\]|]*$/, '')
				.split(/\s+/)
			eol = j - 1
		} else {
			ly.length = 0
		}

		// parse the music line
		while (1) {
			c = mb[bol++]
			if (!c || c == '\n')
				break
			j = 'SRGMPDN'.indexOf(c.toUpperCase())
			if (j >= 0) {			// note found
				sym_out()		// output the previous symbol
				s = {
					typ: C.note,
					stem: 0,
					multi: 0,
					nhd: 0,
					xmx: 0,
//					dur: ,
//					dur_orig: ,
					shhd: 0,
					shac: 0,
					notes: [{
						shhd: 0,
						shac: 0
					}]
				}
				dur = bdur
				parse.istart = bol
				c1 = mb[bol]
//				if (c1 >= '1' && c1 <= '6') {
				if (c1 >= '1' && c1 <= '3') {
					bol++
//					if (j != '0' && j != '4')
						c1 = +c1 - mod_sh[+j]
//					switch (j) {
//					case '3':	// M > F
//						c1 = +c1 - 1
//						break
//					default:
////					case '1':	// R > D
////					case '5':	// D > A
//						c1 = +c1 - 2
//						break
//					case '2':	// G > E
//					case '6':	// N > B
//						c1 = +c1 - 3
//						break
//					}
				} else if (mode && !cfmt.cmn_raga) { // if by accidental
					c1 = mode[j]
				}

				// accidental and note
				if (typeof c1 == "number" && c1)
					s.notes[0].acc = c1
				s.notes[0].pit = j - 2

				// (some) octave
				if (cfmt.cmn_syntax == 1) {	// Gaayaka
					bol++
					if (c < 'Z')
						s.notes[0].pit += 7
					else if (c1 < 'Z')
						s.notes[0].pit -= 7
					switch (mb[bol]) {	// if 3 letters
					case 'a':
					case 'e':
					case 'i':
						bol++
						dur = 2 * bdur
						break
					}
				} else if (cfmt.cmn_case == 1 && c < 'Z') {
					s.notes[0].pit += 7
				} else {
					switch (mb[bol]) {	// if 2 letters
					case 'a':
					case 'e':
					case 'i':
						bol++
						dur = 2 * bdur
						break
					}
				}
				parse.iend = bol

				// add the lyrics
				if (ly.length || (ly1 && ly1.length)) {
					if (!wl.length) {
						p_v = mus.get_curvoice()
						p_v.lyric_restart = s
					}
					if (nw) {		// if new word
						if (cfmt.cmn_syntax == 1)
							nw = 0
						if (ly1 && ly1.length) {
							c = ly1.shift()
							if (ly1.length)
								c += '-'
							wl.push(c)
						} else {
							c = ly.shift()
							if (c) {
							    if (!cfmt.cmn_syntax) {


							// split the word into syl-la-bles
								ly1 = c.match(
//fixme: almost ok, but not all cases
//	/[^aeiouy]*[aeiouy]+(?:[^aeiouy]*$|[^aeiouy](?=[^aeiouy]))?/gi)
	/[^aeiouy]*[aeiouy]+(?:[^aeiouy]*$|[^aeiouyh](?=[^aeiouyh]))?/gi)
								if (ly1 && ly1.length) {
									c = ly1.shift()
									if (ly.length)
										c += '-'
								}
							    }
							    wl.push(c)
							}
						}
					} else {
						wl.push('*')
					}
				}
				continue
			}
			switch (c) {
			case ' ':
			case '\t':
				continue
			case '|':
				sym_out()
				s = {
					typ: C.bar,
					bar_type: '|',
					dur: 0,
					multi: 0
				}
				parse.istart = bol
				if (mb[bol] == '|') {
					s.bar_type = '||'
					bol++
				}
				parse.iend = bol + 1
				mus.sym_link(s)
				s = null
				continue
			case '^':
			case '+':
			case "'":
			case '↑':
			case '\u0307':			// (dotaccentcmb)
				if (s)
					s.notes[0].pit += 7
				break
			case '↓':
			case '\u0323':			// (dotbelowcmb)
				if (s)
					s.notes[0].pit -= 7
				break
			case '-':
				if (s)
					s.beam_end = 1 //true
				else if (last_nt)
					last_nt.beam_end = 1 //true
				sym_out()
				if (cfmt.cmn_syntax == 1) // Gaayaka
					nw = 1		// next lyric word
				break
			case '.':
				if (cfmt.cmn_syntax != 1) { // not Gaayaka
					s.notes[0].pit -= 7
					break
				}
			case ',':
			case ';':
				if (!s) {
					s = {
						typ: C.rest,
						stem: 0,
						multi: 0,
						nhd: 0,
						xmx: 0,
//						dur: ,
//						dur_orig: ,
						notes: [{
							pit: 0
						}]
					}
					parse.istart = bol
					parse.iend = bol + 1
					dur = bdur
					if (c == ';')
						dur += bdur
					else if (c == '.')
						dur -= bdur / 2
					sym_out()
					break
				}
				dur += bdur
				if (c == ';')
					dur += bdur
				else if (c == '.')
					dur -= bdur / 2
				break
			case ':':
				dur /= 2
				break
			case '(':
				sym_out()
				bdur /= 2
				break
			case ')':
				sym_out()
				bdur *= 2
				break
			}
		}
		sym_out()

		// set the lyrics if any
// with cfmt.cmn_syntax==1
//	a syllable goes in a note when '-' in the music
//	if no '-', '*' (continuation/rest)
//	there can be '-' in lyrics between 2 syllables
//	can there be no '-' between syllables? yes!
// otherwise
//	one syllable -> one note
//	'-' alone -> on rest
//	'-' just after syllable -> in next note
//	there may be no '-' between syllables

		if (wl.length) {
			mus.get_lyrics(wl.join(' '))
			wl.length = 0
			bol = eol
		}
		mus.get_curvoice().eoln = 1 //true
		bol--				// back to '\n'
	}

	parse.eol = bol
    } // parse_cmn()

    // start parsing a CMN sequence
    function start_cmn() {
	mus.glovar().ulen = tm.C.BLEN / 4

	parse_cmn()

	if (mus.parse.state == 3) {
		mus.generate()		// generate the last tune
		mus.init_tune()
	}
    } // start_cmn

    // parse a CMN source
    function tosvg(of, fn, mb, bol, eof) {
    var	parse = mus.parse

	parse.mb = mb
	parse.fname = fn
	if (bol == undefined)
		bol = 0
	parse.eol = bol
	start_cmn()
    } // tosvg()

    // get the options
    function set_fmt(of, cmd, parm) {
	if (cmd != "cmni") {
		of(cmd, parm)
		return
	}
    var	i,
	cfmt = mus.cfmt(),
	a = mus.get_bool(parm)

	cfmt.cmni = a
	if (!a || mus.parse.state) {	// not in global state
		if (otosvg) {		// restore the tosvg function
			mus.tosvg = otosvg
			otosvg = null
		}
		return
	}

	// get the parameters
	a = parm.match(/[^\s"=]+=?|"[^"]+"/g)
	if (!a) {
//fixme
//		syntax(1, "Bad %%cmn options")
		return
	}
	while (1) {
		i = a.shift()
		if (!i)
			break
		switch (i) {
		case "case=":
		case "nlines=":
		case "raga=":
		case "syntax=":
//		case "debug=":
			i = 'cmn_' + i.slice(0, -1)
			cfmt[i] = +a.shift()
			break
		}
	}
	if (!otosvg) {
		otosvg = mus.tosvg
		mus.tosvg = tosvg.bind(mus, mus.tosvg)
	}
	start_cmn()
    } //set_fmt()

	// set the hooks
	mus.set_format = set_fmt.bind(mus, mus.set_format)
} // cmni()

if (!tm.mhooks)
	tm.mhooks = {}
tm.mhooks.cmni = tm.cmni
