// gamelan.js - module to output Gamelan (indonesian) music sheets
//
// Copyright (C) 2020-2024 Jean-Francois Moine
//
// This file is part of txtmus.
//
// txtmus is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// txtmus is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with txtmus.  If not, see <http://www.gnu.org/licenses/>.
//
// This module is loaded when "%%gamelan" appears in a ABC source.
//
// Parameters (none)
//	%%gamelan 1
// scale:
// - sléndro - 5 equal tones
//	5-TET	C   D   E     G   A
//	detune	0 +40 +80 . +20 +60 .
// - pélog - 7 unequal tones
//		D _E+ F- ^G+ A _B c+ d
//		1  2  3   4  5  6 7  1
// (first note = ding)
//   Bali
//	selisir ^C D E ^G A	12356
//	tembung	E F A B c	45612
//	sunaren	F G B c d (?)	56723
//   Java
//	bem	D_E ^G A _B	12456
//	barang	_E F A _B c	23567
// numbers
//	??	E F G B c
//	??	C ^C _E G _A (or ^C D E ^G A)
// dot = continuation (not rest)

if (typeof tm == "undefined")
	tm = {}

tm.gamelan = function(mus) {

    var	cde2fcg = new Int8Array([0, 2, 4, -1, 1, 3, 5]),
	cgd2cde = new Int8Array([0, -4, -1, -5, -2, -6, -3,
				0, -4, -1, -5, -2, -6, -3, 0]),
	acc2 = new Int8Array([-2, -1, 3, 1, 2])

// change %%staves and %%score
    function do_pscom(of, p) {
	switch (p.match(/\w+/)[0]) {
	case 'staves':
	case 'score':
		p = p.replace(/\(|\)/g, '')
		break
	}
	of(p)
    } // do_pscom()

// adjust some symbols before the generation
    function output_music(of) {
    var	C = tm.C,
	cur_sy = mus.get_cur_sy(),
	v_tb = mus.get_v_tb()

	if (!mus.cfmt().gamelan) {
		of()
		return
	}

	// expand dots and long notes/rests
	function slice(s) {
	    var	m, n, s2, s3, d, d_orig

		if (s.dur <= C.BLEN * 3 / 8) {
			if ((s.dur_orig / 9 | 0) * 9 != s.dur_orig)
				return
			d = s.dur / 3
			d_orig = s.dur_orig / 3
			s.dur -= d
			s.dur_orig -= d_orig
			n = 1
		} else {
			if (s.dur >= C.BLEN)
				n = 3
			else if (s.dur == C.BLEN / 2)
				n = 1
			else
				n = 2
			d = d_orig = C.BLEN / 4
			s.dur = s.dur_orig = C.BLEN / 4
		}
		for (m = 0; m <= s.nhd; m++)
			s.notes[m].dur = s.dur
		s.beamon = true
		while (--n >= 0) {
			s2 = {
				typ: C.rest,
				v: s.v,
				p_v: s.p_v,
				st: s.st,
				fmt: s.fmt,
				dur: d,
				dur_orig: d_orig,
				stem: 1,
				multi: 0,
				nhd: 0,
				notes: [{
					pit: s.notes[0].pit,
					jn: 8
				}],
				xmx: 0,
				beamon: true,
				noplay: true,
				time: s.time + s.dur,
				prev: s,
				next: s.next
			}
			s.next = s2
			if (s2.next)
				s2.next.prev = s2

			if (!s.ts_next) {
				s.ts_next = s2
				if (s.soln)
					s.soln = false
				s2.ts_prev = s
				s2.seqst = true
			} else {
			    for (s3 = s.ts_next; s3; s3 = s3.ts_next) {
				if (s3.time < s2.time)
					continue
				if (s3.time > s2.time) {
					s2.seqst = true
					s3 = s3.ts_prev
				}
				s2.ts_next = s3.ts_next
				s2.ts_prev = s3
				if (s2.ts_next)
					s2.ts_next.ts_prev = s2
				s3.ts_next = s2
				break
			    }
			}
			s = s2
		}
	} // slice()

	// replace the tied notes by a '.'
	function do_tie(s) {
	    var	end_time = s.time + s.dur
		while (1) {
			s = s.ts_next
			if (!s || s.time > end_time)
				return	// ?!
			if (s.typ == C.note
			 && s.time == end_time)
				break
		}
		s.notes[0].jn = 8
		s.notes[0].jo = 2
	} // do_tie()

	function set_sym(p_v) {
	    var s, s2, note, pit, nn, p, a, m, i,
		sf = p_v.key.k_sf
		delta = cgd2cde[sf + 7] + 16 // = 2 + 14

		delete p_v.key.k_a_acc		// no accidental

		// no (visible) clef
		p_v.clef.invis = true

		// scan the voice
		for (s = p_v.sym; s; s = s.next) {
			s.st = p_v.st
			switch (s.typ) {
			case C.clef:
				s.invis = true
//				continue
			default:
				continue
			case C.key:
				delta = cgd2cde[s.k_sf + 7] - 2
				continue
			case C.rest:
				if (s.notes[0].jn)
					continue
				s.notes[0].jn = 0
				s.notes[0].pit = 3	// "A"
				slice(s)
				continue
			case C.note:		// change the notes
				break
			}

			s.stem = 1
			s.stemless = true

			// set the slurs offset and direction
			if (s.sls) {
				for (i = 0; i < s.sls.length; i++)
					s.sls[i].ty = C.SL_BELOW
			}

			for (m = 0; m <= s.nhd; m++) {
				note = s.notes[m]

				// note head
				p = note.pit
				pit = p + delta
				if (note.jn == undefined) {	// if not tied
					note.jn = ((pit + 77) % 7) + 1	// note number
					note.jo = (pit / 7) | 0	// octave number
				}

				// set a fixed offset to the note
				// for the slurs and decorations
				note.pit = 3			// "A"

				// accidentals
				a = note.acc
				if (a) {
					nn = cde2fcg[(p + 5 + 16 * 7) % 7] - sf
					if (a != 3)
						nn += a * 7
					nn = ((((nn + 1 + 21) / 7) | 0) + 2 - 3 + 32 * 5) % 5
					note.acc = acc2[nn]
				}

				if (note.tie_ty) {
					do_tie(s)
					delete note.tie_ty
				}
			}

			// change the dots and the long notes
			slice(s)

			// replace the staccato dot
			if (s.a_dd) {
				for (i = 0; i < s.a_dd.length; i++) {
					if (s.a_dd[i].glyph == "stc") {
						mus.deco_put("gstc", s)
						s.a_dd[i] = s.a_dd.pop()
					}
				}
			}
		}
	} // set_sym()

	// -- output_music --

	for (v = 0; v < v_tb.length; v++)
		set_sym(v_tb[v])

	of()
    } // output_music()

    function draw_symbols(of, p_voice) {
    var	i, m, nl, note, s, s2, x, y,
	C = tm.C,
	dot = "\ue1e7",
	staff_tb = mus.get_staff_tb(),
	out_svg = mus.out_svg,
	out_sxsy = mus.out_sxsy,
	xypath = mus.xypath

	if (!mus.cfmt().gamelan) {
		of(p_voice)
		return
	}

	// draw the duration lines above the notes
	function draw_dur(s1, y, s2, n, nl) {
	    var s, s3

		xypath(s1.x - 3, y + 24)
		out_svg('h' + (s2.x - s1.x + 8).toFixed(1) + '"/>\n')	// "
		y -= 2.5
		while (++n <= nl) {
			s = s1
			while (1) {
				if (s.nflags && s.nflags >= n) {
					s3 = s
					while (s != s2) {
						if (s.next.bmbrk && n > s.next.bmbrk
						 || (s.next.nflags
						  && s.next.nflags < n))
							break
						s = s.next
					}
					draw_dur(s3, y, s, n, nl)
				}
				if (s == s2)
					break
				s = s.next
			}
		}
	} // draw_dur()

	function out_mus(x, y, p) {
		out_svg('<text x="')
		out_sxsy(x, '" y="', y)
		out_svg('">' + p + '</text>\n')
	} // out_txt()

	function out_txt(x, y, p) {
		out_svg('<text class="bn" x="')
		out_sxsy(x, '" y="', y)
		out_svg('">' + p + '</text>\n')
	} // out_txt()

	function draw_hd(s, x, y) {
	    var	m, note, ym

		for (m = 0; m <= s.nhd; m++) {
			note = s.notes[m]
			out_txt(x - 3.5, y + 8, "01234567."[note.jn])
			if (note.acc) {
				out_svg('<path class="stroke" stroke-width="1.1" d="M')
				if (note.acc > 0) {
					out_sxsy(x - 6, ' ', y + 10)
					out_svg("l12 -6")
				} else {
					out_sxsy(x - 6, ' ', y + 16)
					out_svg("l12 6")
				}
				out_svg('"/>\n')
			}
			if (note.jo > 2) {
				out_mus(x - 1, y + 23, dot)
				if (note.jo > 3) {
					y += 3
					out_mus(x - 1, y + 23.4, dot)
				}
			} else if (note.jo < 2) {
				ym = y + 4
				out_mus(x - 1, ym, dot)
			}
			y += 20
		}
	} // draw_hd()

	// -- draw_symbols --

	for (s = p_voice.sym; s; s = s.next) {
		if (s.invis)
			continue
		switch (s.typ) {
		case C.note:
		case C.rest:
			draw_hd(s, s.x, staff_tb[s.st].y)
			break
		}
	}

	// draw the (pseudo) beams
	for (s = p_voice.sym; s; s = s.next) {
		if (s.invis)
			continue
		switch (s.typ) {
		case C.note:
		case C.rest:
			nl = s.nflags
			if (nl <= 0)
				continue
			s2 = s
			while (s.next && s.next.nflags > 0) {
				s = s.next
				if (s.nflags > nl)
					nl = s.nflags
				if (s.beam_end)
					break
			}
			draw_dur(s2, staff_tb[s.st].y + 7,
				s, 1, nl)
			break
		}
	}
    } // draw_symbols()

// set some parameters
    function set_fmt(of, cmd, param) {
	if (cmd == "gamelan") {
	    var	cfmt = mus.cfmt()

		if (!mus.get_bool(param))
			return
		cfmt.gamelan = true
		cfmt.staffsep = 20
		cfmt.sysstaffsep = 14
		mus.set_v_param("stafflines", "...")
		cfmt.tuplets = [0, 1, 0, 1]	// [auto, slur, number, above]
		return
	}
	of(cmd, param)
    } // set_fmt()

// adjust some values
    function set_pitch(of, last_s) {
	of(last_s)
	if (!last_s
	 || !mus.cfmt().gamelan)
		return			// first time

    var	C = tm.C
	
	for (var s = mus.get_tsfirst(); s; s = s.ts_next) {
		switch (s.typ) {

		// adjust the vertical spacing above the note heads
		case C.note:
			s.ymx = 20 * s.nhd + (s.nflags > 0 ? 30 : 24)
			if (s.notes[s.nhd].jo > 2) {
				s.ymx += 3
				if (s.notes[s.nhd].jo > 3)
					s.ymx += 2
			}
			s.ys = s.ymx		// (for tuplets)
			break
		}
	}
    } // set_pitch()

// set the width of some symbols
    function set_width(of, s) {
	of(s)
	if (!mus.cfmt().gamelan)
		return

	switch (s.typ) {
	case tm.C.clef:
	case tm.C.key:
	case tm.C.meter:
		s.wl = s.wr = .1		// (must not be null)
		break
	}
    } // set_width()

	// set the hooks
	mus.do_pscom = do_pscom.bind(mus, mus.do_pscom)
	mus.draw_symbols = draw_symbols.bind(mus, mus.draw_symbols)
	mus.output_music = output_music.bind(mus, mus.output_music)
	mus.set_format = set_fmt.bind(mus, mus.set_format)
	mus.set_pitch = set_pitch.bind(mus, mus.set_pitch)
	mus.set_width = set_width.bind(mus, mus.set_width)

	// big staccato dot
	mus.glyphs.gstc = '<circle id="gstc" cx="0" cy="-3" r="2"/>'
	mus.decos.gstc = "0 gstc 5 1 1"

	mus.add_style("\n.bn {font-family:sans-serif; font-size:16px}")
} // gamelan

if (!tm.mhooks)
	tm.mhooks = {}
tm.mhooks.gamelan = tm.gamelan
