// mup.js - module to handle MUP tunes
//	http://www.arkkra.com
//
// Copyright (C) 2023 Jean-Francois Moine
//
// This file is part of txtmus.
//
// txtmus is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// txtmus is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with txtmus.  If not, see <http://www.gnu.org/licenses/>.
//
// This module is loaded when "%%mup" appears in a text source.
//

if (typeof tm == "undefined")
	tm = {}

tm.mup = function(mus) {

    var	font_st = {				// font styles
	rom: "normal",
	ital: "italic",
	bold: "bold",
	boldital: "bolditalic"
	}

    // handle a MUP sequence
    function set_fmt(of, cmd, parm) {
	if (cmd != "mup") {
		of(cmd, parm)
		return
	}

    var	cfmt = mus.cfmt()

	cfmt.mup = mus.get_bool(parm)
	if (!cfmt.mup)
		return

    var	bol, eol, in_tune,
	C = tm.C,
	parse = mus.parse,
	mb = parse.mb,			// music source
	w = /\w+/,
	parm_tb = {
		tu: C.BLEN,		// time unit
		un: "in",		// size unit
	}

	// get a font
	function get_font() {
	    var	wd, i,
		ft = ""

		while (1) {
			spsk()
			if (mb[bol] == '"')
				break			// start of text
			if (mb[bol] == '(') {
				i = ++bol
				bol = mb.indexOf(')', bol)
				if (bol < 0) {
					syntax(1, "No end of font size")
					bol = i
					break
				}
				if (ft)
					ft += ' '
				ft += mb.slice(i, bol)
				bol++
				break			// the size is the last item
			}
			wd = get_word()
			i = font_st[wd]
			if (i) {
				if (ft)
					ft += ' '
				ft += i			// font style
			} else {
				if (ft)
					ft += ' '
				ft += wd		// font name
			}
		}
		return ft
	} // get_font()

	// set a parameter (key, value)
	function set_parm(ctx) {
	    var	k, v

		spsk()
		if (mb[bol] == '\n')
			return
		k = get_word()
		if (!k)
			return
		spsk()
		if (mb[bol] != '=')
			return
		bol++
		spsk()
		v = get_word()
		if (!v)
			return

		switch (k) {
//		case "barstyle":
//			break
		case "beamslope":
			v = v.match(/\d+(.\d+)?/)
			if (v)
				mus.set_format(k, v)
			break
		case "bottommargin":
			mus.set_format("botmargin", v + parm_tb.un)
			break
		case "brace":
		case "brace":
			parm_tb[k] = v
//fixme: marquer modif %%staves
			break
		case "cancelkey":
			mus.set_format(k, v)
			break
		case "clef":
			mus.do_pscom(k + " " + v)
			break
//		case "defoct":
////fixme: todo
//			break
//		case "font":
////fixme: todo
//			break
//		case "family":
////fixme: todo
//			break
		case "key":
//fixme: todo
// 4# or 3& or f#min
			if (v[0] >= '0' && v[0] <= '7') {
				k = v[1] == '#'
					? ["C","G","D","A","E","B","F#","C#"]
					: ["C","F","Bb","Eb","Ab","Db","Gb","Cb"]
				v = k[+v[0]]
			} else {
				v = v[0].toUpperCase()
					+ v.slice(1).replace(' '.'')
			}
			mus.get_key(v)
			break
		case "leftmargin":
		case "pageheight":
//fixme: pageheight asks for the module 'page'
		case "pagewidth":
		case "rightmargin":
		case "topmargin":
			mus.set_format(k, v + parm_tb.un)
			break
		case "lyricsfont":
			mus.set_format(k, font_st[v] + " *")
			break
		case "lyricsfontfamily":
			mus.set_format("lyricsfont", v + " *")
			break
		case "lyricssize":
			mus.set_format("lyricsfont", "* " + v)
			break
		case "packexp":
			mus.set_format("notespacingfactor", (+v / 2).toFixed(2) + "38")
			break
		case "scale":
			mus.set_format("pagescale", v)
			break
//		case "size":
//fixme: applies to the context (staff, header,footer..)
//			break
		case "staffscale":
			mus.set_format(k, v)
			break
		case "time":
			if (v == "common")
				v = "C"
			else if (v == "cut")
				v = "C|"
			else
				v = v.replace(/ ?(n|y)/, '')
			mus.new_meter(v)
			break
		case "units":
			parm_tb.un = v == "cm" ? v || "in"
			break
		}
	}

	// get a word or a double quoted string
	function get_word() {
	    var	r, i

		spsk()
		if (mb[bol] == '"') {
			i = bol
			while (1) {
				i = mb.indexOf('"', i + 1)
				if (i < 0)
					break
				if (mb[i - 1] != '\\')
					break
			}
			if (i < 0)
				return ""
			r = bol + 1
			bol = i + 1
			return mb.slice(r, i)
		}
		r = w.exec(mb)
		if (!r)
			return ""
		bol = w.lastIndex
		return r[0]
	}

	// skip spaces
	function spsk() {
		while (1) {
			switch (mb[bol]) {
			case ' ':
			case '\t':
			case '\n':
				bol++
				continue
			}
			break
		}
	} // spsk()

	function block_parse() {
	    var	cmd, ft, txt, txt2, txt3,
		ty = "justify"

		while (1) {
			spsk()
			if (mb[bol] == '\n')
				break
			cmd = get_word()
			if (cmd == "justified" || cmd == "ragged") {
				ty = cmd[0] == 'j' ? "justify" : cmd
				continue
			}
			switch (cmd) {
			case "paragraph":
				ft = get_font()
				if (ft)
					mus.set_format("textfont", ft)
				spsk()
//fixme				if (mb[bol] != '"') ...
				txt = get_word()
				mus.do_begin_end("text", ty, txt)
				break
			case "print":
			case "left":
			case "right":
			case "center":
--fixme:todo
				break
			case "title":
				ft = get_font()
				if (ft)
					mus.set_format("textfont", ft)
				spsk()
//fixme				if (mb[bol] != '"') ...
				txt = get_word()
				txt1 = txt2 = ""
				spsk()
				if (mb[bol] == '"') {
					txt2 = get_word()
					spsk()
					if (mb[bol] == '"') {
						txt3 = get_word()
--fixme: non: meme ligne
						do_pscom("text " + txt)
						do_pscom("center " + txt2)
//fixme						do_pscom("text " + txt3)
					} else {
--fixme: non: meme ligne
						do_pscom("text " + txt)
//fixme						do_pscom("text " + txt2)
					}
				} else {
					mus.do_pscom("center " + txt)
				}
				break
			default:
				return
			}
		}
	} // block_parse()

	function music_parse() {

		// parse a bar
		function bar_parse(ty) {
		    var	b_ty, s, dotted, wd

			if (ty[0] == 'd'
			 && (ty[1] == 'a' || ty[1] == 'o')) {
				dotted = 1 //true
				wd = get_word()
				ty = wd || "bar"
			}

			switch (ty[0]) {
			case 'b':				// bar
			case 'i': b_ty = '|'; break		// invisbar
			case 'd': b_ty = '||'; break		// dblbar
			case 'e': b_ty = '|]'; break		// endbar
			case 'r':
				switch (ty.slice(-1)) {
				case 't': b_ty = '|:'; break	// repeatstart
				case 'h': b_ty = '::'; break	// repeatboth
				case 'd': b_ty = ':|'; break	// repeatend
				}
				break
			}
			s = {
				typ: C.bar,
				bar_type: b_ty,
				dir: 0,
				multi: 0
			}
			if (ty[0] == 'i')
				s.invis = 1 //true
			if (dotted)
				s.dotted = 1 //true

			// extra arguments
			wd = get_word()
			switch (wd) {
//			case "pad":
//				break
			case "ending":
				spsk()
				if (mb[bol] == '"')
					s.text = get_word()
				break
			case "endending":
				s.rbstop = 1 //true
				break
//			default:
//				if (wd[0] == '=') {
////fixme: tag
//				}
//				break
			}
			mus.sym_link()
		} // bar_parse()

		// get the duration
		function get_d() {
		    var	d = mb[bol++]

			if (mb[bol] >= '1' && mb[bol] <= '9')
				d += mb[bol++]
			d = C.BLEN / +d
			if (mb[bol] == '/') {
				bol++
				d *= +mb[bol++]
			}
			while (mb[bol] == '.') {
				d += d / 2
				bol++
			}
//fixme: handle '+' and '-'
			return d
		} // get_d()

		// get the voice(s)
		function get_v() {
//fixme: handle many voices
		    var	v = (+mb[bol++] - 1) * 3	// 3 voices per staff
			if (mb[bol] == ' ') {
				bol++
				v += (+mb[bol++] - 1)
			}
			if (isNaN(v) {
				syntax(1, "Bad staff/voice")
				v = 0
			}
			return v.toString(v + 1)
		} // get_v()

		// parse the music elements of a measure
		function measure_parse() {
		    var	p, v, s, m, nt, c,
			d = parm_tb.tu			// default chord duration

			v = get_v()			// get the voice(s)
			mus.get_voice(v)
//			p_v = mus.get_curvoice()

			// loop on the chords/rests/...
			while (1) {
				spsk()
				if (mb[bol] == '\n')
					break
				if (mb[bol] == '<') {
--fixme
//<<score clef=bass; defoct = 4>> ou autre?
					???
					continue
				}
				if (mb[bol] >= '1' && mb[bol] <= '9')
					d = get_d()	// get a duration
//fixme: handle 'm' (duration of a measure)
				while (1) {
					p = 'cdefgabrsu'.indexOf(mb[bol++])
					if (p < 0) {
						syntax(1, "Unknown note")
						return bol
					}
					if (!s)
						s = {
							fname: parse.fname,
							istart: bol - 1,
							stem: 0
							multi: 0,
//							nhd: 0,
							xmx: 0,
							dur: d,
							dur_orig: d,
							notes: []
						}
					}
					nt = {
//						pit:
//						acc
						shhd: 0,
						shac: 0,
						dur: d
					}
					s.notes.push(nt)
					if (p < 7) {
						s.typ = C.note
						p -= 28
					} else {
						s.typ = C.rest
						if (p >= 8)
							s.invis = 1 //true
					}
					c = mb[bol++]
					switch (c) {
					case '#':
						nt.acc = 1
						break
					case 'x':
						nt.acc = 2
						break
					case '&':
						nt.acc = -1
						if (mb[bol] == '&') {
							nt.acc++
							bol++
						}
						break
					case 'n':
						nt.acc = 3
						break
					}
					if (nt.acc)
						c = mb[bol++]
					if (c >= '1' && c <= '9') {
						p += +c * 7		// octave
						c = mb[bol++]
					}
					if (c != ';')
						break
					spsk()
					if (mb[bol] == ';')
						break
				}
				s.nhd = s.notes.length - 1
				s.iend = bol
				for (m = 0; m <= s.nhd; m++)
					s.notes[m].pit = p
				mus.sym_link(s)

				while (mb[bol + 1] == ';') {	// while same element
					s = mus.clone(s)
					mus.sym_link(s)
					bol++
					spsk()
				}
			}
		} // measure_parse()

		// ----- music_parse -----
		while (1) {
			spsk()
			if (mb[bol] >= '1' && mb[bol] <= '9') {
				measure_parse()
				continue
			}
			w.lastIndex = bol
			r = w.exec(mb)
			if (!r)
				break
			bol = w.lastIndex
			switch (w[0]) {
			case "bar":
			case "dashed":
			case "dblbar":
			case "dotted":
			case "endbar":
			case "invisbar":
			case "repeatstart":
			case "repeatboth":
			case "repeatend":
--fixme: apres bar
//			case "restart":
				bar_parse(w[0])
				break
			case "lyrics":
				break
			case "multirest":
				break
//fixme: + rom, bold, boldital, ...
			}
		}
	} // music_parse()

	// --- parse the MUP file ---

// chaine magique: //!Mup-Arkkra (plus "-7.0")- file extension: .mup
// stepsize = moitie de la distance entre 2 lignes de la portee = 3px
//	; permet de separer les contexts et autres
//	score
//	music
//	1: 4e; 4d; 2c;		EDC2
//	2: ceg;;		[CEG][CEG]
//	bar

// (header)
//	title "titre du morceau"	(shortcut)

// -contexts-
// header/header2
// footer/footer2
// top/top2
// bottom/bottom2
// block
//	title [fontfamily weight/style (size)] "text1" "text2" "text3"
//			text1 = centered
//			text1 text2 = left and right
//			text1 text2 text3 = left, center and right
//	print [location] "text"		- cf location
//	left [location] "text"
//	right [location] "text"
//	center [location] "text"
//	[justified | ragged] paragraph [fontfam w/s (size)] "text
//				sur plusieurs lignes"	- default: justified
//							or same as previous parag.
//			- peut aller dans music (mais dans block!)
// newscore
// score
//		si absent, clef de sol, pas de #/&, 4/4
//	key=n '#'/'&'		nb de # ou &
//	key=d major
//	time=6/8		meter
//	staffs=n		nb de portees
//	clef=treble/bass
//	bracket/brace=2-3	sur staves 2 et 3
// staff S
//		 staff numeros	ex: staff 1,3,5-8
//	clef=
//
// voice S V
// grid2
// headshapes
// symbol
// accidentals
// keymap
// music
//	s v: en debut de ligne s = staff number (optionel),
//		v = no voix dans la portee (1,2 ou 3 seulement - si omis: voice 1)
//			1: top voice, 2: bottom voice, 3: middle voice
//		s et v peuvent etre range (comme 1,2,4-6)
//		et aussi 1 1 & 3 2 => musique va ds s1/v1 et aussi dans s3/v2
//		( -> ABC V:n avec n = (s - 1) * 3 + (v - 1)
//	notes: a..g	 les notes/chords sont séparées pas ';'
//			';;' = repetition note precedente
//			';n;' = repe avec nlle duree
//	accidents # & x && n
//	rest: r, space: s (prend temps, mais pas d'espace)
//		uncollapsible space u (prend temps et espace)
//		avec duree 'm' = whole measure
//		'm rpt' measure repeat
//	duree: n avant la note/rest/chord
//		 = whole / n '4' = noire, '1/2' = breve - '.' derriere
//		si absent = duree note precedente
//	octave: n apres la note c4 = ABC C - + octave au dessus, - en dessous
//		la clef definit l'octave par defaut treble/alto=4, bass=3
//	chgt clef: <<score clef=bass>>
//
//	lyrics 1: d;d;d; "pa-ro-les";	d;d;d est optionel
//	boldital below 1: 3 "mf";	=> ecrire "mf" sous la portee en bold/italic
//					au 3eme beat
//	rom	?
//	bar
//		bar, dblbar, dashed bar, dotted bar,
//		repeatstart, repeatboth, repeatend, endbar
//		(dblbar) repeatbar => ||:
//		invisbar
//		restart			= %%staffbreak
//	    options
//		pad stepsize
//		ending "txt"
//		endending
//		reh / rehearsal
//		hidechanges		= supprime key/meter en fin ligne precedente
//	multirest n

    // contexts
    var	ctx_f = {
	accidentals: function() {
	}, // accidentals()
	block: function(n) {
		if (!n)
			block_parse()
	}, // block()
	bottom: function(n) {
	}, // bottom()
	bottom2: function(n) {
	}, // bottom2()
	footer: function(n) {
	}, // footer()
	footer2: function(n) {
	}, // footer2 ()
	grid2: function(n) {
	}, // grid2()
	header: function(n) {
	}, // header()
	header2: function(n) {
	}, // header2()
	headshapes: function(n) {
	}, // headshapes()
	keymap: function(n) {
	}, // keymap()
	music: function(n) {
		if (!n)
			music_parse()
	}, // music()
	newscore: function(n) {
	}, // newscore()
	score: function(n) {
	}, // score()
	staff: function(n) {
	}, // staff()
	symbol: function(n) {
	}, // symbol()
	top: function(n) {
	}, // top()
	top2: function(n) {
	}, // top2()
	voice: function(n) {
	} // voice()
    } // ctx_f{}

	function parse_mup() {
	    var	f, r, p,
		ctx = 'music'			// default context

		// the tune may directly start with a measure
		bol = parse.eol
//		spsk()
//		if (mb[bol] >= '1' && mb[bol <= '9') {
//			measure_parse()
//			if (bol < 0)
//				return
//		}

		// loop on the contexts
		while (1) {
			while (1) {
				spsk()
				if (mb[bol] == '\n')
					bol++
				spsk()
				if (mb[bol] != '/' || mb[bol + 1] != '/')
					break
				bol = mb[bol].indexOf('\n', bol)
				if (bol < 0)
					return
			}
			if
			w.lastIndex = bol
			r = w.exec(mb)
			if (r) {
				bol = w.lastIndex
				r = r[0]
				if (ctx_f[r]) {
					ctx = r
					f = ctx_f[r]
					spsk()
					if (mb[bol] == '\n')
						bol++
					f(0)
				}
				// else same context
			}
			while (1) {
				set_parm(ctx)
				spsk()
				if (mb[bol] != ';')
					break
			}
			f(1)
			if (mb[bol] == '\n')
				bol++
			if (bol < 0 || !mb[bol])
				break
		}
	} // parse_mup

	// ------ set_fmt ------

	in_tune = parse.state
	if (parse.state < 3) {			// new tune
		mus.info.X = "1"
		parse.state = 1
		mus.get_key('C')		// start the tune body
		cfmt = mus.cfmt()
	}

	parse_mup()				// parse the mup notation

	// generate the tune if pure mup
	if (!in_tune) {
		mus.generate()
		mus.init_tune()
	}

	parse.eol = bol				// continue ABC scan
    } // set_fmt()

	mus.set_format = set_fmt.bind(mus, mus.set_format)
} // mup

if (!tm.mhooks)
	tm.mhooks = {}
tm.mhooks.mup = tm.mup
