// page.js - module to generate pages
//
// Copyright (C) 2018-2024 Jean-Francois Moine
//
// This file is part of txtmus.
//
// txtmus is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// txtmus is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with txtmus.  If not, see <http://www.gnu.org/licenses/>.
//
// This module is loaded when "%%pageheight" appears in a music source.
//
// Parameters
//	%%pageheight <unit>
//	%%footerfont font_name size (default: 'serif 16')
//	%%headerfont font_name size (default: 'serif 16')
//	%%topmargin <unit>
//	%%botmargin <unit>
//	%%gutter <unit>
//	%%oneperpage <bool>
//	%%dateformat <string> (see strftime)

if (typeof tm == "undefined")
	tm = {}

tm.pageheight = function (mus) {

    var	user_out

    // function called at end of generation
    function mus_end(of) {
	if (this.cfmt().pageheight) {
	    var page = mus.page
		if (page && page.in_page)
			close_page(page)
		if (user_out) {
			user.img_out = user_out
			user_out = null
		}
	}
	of()
    } // mus_end()

    // output the SVG tag
    function svg_tag(w, h, ty) {
	w = Math.ceil(w)
	h = Math.ceil(h)
	user_out(
		'<svg xmlns="http://www.w3.org/2000/svg" version="1.1"\n\
 class="'
		+ ty + '" '
		+ (user.imagesize != undefined
			? user.imagesize
			: ('width="' + w + 'px" height="' + h + 'px"')
		)
		+ ' viewBox="0 0 ' + w + ' ' + h + '">'
	)
    } // svg_tag()

    // generate a header or a footer in page.hf and return its height
    function gen_hf(page, ty) {
    var	a, i, j, k, x, y, y0, s, str,
	font = mus.get_font(ty.substr(0, 6)),
	cfmt = mus.cfmt(),
	fh = font.size * 1.1,
	pos = [ '">',
		'" text-anchor="middle">',
		'" text-anchor="end">' ]

	// replace <>& by XML character references
	function clean_txt(txt) {
		return txt.replace(/<|>|&.*?;|&/g, function(c) {
			switch (c) {
			case '<': return "&lt;"
			case '>': return "&gt;"
			case '&': return "&amp;"
			}
			return c
		})
	} // clean_txt()

	// clear a field if $x said so
	function clr(str) {
		return str.indexOf('\u00ff') >= 0 ? '' : str
	} //clr()

	// create the text of a header or a footer
	function header_footer(o_font, str) {
	    var	c, d, i, k, t, n_font, s,
		c_font = o_font,
		nl = 1,
		j = 0,
		r = ["", "", ""]

//		if (str[0] == '"')
//			str = str.slice(1, -1)
		while (1) {
			i = str.indexOf('$', j)
			if (i < 0)
				break
			c = str[++i]
			s = '$' + c		// string to replace
			switch (c) {
			case 'd':
				if (!tm.get_mtime)
					break // cannot know the change time of the file
				d = tm.get_mtime(mus.parse.fname)
				// fall thru
			case 'D':
				if (c == 'D')
					d = new Date()
				if (cfmt.dateformat[0] == '"')
					cfmt.dateformat = cfmt.dateformat.slice(1, -1)
				d = strftime(cfmt.dateformat, d)
				break
			case 'F':
				d = typeof document != "undefined"
					? window.location.href
					: page.abc.parse.fname
				break
			case 'I':
				c = str[++i]
				s += c
				// fall thru
			case 'T':
				t = mus.info()[c]
				if (t)
					t = t[0]
				d = t ? t.t : ''
				break
			case 'P':			// current page number
			case 'Q':			// absolute page number
				t = c == 'P' ? page.pn : page.pna
				switch (str[i + 1]) {
				case '0':
					s += '0'
					d = (t & 1) ? '\u00ff' : t
					break
				case '1':
					s += '1'
					d = (t & 1) ? t : '\u00ff'
					break
				default:
					d = t
					break
				}
				break
			case 'V':
				d = "txtmus-" + tm.version
				break
			default:
				d = ''
				if (c == '0')
					n_font = o_font
				else if (c >= '1' && c < '9')
					n_font = mus.get_font("u" + c)
				else
					break

				// handle the font changes
				if (n_font == c_font)
					break
				if (c_font != o_font)
					d += "</tspan>"
				c_font = n_font
				if (c_font == o_font)
					break
				d += '<tspan class="' +
					font_class(n_font) + '">'
				break
			}
			str = str.replace(s, d)
			j = i
		}
		if (c_font != o_font)
			str += "</tspan>"

		str = str.split('\n')
		r[4] = str.length		// number of lines
		for (j = 0; j < str.length; j++) {
			if (j != 0)
				for (i = 0; i < 3; i++)
					r[i] += '\n'
			t = str[j].split('\t')
			if (t.length == 1) {
				r[1] += clr(t[0])
			} else {
				for (i = 0; i < 3; i++) {
					if (t[i])
						r[i] += clr(t[i])
				}
			}
		}
		return r
	} // header_footer()

	function font_class(font) {
		if (font.class)
			return 'f' + font.fx + cfmt.fullsvg + ' ' + font.class
		return 'f' + font.fx + cfmt.fullsvg
	} // font_class()

	// -- gen_hf --

	if (!(page.pn & 1))
		str = page[ty + '2'] || page[ty]
	else
		str = page[ty]

	if (str[0] == '-') {		// not on 1st page
		if (page.pn == 1)
			return 0
		str = str.slice(1)
	}

	a = header_footer(font, clean_txt(str))
	y0 = font.size * .8
	for (i = 0; i < 3; i++) {
		str = a[i]
		if (!str)
			continue
		if (i == 0)
			x = cfmt.leftmargin
		else if (i == 1)
			x = cfmt.pagewidth / 2
		else
			x = cfmt.pagewidth - cfmt.rightmargin
		y = y0
		k = 0
		while (1) {
			j = str.indexOf('\n', k)
			if (j >= 0)
				s = str.slice(k, j)
			else
				s = str.slice(k)
			if (s)
				page.hf += '<text class="' +
						font_class(font) +
						'" x="' + x.toFixed(1) +
						'" y="' + y.toFixed(1) +
						pos[i] +
						s + '</text>\n'
			if (j < 0)
				break
			k = j + 1
			y += fh
		}
	}
	return fh * a[4]
    } // gen_hf()

    // start a new page
    function open_page(page,
			ht) {	// spacing under the header
    var	h, l,
	cfmt = mus.cfmt(),
	sty = ['line-height:0']

	page.pn++
	page.pna++

	// start a new page
	if (page.first)
		page.first = 0 //false
	else
		sty.push("page-break-before:always")
	if (page.gutter)
		sty.push("margin-left:"
			+ ((page.pn & 1) ? page.gutter : -page.gutter).toFixed(1)
			+ "px")
	if (sty.length)
		user_out('<div style="' + sty.join(';') + '">')
	else
		user_out('<div>')
	page.in_page = 1 //true

	ht += page.topmargin
	page.hmax = cfmt.pageheight - page.botmargin - ht

	// define the header/footer
	page.hf = ''
	if (page.header) {
		l = mus.get_font_style().length
		if (!cfmt.headerfont)
			mus.param_set_font("headerfont", "serif 16")
		h = gen_hf(page, "header")
		if (!h && page.pn == 1 && page.header1)
			h = gen_hf(page, "header1")
		sty = mus.get_font_style().slice(l)		// new style(s)
		if (cfmt.fullsvg || sty != page.hsty) {
			page.hsty = sty
			sty = '<style>' + sty + '\n</style>\n'
		} else {
			sty = ''
		}
		svg_tag(cfmt.pagewidth, ht + h, "header")
		user_out(sty
			+ '<g transform="translate(0,'
				+ page.topmargin.toFixed(1) + ')">'
				+ page.hf + '</g>\n</svg>')
		page.hmax -= h
		page.hf = ''
	} else {
		svg_tag(cfmt.pagewidth, ht, "header")
		user_out('\n</svg>')
	}
	if (page.footer) {
		l = mus.get_font_style().length
		if (!cfmt.footerfont)
			mus.param_set_font("footerfont", "serif 16")
		page.fh = gen_hf(page, "footer")
		sty = mus.get_font_style().slice(l)		// new style(s)
		if (cfmt.fullsvg || sty != page.fsty) {
			page.fsty = sty
			page.ffsty = '<style>' + sty + '\n</style>\n'
		} else {
			page.ffsty = ''
		}
		page.hmax -= page.fh
	}

	page.h = 0
    } // open_page()

    function close_page(page) {
    var	h,
	cfmt = mus.cfmt()

	page.in_page = 0 //false
	if (page.footer) {
		h = page.hmax + page.fh - page.h
		svg_tag(cfmt.pagewidth, h, "footer")
		user_out(page.ffsty
			+ '<g transform="translate(0,'
				+ (h - page.fh).toFixed(1) + ')">'
			+ page.hf + '</g>\n</svg>')
	}
	user_out('</div>')
	page.h = 0
    } // close_page()

    // handle the output flow of the txtmus generator
    function img_in(p) {
    var h, ht, nh,
	page = mus.page,
	cfmt = mus.cfmt()

	// copy a block
	function blkcpy(page) {
		while (page.blk.length)
			user_out(page.blk.shift())
		page.blk = null			// direct output
	} // blkcpy()

	// img_in()
	switch (p.slice(0, 4)) {
	case "<div":				// block of new tune
		if (p.indexOf('newpage') > 0
		 || (page.oneperpage && mus.info().X)
		 || !page.h) {			// empty page
			if (page.in_page)
				close_page(page)
			open_page(page, 0)
		}
		page.blk = []			// in block
		page.hb = page.h		// keep the offset of the start of tune
		break
	case "<svg":				// SVG image
		ht = p.match(/viewBox="0 0 (\d+) (\d+)"/)
		h = +ht[2]			// height
		if (cfmt.pagewidth != +ht[1])
			h *= cfmt.pagewidth / +ht[1]
		while (h + page.h >= page.hmax) { // if (still) page overflow
			ht = page.blk ? 0 :
				mus.cfmt().topspace // tune continuation

			if (page.blk) {
				if (!page.hb) {	// overflow on the first page
					blkcpy(page)
					nh = 0
				} else {
					nh = page.h - page.hb
					page.h = page.hb
				}
			}
			close_page(page)
			open_page(page, ht)

			if (page.blk) {		// if inside a block
				blkcpy(page)	// output the beginning of the tune
				page.h = nh
			}
			if (h > page.hmax)
				break		// error
		}

		// if no overflow yet, keep the block
		if (page.blk)
			page.blk.push(p)
		else
			user_out(p)
		page.h += h
		break
	case "</di":				// end of block
		if (page.blk)
			blkcpy(page)
		break
//	default:
////fixme: %%beginml cannot be treated (no information about its height)
//		break
	}
    } // img_in()

    // handle the page related parameters
    function set_fmt(of, cmd, parm) {
    var	v,
	cfmt = mus.cfmt(),
	page = mus.page

	if (cmd == "pageheight") {
		v = mus.get_unit(parm)
		if (isNaN(v)) {
			mus.syntax(1, mus.errs.bad_val, '%%' + cmd)
			return
		}
		if (!user.img_out || !tm.mus_end)
			v = 0
		cfmt.pageheight = v
		if (!v) {
			if (user_out) {
				user.img_out = user_out
				user_out = null
			}
			delete mus.page
			return
		}

		// if first definition, install the hook
		if (!page || !user_out) {
			mus.page = page = {
				mus: mus,
				topmargin: 38,	// 1cm
				botmargin: 38,	// 1cm
//				gutter: 0,
				h: 0,		// current page height
				pn: 0,		// page number
				pna: 0,		// absolute page number
				first: 1 //true	// no skip to next page
			}

			// don't let the backend handle the header/footer
			if (cfmt.header) {
				page.header = cfmt.header
				cfmt.header = null
			}
			if (cfmt.footer) {
				page.footer = cfmt.footer
				cfmt.footer = null
			}
			if (cfmt.header1) {
				page.header1 = cfmt.header1
				cfmt.header1 = null
			}
			if (cfmt.header2) {
				page.header2 = cfmt.header2
				cfmt.header2 = null
			}
			if (cfmt.footer2) {
				page.footer2 = cfmt.footer2
				cfmt.footer2 = null
			}

			// get the previously defined page parameters
			if (cfmt.botmargin != undefined) {
				v = mus.get_unit(cfmt.botmargin)
				if (!isNaN(v))
					page.botmargin = v
			}
			if (cfmt.topmargin != undefined) {
				v = mus.get_unit(cfmt.topmargin)
				if (!isNaN(v))
					page.topmargin = v
			}
			if (cfmt.gutter != undefined) {
				v = mus.get_unit(cfmt.gutter)
				if (!isNaN(v))
					page.gutter = v
			}
			if (cfmt.oneperpage)
				page.oneperpage = mus.get_bool(cfmt.oneperpage)
			if (!cfmt.dateformat)
				cfmt.dateformat = "%b %e, %Y %H:%M"

			// set the hooks
			user_out = user.img_out
			user.img_out = img_in.bind(mus)
			tm.mus_end = mus_end.bind(mus, tm.mus_end)
		}
		return
	}
	if (page) {
		switch (cmd) {
		case "header":
		case "footer":
		case "header1":
		case "header2":
		case "footer2":
			if (parm[0] == '"')
				parm = parm.slice(1, -1)
			page[cmd] = parm
			return
		case "newpage":
			if (!parm)
				break
			v = +parm
			if (isNaN(v)) {
				mus.syntax(1, mus.errs.bad_val, '%%' + cmd)
				return
			}
			page.pn = v - 1
			return
		case "gutter":
		case "botmargin":
		case "topmargin":
			v = mus.get_unit(parm)
			if (isNaN(v)) {
				mus.syntax(1, mus.errs.bad_val, '%%' + cmd)
				return
			}
			page[cmd] = v
			return
		case "oneperpage":
			page[cmd] = mus.get_bool(parm)
			return
		}
	}
	of(cmd, parm)
    } // set_fmt()

	// set the hooks
	mus.set_format = set_fmt.bind(mus, mus.set_format)

	user.page_format = 1 //true		// do page formatting
	mus.set_pagef()
} // pageheight()

if (!tm.mhooks)
	tm.mhooks = {}
tm.mhooks.page = tm.pageheight
