// sth.js - module to set the stem heights
//
// Copyright (C) 2018-2024 Jean-Francois Moine
//
// This file is part of txtmus.
//
// txtmus is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// txtmus is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with txtmus.  If not, see <http://www.gnu.org/licenses/>.
//
// This module is loaded when "%%sth" appears in a ABC source.
//
// Parameters
//	%%sth h1 h2 h3 ...
// The values h1, h2, .. are applied to the following notes which
// have a stem and which are not inside a beam.
// The value may be '*' for keeping the original stem length.

if (typeof tm == "undefined")
	tm = {}

tm.sth = function(mus) {

// function called after beam calculation
    function recal_beam(bm, s) {
    var staff_tb = mus.get_staff_tb(),
	y = staff_tb[s.st].y,
	s2 = bm.s2,
	y2 = staff_tb[s2.st].y

	if (s.sth != undefined)
		s.ys = s.sth
	if (s2.sth != undefined)
		s2.ys = s2.sth
	bm.a = (s.ys + y - s2.ys - y2) / (s.xs - s2.xs)
	bm.b = s.ys - s.xs * bm.a + y
	while (1) {
		s.ys = bm.a * s.xs + bm.b - y
		if (s.stem > 0)
			s.ymx = s.ys + 2.5
		else
			s.ymn = s.ys - 2.5
		s = s.next
		if (s == s2)
			break
	}
    } // recal_beam()

// function called after the stem heights have been computed
    function set_sth() {
    var s, h, v, sth_a, p_voice,
	v_tb = mus.get_v_tb()

	for (v = 0; v < v_tb.length; v++) {
		p_voice = v_tb[v]
		if (p_voice.sth != null)	// if no stem length in this voice
			continue
		sth_a = []
		for (s = p_voice.sym; s; s = s.next) {
			if (s.sth) {
				sth_a = s.sth
				s.sth = null
			}
			if (sth_a.length == 0
			 || s.nflags <= -2 || s.stemless
			 || !(s.beam_st || s.beam_end))
				continue
			h = sth_a.shift()
			if (h == '*')
				continue	// no change
			if (h == '|') {		// skip to the next measure bar
				for (s = s.next; s; s = s.next) {
					if (s.bar_type)
						break
				}
				continue
			}
			h = Number(h)
			if (isNaN(h) || !h)
				continue	// fixme: error
			if (s.stem >= 0) {
				s.ys = s.y + h
				s.ymx = (s.ys + 2.5) | 0
			} else {
				s.ys = s.y - h
				s.ymn = (s.ys - 2.5) | 0
			}
			s.sth = s.ys
		}
	}
    } // set_sth()

    function calculate_beam(of, bm, s1) {
    var	done = of(bm, s1)

	if (done && bm.s2 && (s1.sth || bm.s2.sth))
		recal_beam(bm, s1)
	return done
    } // calculate_beam()

    function new_note(of, gr, sls) {
    var	C = tm.C,
	p_v,
	s = of(gr, sls)

	if (s) {
		p_v = s.p_v
		if (p_v.sth && s.typ == C.note) {
			s.sth = p_v.sth
			p_v.sth = null	// some stem widths in this voice
		}
	}
	return s
    } // new_note()

    function set_fmt(of, cmd, param) {
	if (cmd == "sth") {
	    var	curvoice = mus.get_curvoice()
		if (curvoice)
			curvoice.sth = param.split(/[ \t;-]+/)
		return
	}
	of(cmd, param)
    } // set_fmt()

    function set_stems(of) {
	of()
	set_sth()
    } // set_stems()

	// set the hooks
	mus.calculate_beam = calculate_beam.bind(mus, mus.calculate_beam)
	mus.new_note = new_note.bind(mus, mus.new_note)
	mus.set_format = set_fmt.bind(mus, mus.set_format)
	mus.set_stems = set_stems.bind(mus, mus.set_stems)
} // sth

if (!tm.mhooks)
	tm.mhooks = {}
tm.mhooks.sth = tm.sth
