// txtmus - strtab.js - tablature for string instruments
//
// Copyright (C) 2020-2024 Jean-Francois Moine
//
// This file is part of txtmus.
//
// txtmus is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// txtmus is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with txtmus.  If not, see <http://www.gnu.org/licenses/>.
//
// This module is loaded by %%strtab.
// The command %%strtab changes the display of the voice to a tablature.
// Syntax:
//	%%strtab <string list> [diafret]
// <string list> is the list of the strings as ABC notes
// diafret indicates the instrument has diatonic frets
// The fret may be forced by a decoration format
//	"!" digit "s!"
// where 'digit' is the string range in the string list (last string is '1')
//
// The command %%minfret permits to force upper frets.
// Syntax:
//	%%minfret [ <string number> : <fret number> ]*
// Each command replaces the previous value.
//
// This module accepts Willem Vree's tablature syntax:
//	https://wim.vree.org/svgParse/abc2xml.html#tab

if (typeof tm == "undefined")
	tm = {}

tm.strtab = function(mus) {

    // draw the tablature
    function draw_symbols(of, p_v) {
    var	s, m, not, stb, x, y, g,
	C = tm.C

	// draw the note heads
	function draw_heads(stb, s) {
	    var	m, not, x, y

		for (m = 0; m <= s.nhd; m++) {
			not = s.notes[m]
			if (not.nb < 0)
				continue
			x = s.x - 3
			if (not.nb >= 10)
				x -= 3
			y = 3 * not.pit
			mus.out_svg('<text class="bg' + mus.bgn +
				'" x="')
			mus.out_sxsy(x, '" y="', stb + y - 2.5)
			mus.out_svg('">' + not.nb + '</text>\n')
		}
	} // draw_heads()

	// draw the stems, flags and beams
	function draw_stems(stb, s) {
		if (!s.tabst)
			return			// don't draw any stem
	    var	s1, s2, nfl, l,
		y = stb + 3 * (s.notes[0].pit - 3)
				* s.p_v.staffscale,
		h = (11 + 3 * s.notes[0].pit) * s.p_v.staffscale

		mus.out_svg('<path class="sW" d="M')
		mus.out_sxsy(s.x, ' ', y)
		mus.out_svg('v' + h.toFixed(1) +'"/>\n')

		// draw the flag(s)
		if (s.nflags <= 0
		 || !s.beam_end)		// inside beam
			return
		y -= h
		if (s.beam_st) {		// no beam
			mus.out_svg('<text class="f'
				+ mus.get_font('music').fid
				+ '" transform="translate(')
			mus.out_sxsy(s.x, ',', y)
			mus.out_svg(') scale('
				+ (s.grace ? '.5,.5' : '.7,.7') + ')">'
				+ String.fromCharCode(0xe23f + 2 * s.nflags)
				+ '</text>\n')
			return
		}

		// draw the beam(s)
		s2 = s
		nfl = s.nflags
		while (1) {
			if (s.nflags > nfl)
				nfl = s.nflags
			if (s.beam_st)
				break
			s = s.prev
		}
		s1 = s
		l = (s2.x - s1.x).toFixed(1)
		mus.out_svg('<path d="M')
		mus.out_sxsy(s1.x, ' ', y)
		mus.out_svg('h' + l
				+ 'v-3h-' + l
				+ 'v3"/>\n')
		if (nfl == 1)
			return
//fixme: do more levels
		y += 5
		while (1) {
			while (s.nflags < 2) {
				s = s.next
				if (s == s2)
					break
			}
			if (s == s2)
				break
			s1 = s
			while (s.next.nflags >= 2) {
				s = s.next
				if (s == s2)
					break
			}
			l = (s.x - s1.x).toFixed(1)
			mus.out_svg('<path d="M')
			mus.out_sxsy(s1.x, ' ', y)
			mus.out_svg('h' + l
					+ 'v-3h-' + l
					+ 'v3"/>\n')
			if (s == s2)
				break
			s = s.next
			if (s == s2)
				break
		}
	} // draw_stems()

	if (!p_v.tab) {
		of(p_v)
		return
	}

	// define the 'bgx' filter if not done yet
	m = mus.cfmt().bgcolor || "white"
	if (mus.bgt != m) {
		if (!mus.bgn)
			mus.bgn = 1
		else
			mus.bgn++
		mus.bgt = m
		mus.defs_add('\
<filter x="-0.1" y="0.1" width="1.2" height=".8" id="bg' + mus.bgn + '">\n\
<feFlood flood-color="' + m + '"/>\n\
<feComposite in="SourceGraphic" operator="over"/>\n\
</filter>')
		mus.add_style('\n.bg' + mus.bgn + '{filter:url(#bg' + mus.bgn + ')}')
	}

	// adjust the symbol before generation
	for (s = p_v.sym; s; s = s.next) {
		switch (s.typ) {
		case C.key:
		case C.meter:
		case C.rest:
			s.invis = 1 //true
			break
		}
	}
	of(p_v)

	mus.glout()			// output the stems

	stb = mus.get_staff_tb()[p_v.st].y

	// draw the stems and beams
	mus.set_sscale(-1)		// no scale
	for (s = p_v.sym; s; s = s.next) {
		switch (s.typ) {
		case C.grace:
			for (g = s.grasym; g; g = g.next)
				draw_stems(stb, g)
			break
		case C.note:
			draw_stems(stb, s)
			break
		}
	}

	// draw the note heads
	mus.set_scale(p_v.sym)
	mus.out_svg('<g class="bn">\n')
	for (s = p_v.sym; s; s = s.next) {
		switch (s.typ) {
		case C.grace:
			for (g = s.grasym; g; g = g.next)
				draw_heads(stb, g)
			break
		case C.note:
			draw_heads(stb, s)
			break
		}
	}
	mus.out_svg('</g>\n')
    } // draw_symbols()

    // change the font size of the chord symbols
    function csan_bld(of, s) {
	if (s.p_v.tab) {
	    var	i, gch,
		fmt = mus.cfmt()

		for (i = 0; i < s.a_gch.length; i++) {
			gch = s.a_gch[i]
			if (gch.type != 'g')
				continue

			// create the smaller font if not done yet
			if (!fmt.cstabfont) {
			    var	f = gch.f

				mus.param_set_font("cstabfont",
						f.name + ' ' + (f.size / 1.6).toFixed(1))
			}
			gch.f = mus.get_font("cstab")
		}
	}
	of(s)
    } // set_csan()

    // set a format parameter
    function set_fmt(of, cmd, parm) {
	switch (cmd) {
	case "cstabfont":
		mus.param_set_font("cstabfont", parm)
		return
	case "strtab":
		if (!parm)
			return
	    var	p_v = mus.get_curvoice()
		if (!p_v) {
			mus.parse.tab = parm
			return
		}
		mus.set_v_param("clef", "tab")
		if (parm.indexOf("diafret") >= 0) {
			mus.set_v_param("diafret", 1/*true*/)
			parm = parm.replace(/\s*diafret\s*/, "")
		}
		mus.set_v_param("strings", parm)
		return
	case "minfret":
		mus.set_v_param("minfret", parm)
		return
	}
	of(cmd, parm)
    } // set_fmt()

    // change the notes when the global generation settings are done
    function set_stems(of) {
    var	p_v, i, m, n, strss, g,
	C = tm.C,
	v_tb = mus.get_v_tb(),
	s = mus.get_tsfirst(),		// first symbol
	strs = [],			// notes per staff - index = staff
	lstr = []			// lowest string per staff - index staff

	// set a string (pitch) and a fret number
	function set_pit(p_v, s, nt, i) {
	    var	st = s.st

		if (i >= 0) {
			nt.nb = (p_v.diafret ? nt.pit : nt.midi) - p_v.tab[i]
			if (p_v.diafret && nt.acc)
				n += '+'
			nt.pit = i * 2
		} else {
			nt.nb = -1
			nt.pit = 0
		}
		nt.acc = 0
		nt.invis = 1 //true
		if (!s.grace)
			strss[i] = s.time + s.dur
		if (s.dur <= C.BLEN / 2 && !s.stemless) {
			if (!lstr[st])
				lstr[st] = [ 10 ]
			if (lstr[st][0] > i) {
				lstr[st][0] = i
				lstr[st][1] = s
			}
			s.stemless = 1 //true
		}
	} // set_pit()

	function set_notes(p_v, s) {
	    var	i, bi, bn, nt, m, n, ns

		s.stem = -1			// down stems

		// handle the fret numbers as chord decoration
		if (!s.nhd && s.a_dd) {
			i = s.a_dd.length
			while (--i >= 0) {
				bi = strnum(s.a_dd[i].name)
				if (bi >= 0) {
					nt = s.notes[0]
					set_pit(p_v, s, nt, bi)
					break
				}
			}
		}
		delete s.a_dd

		if (s.sls) {			// set the slurs above the staff
			for (i = 0; i < s.sls.length; i++) {
				s.sls[i].ty &= ~0x07
				s.sls[i].ty |= C.SL_ABOVE
			}
		}

	ls:	for (m = 0; m <= s.nhd; m++) {
			nt = s.notes[m]
			if (nt.sls) {			// if start of slur
				for (i = 0; i < nt.sls.length; i++) {
					nt.sls[i].ty &= ~0x07
					nt.sls[i].ty |= C.SL_ABOVE
				}
			}
			if (nt.nb != undefined)
				continue
			if (nt.a_dd) {
				i = nt.a_dd.length
				while (--i >= 0) {
					bi = strnum(nt.a_dd[i].name)
					if (bi >= 0) {
						set_pit(p_v, s, nt, bi)
						delete nt.a_dd
						continue ls
					}
				}
				delete nt.a_dd
			}

			// search the best string
			bn = 100
			bi = -1
			ns = i = p_v.tab.length
			while (--i >= 0) {
				if (strss[i] && strss[i] > s.time)
					continue
				n = (p_v.diafret ? nt.pit : nt.midi) - p_v.tab[i]
				if (n >= 0 && n < bn
				 && (!p_v.minfret
				  || !p_v.minfret[ns - i]
				  || n >= p_v.minfret[ns - i])) {
					bi = i
					bn = n
				}
			}
			set_pit(p_v, s, nt, bi)
		}

		s.y = 3 * nt.pit
		s.ymn = 0		// don't get space below the tablature
	} // set_notes()

	// get the string number from the decoration
	// format is either !<n>s! or !<n>!
	function strnum(n) {
		n = n.match(/^([1-9])s?$/)
		return n ? p_v.tab.length - n[1] : -1
	} // strnum()

	// change the notes of the strings when a capo
	for (n = 0; n < v_tb.length; n++) {
		if (!v_tb[n].tab)
			continue
		m = v_tb[n].capo
		if (m) {
			for (i = 0; i < v_tb[n].tab.length; i++)
				v_tb[n].tab[i] += m
		}
	}

	of()				// do the normal work

	// set the pitches
	for ( ; s; s = s.ts_next) {

		// let a stem on the lowest strings
		if (s.seqst || (s.ts_prev && s.ts_prev.typ == C.grace)) {
			for (i = 0; i < lstr.length; i++) {
				if (lstr[i]) {
					lstr[i][1].tabst = 1
					lstr[i] = null
				}
			}
		}

		p_v = s.p_v
		if (!p_v.tab)
			continue
		strss = strs[s.st]
		if (!strss)
			strss = strs[s.st] = []
		switch (s.typ) {
		case C.key:
		case C.rest:
		case C.time:
			s.invis = 1 //true
		default:
			delete s.a_dd
			break
		case C.grace:
			if (p_v.pos.gst == C.SL_HIDDEN)
				s.sappo = 0
			for (g = s.grasym; g; g = g.next) {
				if (p_v.pos.gst == C.SL_HIDDEN)
					g.stemless = 1 //true
				set_notes(p_v, g)
			}
			break
		case C.note:
			set_notes(p_v, s)
			break
		}
	}
    } // set_stems()

    // get the parameters of the current voice
    function set_vp(of, a) {
    var	e, g, tab, strs, ok,
	p_v = mus.get_curvoice(),
	i = 0

	// convert a list of ABC notes into a list of MIDI pitches
	function not2tab(p) {
	    var	i, c, a,
		t = []

		if (p_v.diafret) {
		    for (i = 0; i < p.length; i++) {
			c = p[i]
			c = tm.C.ntb.indexOf(c)
			if (c < 0)
				return // null
			c += 16
			while (1) {
				if (p[i + 1] == "'") {
					c += 7
					i++
				} else if (p[i + 1] == ",") {
					c -= 7
					i++
				} else {
					break
				}
			}
			t.push(c)
		    }
		} else {
		    for (i = 0; i < p.length; i++) {
			c = p[i]
			switch (c) {
			case '^':
			case '_':
				a = c == '^' ? 1 : -1
				c = p[++i]
				break
			default:
				a = 0
				break
			}
			c = "CCDDEFFGGAABccddeffggaab".indexOf(c)
			if (c < 0)
				return // null
			c += 60 + a
			while (1) {
				if (p[i + 1] == "'") {
					c += 12
					i++
				} else if (p[i + 1] == ",") {
					c -= 12
					i++
				} else {
					break
				}
			}
			t.push(c)
		    }
		}
		return t
	} // not2tab()

	// convert an array of <note name><octave> into a list of MIDI pitches
	function str2tab(a) {
	    var	str, p, o,
		t = []

		if (p_v.diafret) {
		    while (1) {
			str = a.shift()
			if (!str)
				break
			p = "CDEFGAB".indexOf(str[0])
			o = +str[1]
			if (p < 0 || isNaN(o))
				return // null
			t.push(o * 7 + p - 12)	// C4 = 16 (12 = 4 * 7 - 16)
		    }
		} else {
		    while (1) {
			str = a.shift()
			if (!str)
				break
			p = "CCDDEFFGGAAB".indexOf(str[0])
			if (p < 0)
				return // undefined
			o = str[1]
			switch (o) {
			case '#':
			case 'b':
				p += o == '#' ? 1 : -1
				o = +str[2]
				break
			default:
				o = Number(o)
				break
			}
			if (isNaN(o))
				return // undefined
			t.push((o + 1) * 12 + p)	// C4 = 60
		    }
		}
		return t
	} // str2tab()

	// convert the list of <string number> ':' <fret number>
	function minfret(a) {
	    var	sf,
		sfa = a.split(' ')

		p_v.minfret = {}
		while (1) {
			sf = sfa.shift()
			if (!sf)
				break
			sf = sf.split(':')
			if (sf.length != 2)
				break		//fixme: error
			p_v.minfret[sf[0]] = sf[1]
		}
	} //minfret()
 
	while (1) {
		if (!a[i])
			break
		switch (a[i++]) {
		case "clef":
			if (a[i] != '=')
				continue
			i++
//			e = a[i]
//			if (e != "tab")
//				break
//			a.splice(i, 1)
			if (a[i++] != 'tab')
				continue
			// fall thru
		case "tab":
//			a.splice(i, 1)
//			i--
			ok = 1 //true
			break
		case "strings":
			if (a[i] != '=')
				continue
			strs = a[++i]
			ok = 1 //true
			break
		case "nostems":
			p_v.pos.stm = p_v.pos.gst = tm.C.SL_HIDDEN
			break
		case "capo":
			if (a[i] != '=')
				continue
			i++
			p_v.capo = +a[i++]
			break
		case "diafret":
			if (a[i] == '=')
				i += 2
//		case "diafret":
			p_v.diafret = 1 //true
			break
		case "minfret":
			if (a[i] != '=')
				continue
			i++
			minfret(a[i++])
 			break
		}
	}

	// define the elements of the tablature
	if (ok) {
		if (!strs && mus.parse.tab) {		// if a global definition
			strs = mus.parse.tab
			if (strs.indexOf("diafret") >= 0) {
				p_v.diafret = 1 //true
				strs = strs.replace(/\s*diafret\s*/, "")
			}
		}
		if (strs) {
			e = strs.slice(-1)
			if (e >= '1' && e <= '9')
				tab = str2tab(strs.split(','))	// W.V.'s syntax
			else
				tab = not2tab(strs)		// ABC syntax
			if (!tab) {
				mus.syntax(1, "Bad strings in tablature")
				ok = 0 //false
			}
		} else if (!p_v.tab) {
			tab = p_v.diafret ?
				[10, 14, 17] :		// dulcimer
				[40, 45, 50, 55, 59, 64] // guitar strings
		} else {
			tab = p_v.tab
		}
	}
	if (ok) {
		if (p_v.capo) {
			p_v.tab = []
			for (i = 0; i < tab.length; i++)
				p_v.tab.push(tab[i] + p_v.capo)
		} else {
			p_v.tab = tab
		}

		a.push("clef")				// set the clef
		a.push("=")
		g = mus.glyphs
		if (tab.length == 3) {
			a.push('"tab3"')
			if (!g.tab3)
// SMuFL: -none-
				g.tab3 = '<text id="tab3"\
 x="-2,-2,-2" y="-4,3,10"\
 style="font:bold 8px sans-serif">TAB</text>'
		} else if (tab.length == 4) {
			a.push('"tab4"')
			if (!g.tab4)
// SMuFL: \ue06e
				g.tab4 = '<text id="tab4"\
 x="-3,-3,-3" y="-8,1,10"\
 style="font:bold 12px sans-serif">TAB</text>'
		} else if (tab.length == 5) {
			a.push('"tab5"')
			if (!g.tab5)
// SMuFL: -none-
				g.tab5 = '<text id="tab5"\
 x="-4,-4,-4" y="-11,-2,7"\
 style="font:bold 12px sans-serif">TAB</text>'
		} else {
			a.push('"tab6"')
			if (!g.tab6)
// SMuFL: \ue06d
				g.tab6 = '<text id="tab6"\
 x="-4,-4,-4" y="-14.5,-4,5.5"\
 style="font:bold 13px sans-serif">TAB</text>'
		}
		a.push("stafflines")
		a.push("=")
		a.push("|||||||||".slice(0, tab.length))
		p_v.staffscale = 1.6
//		p_v.scale = .6
//		p_v.straightflags = 1 //true
	}
	of(a)
    } // set_vp()

	// set the hooks
	mus.draw_symbols = draw_symbols.bind(mus, mus.draw_symbols)
	mus.gch_build = csan_bld.bind(mus, mus.gch_build)
	mus.set_format = set_fmt.bind(mus, mus.set_format)
	mus.set_stems = set_stems.bind(mus, mus.set_stems)
	mus.set_vp = set_vp.bind(mus, mus.set_vp)

	// define specific decorations used to force the string number
    var	decos = mus.decos
	decos["1s"] = "0 nil 0 0 0"
	decos["2s"] = "0 nil 0 0 0"
	decos["3s"] = "0 nil 0 0 0"
	decos["4s"] = "0 nil 0 0 0"
	decos["5s"] = "0 nil 0 0 0"
	decos["6s"] = "0 nil 0 0 0"

	mus.add_style("\n.bn{font:bold 7px sans-serif}")
} // strtab

if (!tm.mhooks)
	tm.mhooks = {}
tm.mhooks.strtab = tm.strtab
