// txtmus - tomnx.js - convert text music to MNX
//
// MNX documentation: https://w3c.github.io/mnx/docs/
//
// Copyright (C) 2021-2024 Jean-Francois Moine
//
// This file is part of txtmus.
//
// txtmus is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// txtmus is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with txtmus.  If not, see <http://www.gnu.org/licenses/>.

// constants from core/tm.js
    var	C = tm.C,

	OPEN_BRACE = 0x01,
	CLOSE_BRACE = 0x02,
	OPEN_BRACKET = 0x04,
	CLOSE_BRACKET = 0x08,
	OPEN_PARENTH = 0x10,
	CLOSE_PARENTH = 0x20,
	STOP_BAR = 0x40,
	FL_VOICE = 0x80,
	OPEN_BRACE2 = 0x0100,
	CLOSE_BRACE2 = 0x0200,
	OPEN_BRACKET2 = 0x0400,
	CLOSE_BRACKET2 = 0x0800,
	MASTER_VOICE = 0x1000

user.img_out = function() {}	// do the generation with no output

	// global variables/constants
    var	id = 0,			// next available identifier
	bar_ty = {
		"||": "light-light",
		"|]": "light-heavy",
		"": "none"
	}

// dump the global element
function global_dump() {
    var	s, meas, val, volt,
	bartim = 0,
	tsfirst = this.get_tsfirst(),
	voice_tb = this.voice_tb

	// decorations
	function deco(s) {
	    var	i, dd

		for (i = 0; i < s.a_dd.length; i++) {
			dd = s.a_dd[i]
			switch (dd.name) {
			case "D.S.alfine":
				if (s.time == bartim)
					continue
				meas += '\t<jump type="dsalfine" location="'
				break
			case "fine":
				if (s.time == bartim)
					continue
				meas += '\t<fine location="'
				break
			case "segno":
				if (s.time != bartim)	// at start of measure
					continue
				meas += '\t<segno location="'
				break
			default:
				continue
			}
			meas += ((s.time - bartim) / C.BLEN) + '"/>\n'
		}
	} // deco()

	// -- main code of global_dump()
	tm.print('\
<mnx>\n\
  <global>')

	// for the 1st measure
	meas = ''
	if (voice_tb[0].key.k_sf)
		meas = '\t<key fifths="' + voice_tb[0].key.k_sf + '"/>\n'

	s = voice_tb[0].meter
	if (s.wmeasure != 1) {
		val = s.a_meter[0].top + '/' + s.a_meter[0].bot
		if (s.meas_dur)
			val += '" measure="' + s.meas_dur
		meas += '\t<time signature="' + val + '"/>\n'
	}

	for (s = tsfirst; s; s = s.ts_next) {
		if (s.a_dd)
			deco(s)

		switch (s.typ) {
		case C.bar:
//fixme: MNX: what/how if a bar inside a measure?
			if (s.bar_type[0] == ':')
				meas += '\t<repeat type="end"/>\n'
			if (volt && s.rbstop) {
				volt = 0
				if (s.rbstop == 2)
					meas += '\t<ending type="stop"/>\n'
				else
					meas += '\t<ending type="discontinue"/>\n'
			}

			if (s.time) {		// if not at start of tune
				val = ''
				if (!s.next) {
					switch (s.bar_type) {
					case '|':
						val = "regular"
						break
					default:
						val = bar_ty[s.bar_type]
					case '|]':
						break
					}
				} else {
					val = bar_ty[s.bar_type] // barline
				}
				if (s.bar_dotted)
					val = 'dashed'
//--fixme: 'index' and 'number' are missing in 'measure-global'
				if (!val) {
					if (meas)
						tm.print('\
    <measure-global>\n\
      <directions-global>\n'
							+ meas + '\
      </directions-global>\n\
    </measure-global>')
					else
						tm.print('\
    <measure-global/>')
				} else {
					val = ' barline="' + val + '"'
					if (meas)
						tm.print('\
    <measure-global' + val + '>\n\
      <directions-global>\n'
							+ meas + '\
      </directions-global>\n\
    </measure-global>')
					else
						tm.print('\
    <measure-global' + val + '/>')
				}
			
				meas = ''
			}

			if (s.bar_type.slice(-1) == ':')
				meas += '\t<repeat type="start"/>\n'
			if (s.text) {
				meas += '\t<ending type="start" number="'
					+ s.text + '"/>\n'
				volt = 1
			}
			bartim = s.time
			if (s.a_dd)
				deco(s)
			while (s.ts_next && s.ts_next.typ == C.bar)
				s = s.ts_next
			break
		case C.key:
			meas += '\t<key fifths="'
				+ s.k_sf + '"/>\n'
			while (s.ts_next && s.ts_next.typ == C.key)
				s = s.ts_next
			break
		case C.meter:
			val = s.a_meter[0].top + '/' + s.a_meter[0].bot
			if (s.meas_dur)
				val += '" measure="' + s.meas_dur
			meas += '\t<time signature="' + val + '"/>\n'
			while (s.ts_next && s.ts_next.typ == C.meter)
				s = s.ts_next
			break
		case C.tempo:
			if (!s.invis
			 && s.tempo_notes)
				meas += '\t<tempo bpm="' + s.tempo
					+ '" value="' + dur_dump(s.tempo_notes[0])
					+ '">\n'
			while (s.ts_next && s.ts_next.typ == C.tempo)
				s = s.ts_next
			break
		}
	}
	tm.print('\
  </global>')
} // global_dump()

function beam_dump(s) {
    var s2, ln, bm, nbm, ind

	function b1_dump(s2, n, ind) {
	    var	ln, i, s_pr,
		s = s2

		while (s) {
//			if (s.beam_end)
//				break
			if (!s.dur || s.nflags < n) {
				s = s.next
				continue
			}
			ln = ''
			i = 0
			while (1) {
			    if (s.dur) {
				if (ln)
					ln += ' '
				ln += 'id' + s.id
				i++
				if (s.beam_end)
					break
				s_pr = s	// previous note/rest
				s = s.next
				if (!s || s.nflags < n
				 || s.beam_br1
				 || (s.beam_br2 && n > 2))
					break
			    }
			}
			if (i == 1) {		// if beam stub (hook)
				if (s.nflags < n)
				tm.print(ind + '\
<beam-hook direction="right" event="id' + s_pr.id + '"/>')
				else
				tm.print(ind + '\
<beam-hook direction="left" event="id' + s.id + '"/>')
			} else if (n == nbm) {
				tm.print(ind + '\
<beam events="' + ln + '"/>')
			} else {
				tm.print(ind + '\
<beam events="' + ln + '">')
				b1_dump(s2, n + 1, ind + '  ')
				tm.print(ind + '\
</beam>')
			}
			if (s.beam_end)
				break
		}
	} // b1_dump()

	while (s) {
		if (s.beam_st && !s.beam_end) {
			if (!bm) {
				bm = 1
				tm.print('\t<beams>')
			}
			ind = '\t  '
			ln = ''
			s2 = s
			nbm = 1
			while (s) {
			    if (s.dur) {
				if (s.nflags > nbm)
					nbm = s.nflags
				if (!s.id)
					s.id = ++id
				if (ln)
					ln += ' '
				ln += 'id' + s.id
				if (s.beam_end)
					break
			    }
				s = s.next
			}
			if (nbm == 1) {
				tm.print(ind + '\
<beam events="' + ln + '"/>')
			} else {
				tm.print(ind + '\
<beam events="' + ln + '">')
				b1_dump(s2, 2, ind + '  ')
				tm.print(ind + '\
</beam>')
			}
		}
		s = s.next
	}
	if (bm)
		tm.print('\t</beams>')
} // beam_dump()

// dump a clef
function clef_dump(s, ind) {
    var	o,
	ln = ind + '<clef sign="'

	switch (s.clef_type) {
	case 't': ln += 'G'; break
	case 'c': ln += 'C'; break
	case 'b': ln += 'F'; break
	case 'p': ln += 'percussion'; break
//fixme: TAB
	default:  return ''
	}
	if (s.clef_octave) {
		if (s.clef_octave > 0)
			o = s.clef_octave > 0
				? (s.clef_octave + 1)
				: (s.clef_octave - 1)
		ln += '" octave="' + o
	}
	ln += '" line="' + s.clef_line + '"/>\n'
	return ln
} // clef_dump()

// return a measure location
// @pr: if set, return the previous event
function loc(s, f, pr) {
    var	s2, s3,
	ln = ''

	// search the location
	for (s2 = s.next; s2; s2 = s2.next) {
		if (f(s2))
			break
		if (s2.bar_type)
			s3 = s2
	}
	if (!s2)
		return "Error: no end of decoration"
	if (s3) {
		ln += s3.bar_num.toString() + ':'	// further measure
	} else {
		for (s3 = s; s3.prev; s3 = s3.prev) {	// in this measure
			if (s3.bar_type)
				break
		}
	}
	if (pr)
		s2 = s2.prev
	return ln + ((s2.time - s3.time) / C.BLEN).toFixed(2)
} // location()

// dump the decorations
function dir_dump(s) {
    var	i, dd, val, s2, s3,
	ind = '\t  ',
	ln = ''

	if (s.typ == C.clef)
		ln += clef_dump(s, ind)

    if (s.a_dd) {
	for (i = 0; i < s.a_dd.length; i++) {
		dd = s.a_dd[i]
		switch (dd.name) {
		case "8va(":
		case "8vb(":
		case "15ma(":
		case "15mb(":
			ln += ind + '<octave-shift type="'
			if (dd.name[0] == '8')
				ln += dd.name.slice(-2) == 'a('
					? '8' : '-8'
			else
				ln += dd.name.slice(-2) == 'a('
					? '15' : '-15'

			ln += '" end="'
				+ loc(s, function(ss) {
						return ss.ottava != undefined
						}, 1)	// get the previous event
				+ '"/>\n'
			break
		case "p":
		case "pp":
		case "ppp":
		case "pppp":
		case "f":
		case "ff":
		case "fff":
		case "ffff":
		case "mf":
		case "mp":
		case "sfz":
			if (s.fmt.pos && s.fmt.pos.dyn)
				val = '" orient="'
					+ (((s.fmt.pos.dyn & 0x07) == C.SL_ABOVE)
						? 'up' : 'down')
			else
				val = ''
			ln += ind + '<dynamics type="'
				+ dd.name + val + '"/>\n'
			break
		case "crescendo(":
		case "<(":
		case "diminuendo(":
		case ">(":
			ln += ind + '<wedge type="'
				+ ((dd.name[0] == 'c' || dd.name[0] == '<')
					? "crescendo" : "diminuendo")
			if (s.pos.dyn)
				ln += '" orient="'
					+ (((s.pos.dyn & 0x07) == C.SL_ABOVE)
						? 'up' : 'down')
			val = dd.name.slice(0, -1) + ')'
			ln += '" end="'
				+ loc(s, function(ss) {
						if (!ss.a_dd)
							return
						for (var i = 0; i < ss.a_dd.length; i++) {
							if (ss.a_dd[i].name == val)
								return 1
						}
						return
						})
				+ '"/>\n'
			break
		}
	}
    } // s.a_dd

// <instruction text-font-size="2.5st">Allegro</instruction>
// (see https://github.com/w3c/mnx/issues/263)
// <expression>subito</expression>
// (see old spec)

	if (ln)
		tm.print('\t<directions>\n'
			+ ln
			+'\t<directions>')
} // deco_dump()

// dump a duration
function dur_dump(dur) {
    var	nf, d,			// nf = number of flags
	t = (dur / 12) | 0

	for (nf = 5; !(t & 1); t >>= 1, nf--)
		;
	if (nf == -4)
		d = '*4'
	else if (nf == -3)
		d = '*2'
	else
		d = '/' + (1 << (nf + 2)).toString()
	switch (t) {
	case 15: d += 'd'
	case 7: d += 'd'
	case 3: d += 'd'
		break
	}
	return d
} // dur_dump()

function note_dump(s, ind) {
    var	m, not,
	evid = s.id ? (' id="id' +s.id + '"') : ''

	tm.print(ind + '<event' + evid + ' value="'
		+ dur_dump(s.dur_orig)
		+ '">')
	if (s.sls || s.sl1)			// if start of slur
		slur_dump(s, ind)

	for (m = 0; m <= s.nhd; m++) {
		not = s.notes[m]
		evid = not.id ? (' id="id' +not.id + '"') : ''
		if (not.tie_e) {
			tm.print(ind + '  <note'
				+ evid
				+ pit_dump(not)
				+ '>')
			not = not.tie_e
			if (!not.id)
				not.id = ++id
			tm.print(ind + '    <tied target="id'
					+ not.id + '"/>\n'
				+ ind + '  </note>')
		} else {
			tm.print(ind + '  <note'
				+ evid
				+ pit_dump(not)
				+ '/>')
		}
	}
	tm.print(ind +'</event>')
} // note_dump()

function pit_dump(not) {
    var pit = (not.opit || not.pit) + 12,
	d = (not.midi % 12) - [0, 2, 4, 5, 7, 9, 11][pit % 7],
	ln = ' pitch="' + "CDEFGAB"[pit % 7]

	switch (d) {
	case -2: ln += 'bb'; break
	case -1: ln += 'b'; break
	case 1: ln += '#'; break
	case 2: ln += '##'; break
	case 3: ln += '='; break
	}
	ln += ((pit / 7) | 0).toString() + '"'

//fixme: add microtone value
	switch (not.acc) {
	case -2: ln += ' accidental="dbleflat"'; break
	case -1: ln += ' accidental="flat"'; break
	case 1: ln += ' accidental="sharp"'; break
	case 2: ln += ' accidental="dblesharp"'; break
	case 3: ln += ' accidental="natural"'; break
	}
	return ln
} // pit_dump()

function slur_dump(s, ind) {
    var	 i, m, sl, s2, not, not2

	function slur_out(ty, r1, r2) {
	    var	ln = ''

		if (ty & C.SL_DOTTED)
			ln += ' line-type="dotted"'
		switch (ty & 0x07) {
		case C.SL_ABOVE:
			ln += ' side="up"'
			break
		case C.SL_BELOW:
			ln += ' side="down"'
			break
		}
		return ln
	} // slur_out()

	if (s.sls) {				// start from a chord
		for (i = 0; i < s.sls.length; i++) {
			sl = s.sls[i]
			not2 = sl.note
			s2 = not2.s
			if (!s2.id)
				s2.id = ++id	// target event
			if (sl.is_note		// end on a note
			 && !not2.id)
				not2.id = ++id
			tm.print(ind + '  <slur'
				+ slur_out(sl.ty)
				+ (sl.is_note
					? (' end-note="id' + not2.id + '"')
					: '')
				+ ' target="id'
				+ s2.id
				+ '"/>')
		}
	}
	if (s.sl1) {				// from a note
		for (m = 0; m <= s.nhd; m++) {
			not = s.notes[m]
			if (!not.sls)
				continue
			if (!not.id)
				not.id = ++id
			for (i = 0; i < not.sls.length; i++) {
				sl = not.sls[i]
				not2 = sl.note
				s2 = not2.s
				if (!s2.id)
					s2.id = ++id	// target event
				if (sl.is_note		// end on a note
				 && !not2.id)
					not2.id = ++id
				tm.print(ind + '  <slur'
					+ slur_out(sl.ty)
					+ ' start-note="id' + not.id + '"'
					+ (sl.is_note
						? (' end-note="id' + not2.id + '"')
						: '')
					+ ' target="id'
					+ s2.id
					+ '"/>')
			}
		}
	}
} // slur_dump()

function system_dump(s, voice_tb) {
    var	sy, st, v, fl,
	ind = '  '

	while (1) {
		if (s.typ == C.staves)
			break
		s = s.next
		if (!s)
			return
	}

	tm.print(ind + '<system-layout id="sy1">')

	sy = s.sy
	ind += '  '
	for (st = 0; st <= sy.nstaff; st++) {
		fl = sy.staves[st].flags
//console.log('fl '+fl.toString(16))
		if (fl & (OPEN_BRACKET | OPEN_BRACE)) {
			tm.print(ind + '<group-layout symbol="'
				+ ((fl & OPEN_BRACKET) ? 'bracket' : "brace")
				+ '">')
			ind += '  '
		}
		if (fl & (OPEN_BRACKET2 | OPEN_BRACE2)) {
			tm.print(ind + '<group-layout symbol="'
				+ ((fl & OPEN_BRACKET2) ? 'bracket' : "brace")
				+ '">')
			ind += '  '
		}

		tm.print(ind + '<staff-layout>')
		ind += '  '
		for (v = 0; v < sy.voices.length; v++) {
			if (sy.voices[v].st == st)
				tm.print(ind + '<part-layout part="v'
					+ voice_tb[v].id
					+ '"/>')
		}
		ind = ind.slice(0, -2)
		tm.print(ind + '</staff-layout>')
		if (fl & (CLOSE_BRACKET2 | CLOSE_BRACE2)) {
			ind = ind.slice(0, -2)
			tm.print(ind + '</group-layout>')
		}
		if (fl & (CLOSE_BRACKET | CLOSE_BRACE)) {
			ind = ind.slice(0, -2)
			tm.print(ind + '</group-layout>')
		}
	}
	ind = ind.slice(0, -2)
	tm.print(ind + '</system-layout>')
} // system_dump()

// dump the (first) tune
function mns_gen(of) {
   var	s, s2, g, v, p_v, val, m, not,
	ind = '\t',			// default <event> indentation
	tsfirst = this.get_tsfirst(),
	voice_tb = this.voice_tb,
	staff_tb = this.get_staff_tb()

	global_dump.call(this)

	// loop on the voices
//fixme: what about instruments with many voices?
	for (v = 0; v < voice_tb.length; v++) {
		p_v = voice_tb[v]
		s = p_v.sym
		if (!s)
			continue
		val = ''
		if (voice_tb.length > 1)
			val += ' id="v' + p_v.id + '"'
		if (p_v.nm)
			val += ' name="' + p_v.nm + '"'
		if (p_v.snm)
			val += ' short-name="' + p_v.snm + '"'
		tm.print('  <part'
			+ val + '>\n\
    <measure>')
		val = ''
//fixme: only if many sequences? (- many voices in the group)
//		if (voice_tb.length > 1)
//			val = ' voice="' + p_v.id + '"'
		tm.print( '      <directions-part>\n'
			+ clef_dump(staff_tb[s.st].clef, '\t')
			+ '      </directions-part>\n\
      <sequence' + val + '>')

		// get the first beams
		for (s2 = s; s2; s2 = s2.next) {
			if (s2.bar_type)
				break
			if (s2.grasym)
				beam_dump(s2.grasym)
			if (s2.beam_st && !s2.beam_end) {
				beam_dump(s2)
				break
			}
		}

		for ( ; s; s = s.next) {

			// handle the tuplets
			if (s.tp) {
//fixme: should handle many levels...
				for (s2 = s; s2; s2 = s2.next) {
					if (s2.tpe)
						break
				}
				val = (s2.time + s2.dur - s.time) / s.tp[0].q
				m = ''
				if (s.tp[0].f[0] == 1) {	// when = never
					m += '\
 bracket="no" show-number="no" show-value="no"'
				} else {
					if (s.tp[0].f[1] == 1) // what = slur
						m += ' bracket="no"'
					if (s.tp[0].f[2] == 1) // which = none
						m += ' show-number="no"'
					else if (s.tp[0].f[2] == 2) // which = p:q
						m += ' show-number="both"'
					switch (+s.tp[0].f[3]) {
					case C.SL_ABOVE:
						m += ' orient="up"'
						break
					case C.SL_BELOW:
						m += ' orient="down"'
						break
					}
				}
				tm.print(ind
					+ '<tuplet inner="'
					+ s.tp[0].p
					+ dur_dump(val)
					+ '" outer="'
					+ s.tp[0].q
					+ dur_dump(val)
					+ m + '">')
				ind = ind + '  '
			}

			// dump the directions
			// (deco, annotations, chord symbols...)
			dir_dump(s)

			switch (s.typ) {
			case C.bar:
				if (!s.time || !s.next)
					break
				tm.print('\
      </sequence>\n\
    </measure>\n\
    <measure>\n\
      <sequence>')
				beam_dump(s)
				break
			case C.grace:
				tm.print(ind + '<grace>')
				for (g = s.grasym; g; g = g.next)
					note_dump(g, ind + '  ')
				tm.print('\t</grace>')
				break
			case C.note:
				note_dump(s, ind)
				break
			case C.rest:
			    var	evid = s.id ? (' id="id' +s.id + '"') : ''
				if (!s.invis) {
					tm.print(ind
						+ '<event'
						+ evid
						+ ' value="'
						+ dur_dump(s.dur_orig)
						+ '">\n'
						+ ind + '  <rest/>\n'
						+ ind + '</event>')
				} else {
					tm.print(ind
						+ '<forward duration="'
						+ dur_dump(s.dur_orig)
						+ '"\>')
				}
				break
//fixme: no multi-measure rests?
//			case C.staves:
//				break
			}
			if (s.tpe) {
				ind = ind.slice(-2)
				tm.print(ind + '</tuplet>')
			}
		}
		tm.print('\
      </sequence>\n\
    </measure>\n\
  </part>')
	}

	// system layout
	if (voice_tb.length > 1)
		system_dump(tsfirst, voice_tb)

	tm.print('</mnx>')
	tm.quit()
} // mns_gen()

// -- local functions
tm.mus_init = function(args) {
	mus.set_stems = mns_gen.bind(mus, mus.set_stems)
}

tm.mus_end = function() {}

tm.abort = function(e) {
	tm.printErr(e.message + "\n*** Abort ***\n" + e.stack)
	tm.quit()
}
