// sndgen.js - sound generation
//
// Copyright (C) 2019-2024 Jean-Francois Moine
//
// This file is part of txtmus.
//
// txtmus is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// txtmus is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with txtmus.  If not, see <http://www.gnu.org/licenses/>.

// This script generates the play data which are stored in the music symbols:
// - in all symbols
//	s.ptim = play time
// - in NOTE and REST
//	s.pdur = play duration

if (!tm)
    var	tm = {}			// txtmus structures

function ToAudio() {
 return {

   // generate the play data of a tune
   add: function(first,		// starting symbol
		voice_tb,	// voice table
		cfmt) {		// tune parameters
    var	toaud = this,
	C = tm.C,
	p_time = 0,		// last playing time
	abc_time = 0,		// last ABC time
	play_fac = C.BLEN / 4 * 120 / 60, // play time factor - default: Q:1/4=120
	i, n, dt, d, v,
	s = first,
	rst_fac,		// play factor on repeat restart
	b_tim,			// time of last measure bar
	b_typ			// type of last measure bar

	function get_beat() {
	    var	s = first.p_v.meter

		if (!s.a_meter[0])
			return C.BLEN / 4
		if (!s.a_meter[0].bot)
			return (s.a_meter[1]
				&& s.a_meter[1].top == '|')	// (cut time)
					? C.BLEN / 2 : C.BLEN / 4
		if (s.a_meter[0].bot == "8"
		 && s.a_meter[0].top % 3 == 0)
			return C.BLEN / 8 * 3
		return C.BLEN / s.a_meter[0].bot | 0
	} // get_beat()

	// create the starting beats
	function def_beats() {
	    var	i, s2, s3, tim,
		beat = get_beat(),		// time between two beats
		d = first.p_v.meter.wmeasure,	// duration of a measure
		nb = d / beat | 0,		// number of beats in a measure
		v = voice_tb.length,		// beat voice number
		p_v = {				// voice for the beats
			id: "_beats",
			v: v,
//			time:
			sym: {
				typ: C.block,
				v: v,
//				p_v: p_v,
				subtype: "midiprog",
				chn: 9,			// percussion channel
				instr: 16384,	// percussion bank
//				time:
//				next:
				ts_prev: first
//				ts_next: 
			}
//			vol:
		},
		s = {
			typ: C.note,
			v: v,
			p_v: p_v,
//			time:
			dur: beat,
			nhd: 0,
			notes: [{
				midi: 37	// Side Stick
			}]
		}

		abc_time = -d			// start time of the beat ticks

		// check for an anacrusis
		for (s2 = first; s2; s2 = s2.ts_next) {
			if (s2.bar_type && s2.time) {
				nb = (2 * d - s2.time) / beat | 0
				abc_time -= d - s2.time
				break
			}
		}

		// add the tempo
		s2 = p_v.sym			// midiprog
		for (s3 = first; s3 && !s3.time; s3 = s3.ts_next) {
			if (s3.typ == C.tempo) {
				s3 = Object.create(s3)	// new tempo
				s3.v = v
				s3.p_v = p_v
				s3.prev =
					s3.ts_prev = s2
				s2.next =
					s2.ts_next = s3
				s2 = s3
				play_fac = set_tempo(s2)
				break
			}
		}

		voice_tb[v] =
			p_v.sym.p_v = p_v
		first.time = s2.time = tim = abc_time
		if (s3)
			p_v.sym.time = tim
		for (i = 0; i < nb; i++) {
			s3 = Object.create(s)	// new beat tick
			s3.time = tim
			s3.prev =
				s3.ts_prev = s2
			s2.next =
				s2.ts_next = s3
			s2 = s3
			tim += beat
		}
		s2.ts_next = first.ts_next
		s2.ts_next.ts_prev = s2
		first.ts_next = p_v.sym
	} // def_beats()

	// build the information about the parts (P:)
	function build_parts(first) {
	    var	i, j, c, n, v,
		s = first,
		p = s.parts,
		st = [],
		r = ""

		// build a linear string of the parts
		for (i = 0; i < p.length; i++) {
			c = p[i]
			switch (c) {
			case '.':
				continue
			case '(':
				st.push(r.length)
				continue
			case ')':
				j = st.pop()
				if (j == undefined)
					j = r.length
				continue
			case '[':			// convert "A[23" into "2A3A"
				c = r[i - 1]
				r = r.slice(0, -1)
				while (1) {
					n = p[i + 1]
					if (!n || n < '1' || n > '9')
						break
					r += n + c
					i++
				}
				continue
			}
			if (c >= 'A' && c <= 'Z') {
				j = r.length
				r += c
				continue
			}
			n = +c
//fixme:one digit is enough!
//			while (1) {
//				c = p[i + 1]
//				if (c < '0' || c > '9')
//					break
//				n = n * 10 + +c
//				i++
//			}
			if (isNaN(n))
				break
			v = r.slice(j)
			if (r.length + v.length * n > 128)
				continue
			while (--n > 0)
				r += v
		}
		s.parts = r
//console.log('parts:'+r)

		// build the part table in the first symbol
		// and put the reverse pointers in the P: symbols
		s.p_s = []			// pointers to the parts
		while (1) {
			if (!s.ts_next) {
				s.part1 = first	// end of tune = end of part
				break
			}
			s = s.ts_next
			if (s.typ == C.part) {
				s.part1 = first		// reverse pointer
				v = s.text[0]		// 1st letter only
				for (i = 0; i < r.length; i++) {
					if (r[i] == v)
						first.p_s[i] = s
				}
			}
		}
	} // build_parts()

	// generate the grace notes
	function gen_grace(s) {
	    var	g, i, n, t, d, s2,
		next = s.next

		// before beat
		if (s.sappo) {
			d = C.BLEN / 16
		} else if ((!next || next.typ != C.note)
			&& s.prev && s.prev.typ == C.note) {
			d = s.prev.dur / 2

		// on beat
		} else {
			d = next.dur / 12
			if (!(d & (d - 1)))
				d = next.dur / 2	// no dot
			else
				d = next.dur / 3
			if (s.p_v.key.k_bagpipe)
				d /= 2
			next.time += d
			next.dur -= d
		}
//fixme: assume the grace notes in the sequence have the same duration
		n = 0
		for (g = s.grasym; g; g = g.next)
			n++
		d /= n * play_fac
		t = p_time
		for (g = s.grasym; g; g = g.next) {
			g.ptim = t
			g.pdur = d
			t += d
		}
	} // gen_grace()

	// change the tempo
	function set_tempo(s) {
	    var	i,
		d = 0,
		n = s.tempo_notes.length

		for (i = 0; i < n; i++)
			d += s.tempo_notes[i]
		return d * s.tempo / 60
	} // set_tempo()

	function set_variant(s, n) {
		s.var = 0
		n = n.match(/[1-8]-[2-9]|[1-9,.]|[^\s]+$/g)
		while (1) {
		    var	d = n.shift()
			if (!d)
				break
			if (d[1] == '-')
				for (i = d[0]; i <= d[2]; i++)
					s.var |= 1 << i
			else if (d >= '1' && d <= '9')
				s.var |= 1 << +d
		}
	} // set_variant()

	// add() main

	// if some chord stuff, set the accompaniment data
	if (cfmt.chord)
		abc2svg.chord(s, voice_tb, cfmt)

	// if %%playbeats, create the sounds
	if (cfmt.playbeats)
		def_beats()

	if (s.parts)
		build_parts(s)

	// set the time parameters
	rst_fac = play_fac
	while (s) {
		if (s.noplay) {			// in display macro sequence
			s = s.ts_next
			continue
		}

		dt = s.time - abc_time
		if (dt) {			// may go backwards after grace notes
			p_time += dt / play_fac
			abc_time = s.time
		}
		s.ptim = p_time

		if (s.part1)			// new part
			rst_fac = play_fac
		switch (s.typ) {
		case C.bar:
			if (s.time != b_tim) {
				b_tim = s.time
				b_typ = 0
			}
			if (s.text			// if new variant
			 && s.text[0] != '1') {
				if (b_typ & 1)
					break
				b_typ |= 1
				set_variant(s, s.text)
				play_fac = rst_fac
			}

			// right repeat
			if (s.bar_type[0] == ':') {
				if (b_typ & 2)
					break
				b_typ |= 2
			}

			// 1st variant
			if (s.text) {
			    if (s.text[0] == '1') {
				if (b_typ & 1)
					break
				b_typ |= 1
				set_variant(s, s.text)
				rst_fac = play_fac
			    }

			// left repeat
			} else if (s.bar_type.slice(-1) == ':') {
				if (b_typ & 4)
					break
				b_typ |= 4
				rst_fac = play_fac
			} else if (s.rbstop == 2) {
				if (b_typ & 8)
					break
				b_typ |= 8
				rst_fac = play_fac
			}
			break
		case C.grace:
			if (!s.time		// if before beat at start time
			 && !abc_time) {
				dt = 0
				if (s.sappo)
					dt = C.BLEN / 16
				else if (!s.next || s.next.typ != C.note)
					dt = d / 2
				abc_time -= dt
			}
			gen_grace(s)
			break
		case C.rest:
		case C.note:
			d = s.dur
			if (s.next && s.next.typ == C.grace) {
				dt = 0
				if (s.next.sappo)
					dt = C.BLEN / 16
				else if (!s.next.next || s.next.next.typ != C.note)
					dt = d / 2
				s.next.time -= dt
				d -= dt
			}
			d /= play_fac
			s.pdur = d
			v = s.v
			break
		case C.tempo:
			if (s.tempo)
				play_fac = set_tempo(s)
			break
		}
		s = s.ts_next
	} // loop
   } // add()
 } // return
} // ToAudio()

// play some next symbols
//
// This function is called to start playing.
// Playing is stopped on either
// - reaching the 'end' symbol (not played) or
// - reaching the end of tune or
// - seeing the 'stop' flag (user request).
//
// The po object (Play Object) contains the following items:
// - variables
//  - stop: stop flag
//		set by the user to stop playing
//  - s_cur: current symbol (next to play)
//		must be set to the first symbol to be played at startup time
//  - s_end: stop playing on this symbol
//		this symbol is not played. It may be null.
//  - conf
//    - speed: current speed factor
//		must be set to 1 at startup time
//    - new_conf: new speed factor
//		set by the user
// - internal variables
//  - i_p: current part index
//  - stim: start time
//  - repn: current variant number
//  - repv: bitmap of variants
//  - repe: repeat without :| at end of variant
//  - reps: repeat start symbol (|: ..)
//  - timouts: array of the current timeouts
//		this array may be used by the upper function in case of hard stop
//  - p_v: voice table used for MIDI control
// - methods
//  - onend: (optional)
//  - onnote: (optional)
//  - note_run: start playing a note
//  - get_time: return the time of the underlaying sound system
tm.play_next = function(po) {

	// return the symbol at the beginning of the sequence to be repeated
	function starep(s) {
//console.log('starep V:'+s.p_v.id+' '+s.time+' '+s.typ.nm)
		if (!po.reps) {				// if no start seen yet
			while (s.ts_prev) {
//console.log('   ..  V:'+s.p_v.id+' '+s.time+' '+s.typ.nm)
				if (s.part1			// if start of part
				 || (s.bar_type			// or left repeat bar
				  && (s.bar_type.slice(-1) == ':'
				   || s.bar_type == "||"))) {	// or abc-version < 2.2
//--fixme:pb ici:
//-- si s.part1, on croit que fin de part
//-- si note qui suit, elle est sautée
///-- il faudrait tester s2.time < s.time

//fixme:--fix
//					s = s.ts_next

					break	
				}
				s = s.ts_prev
			}
//console.log(' new rs V:'+s.p_v.id+' '+s.time+' '+s.typ.nm)
		} else {
			s = po.reps
//console.log(' old rs V:'+s.p_v.id+' '+s.time+' '+s.typ.nm)
		}

//fixme:--fix: cannot go backwards before the first part
//		while (s.time < 0)
//			s = s.ts_next

//fixme:--fix: as backwards, don't loose the first note of the part
//		if (s.part1)
//			s = s.next

//console.log('       V:'+s.p_v.id+' '+s.time+' '+s.typ.nm)
		po.reps = s
		return s
	} // starep()

	// a new variant and/or a right repeat bar has been found,
	// return where to go, or null if end of tune
	function var_next(s) {
	    var	i, n, s2

//console.log('     var_next '+s.bar_type+(s.text||'')
//+' repn: '+po.repn+(po.norp?' norp':'')
// +(s.var?(' s.var: 0x'+s.var.toString(16)):''))
//if(po.repn> 5){
//console.log('repn too big')
//return //undefined
//}

		// if forced variant and repeat bar, skip to the end of the part
		if (s.bar_type[0] == ':'		// if repeat bar
		 && po.norp)
			return po.norp			// next part

		// if last repeat bar, go back to the start of sequence
		// or continue playing
		if (s.bar_type[0] == ':'		// if this right repeat bar
		 && !s.var) {				// is the last one
			po.repn++			// repeat number
//console.log(' :| repn:'+po.repn+' repe:'+po.repe+' repv:0x'+po.repv.toString(16))
			if (po.repv) {
				if (po.repv & (1 << (po.repn - po.repe))) {
//console.log('** here is the repeat')
					s2 = starep(s)	// repeat the sequence
					if (!s2.var)	// if not the start of a variant
						return s2
					po.repn--
				}
			} else {
				if (s.bar_type[1] == ':') // 3 or 4 times
					i = s.bar_type[2] == ':' ? 4 : 3
				else
					i = 2
				if (po.repn <= i)
					return starep(s)
			}
			if (!s2) {
				po.repn = po.repe = 1	// repeat done
				po.repv = 0
				po.reps = s
				return s		// play the next notes
			}
		}

		n = 1 << po.repn			// repeat map
		if (!s2 && s.var) {			// if variant, not on seq restart
			po.repv |= s.var
			if (s.bar_type[0] != ':'	// if no :|
			 && s.text != '1')		// at end of variant
				po.repe = 0		// stop on last repeat
			if (s.var & n)			// the searched variant
				return s		// is just here
		}

		if (!s2 && s.bar_type[0] != ':') {	// if not a right repeat bar
			for ( ; s && s != po.s_end; s = s.next) {
//console.log('>> '+s.time+' '+s.typ.nm
// +(s.bar_type?(' '+s.bar_type+(s.text||'')):'')+(s.rbstop?' stop':'')
// +' n:'+po.repn)
				if (!s.bar_type)
					continue
				if (s.bar_type[0] == ':'
				 && !s.var)		//  end of sequence found
					break
				if ((s.var & n)		// twice the same number
				 || (!s.var
				  && s.rbstop == 2))	// explicit end of variant
					return s
			}
			if (!s || s == po.s_end)
				return //undefined
//console.log('-- loop end -- '+s.time+' '+s.typ.nm)
		}
		n = 1 << ++po.repn			// search the next variant

//--fixme:fix???
if(0)
		if (!(po.repv & n)) {
			po.repn = po.repe = 1		// repeat done
			po.repv = 0
			po.reps = s.next
//fixme: if new variant (s.var), var_next() must be recalled but without treating :|
//console.log('-- go next 2 -- '+s.time+' '+s.typ.nm
//+' nxt: '+s.next)
			return s.next			// play the next notes
		}
		if (s2)
			s = s2				// starep() already done
		else
			s = starep(s)			// restart the sequence
//console.log('     starep '+s.time+' '+s.typ.nm)
		if (!s.var)
			return s
		for ( ; s && s != po.s_end; s = s.next) {
//console.log('<< '+s.time+' '+s.typ.nm
// +(s.bar_type?(' '+s.bar_type+(s.text||'')):''))
			if (!s.bar_type)
				continue
			if (s.var) {
//console.log('    .var: '+s.var.toString(16)+' n:'+n.toString(16))
				po.repv |= s.var	// update the variants
				if (s.var & n)		// if variant found
					return s
			} else if (s.bar_type[0] == ':') { // if last repeat bar
//console.log('on '+s.time+' '+s.typ.nm
// +(s.bar_type?(' '+s.bar_type+(s.text||'')):''))
				return s.next		// play on
			}
		}
//		return //undefined			// end of sequence to play
	} // var_next()

	// handle a tie
	function do_tie(not_s, d) {
	    var	i, s2,
		s = not_s.s,
		C = tm.C,
		v = s.v,
		end_time = s.time + s.dur

//console.log('do_tie '+s.time+' '+s.typ.nm
// +(s.bar_type?(' '+s.bar_type+(s.text||'')):''))
		// search the end of the tie
		while (1) {
			s = s.ts_next
			if (!s || s.time > end_time)
				break
			if (s.bar_type) {
				if (s.var)
					po.repv |= s.var
				if (s.bar_type[0] == ':'	// right repeat
				 || s.var) {			// or new variant
					s2 = var_next(s)
					if (!s2)
						return d

					// if tie to symbol before right repeat
					if (s2.time < s.time) {
//console.log('      back '+s2.time+' '+s2.typ.nm)
						while (!s2.dur)
							s2 = s2.ts_next
						i = s2.notes.length
						while (--i >= 0) {
							note = s2.notes[i]
							if (note.midi == not_s.midi) {
								d += s2.pdur
									/ po.conf.speed
//if(!note.tie_ty)
//console.log('     back d: '+d)
								note.tie_b = 1 //true
								return note.tie_ty
									? do_tie(note, d)
									: d
							}
						}
					}
					if (s != s2) {
						s = s2
						end_time = s.time
//console.log('      tie skip '+s.time+' '+s.typ.nm
// +' endtim: '+end_time)
					}
				}
			}
			if (s.time < end_time
			 || !s.dur
			 || !s.ti2)			// if not end of tie
				continue

			i = s.notes.length
			while (--i >= 0) {
				note = s.notes[i]
				if (note.tie_s == not_s) {
					d += s.pdur / po.conf.speed
//if(!note.tie_ty)
//console.log('      d: '+d)
					return note.tie_ty ? do_tie(note, d) : d
				}
			}
		}
//console.log('        !!no end d: '+d+' ti: '+s.time+' '+end_time)
		return d
	} // do_tie()

	// set the MIDI controls up to now
	function set_ctrl(po, s2, t) {
	    var	i,
		p_v = s2.p_v,
		s = {
			subtype: "midictl",
			p_v: p_v,
			v: s2.v
		}

		for (i in p_v.midictl) { // MIDI controls at voice start time
			s.ctrl = +i
			s.val = p_v.midictl[+i]
			po.midi_ctrl(po, s, t)
		}

		po.p_v[s2.v] = 1 //true	// synchronization done

		for (s = p_v.sym; s != s2; s = s.next) {
			if (s.subtype == "midictl")
				po.midi_ctrl(po, s, t)
			else if (s.subtype == "midiprog")
				po.midi_prog(po, s)
		}

		// if no %%MIDI, set 'grand acoustic piano' as the instrument
		i = po.v_c[s2.v]
		if (i == undefined)
			po.v_c[s2.v] = i = s2.v < 9 ? s2.v : s2.v + 1
		if (po.c_i[i] == undefined)
			po.c_i[i] = 0	// piano
	} // set_ctrl()

    // start and continue to play
    function play_cont(po) {
    var	d, i, st, m, note, g, s2, s3, t, maxt, now, flg,
	C = tm.C,
	s = po.s_cur

	if (po.stop) {
		if (po.onend)
			po.onend(po.repn)
		return
	}

	while (s.noplay) {
		s = s.ts_next
		if (!s || s == po.s_end) {
			if (po.onend)
				po.onend(po.repn)
			return
		}
	}
	t = po.stim + s.ptim / po.conf.speed	// start time
	now = po.get_time(po)

	// if speed change, shift the start time
	if (po.conf.new_speed) {
		po.stim = now - (now - po.stim) *
					po.conf.speed / po.conf.new_speed
		po.conf.speed = po.conf.new_speed
		po.conf.new_speed = 0
		t = po.stim + s.ptim / po.conf.speed
	}

	maxt = t + po.tgen		// max time = now + 'tgen' seconds
	po.timouts = []

//var cnt=40

	while (1) {
//console.log('      curr '+s.time+' '+s.typ.nm
// +(s.bar_type?(' '+s.bar_type+(s.text||'')):'')
// +' repv: 0x'+po.repv.toString(16)+' repn: '+po.repn
// +' part1:'+s.part1)
//if(--cnt<0)toto()
		if (!po.p_v[s.v])		// if new voice
			set_ctrl(po, s, t)	// set the MIDI controls

//		s2 = null
		switch (s.typ) {
		case C.bar:
			if (s.bar_type.slice(-1) == ':'	// left repeat
			 && s.bar_type[0] != ':') {
				po.repn = po.repe = 1
				po.repv = 0
				po.reps = s
			}
// (there may be "|:1 .." and even ".. :|:1 ..")
			if (s.bar_type[0] == ':'	// right repeat
			 || s.var) {			// new variant
				s2 = var_next(s)
				if (!s2)		// end of tune
					s = po.s_end
//fixme 24-02-22
				if (s2 == s)
					s2 = null	// no jump
			}
			break
		case C.block:
//console.log('bl '+s.subtype+' c: '+(s.chn||'-')+' p: '+(s.instr||'-'))
			if (s.subtype == "midictl")
				po.midi_ctrl(po, s, t)
			else if (s.subtype == "midiprog")
				po.midi_prog(po, s)
			break
		case C.grace:
			for (g = s.grasym; g; g = g.next) {
				d = g.pdur / po.conf.speed
				for (m = 0; m <= g.nhd; m++) {
					note = g.notes[m]
					if (!note.noplay)
						po.note_run(po, g,
							note.midi,
							t + g.ptim - s.ptim,
//fixme: there may be a tie...
							d)
				}
			}
			break
		case C.note:
		case C.rest:
//console.log('oo '+s.time)
			d = s.pdur / po.conf.speed
			if (s.typ == C.note) {
				for (m = 0; m <= s.nhd; m++) {
					note = s.notes[m]
					if (note.tie_s	 // end of tie
					 || note.noplay) // (%%voicecombine)
						continue // already generated
					if (note.tie_b) { // if tie before :|
						note.tie_b = 0 //false
						continue
					}

					// (don't change the repeat variables)
					po.note_run(Object.create(po),
						s,
						note.midi,
						t,
						note.tie_ty ?
							do_tie(note, d) : d)
				}
			}

			// follow the notes/rests while playing
			if (po.onnote && s.istart) {
				i = s.istart
				st = (t - now) * 1000
				po.timouts.push(setTimeout(po.onnote,
							st, i, 1 /*true*/))
				if (d > 2) // problem when loop on one long note
					d -= .1
				setTimeout(po.onnote,
					st + d * 1000, i, 0 /*false*/)
			}
			break
		}

		// handle the repeats		
		if (s != po.s_end) {
//if(s2)
//console.log('        jump '+s2.time+' '+s2.typ.nm
// +(s2.bar_type?(' '+s2.bar_type+(s2.text||'')):''))

//fixme:--fix - no
//if(0)
//fixme:--fix
//			if (s2 && s2 != s) {	// if time jump
			if (s2 && s2.time < s.time) {	// if jump backwards

//console.log('  - jump -')
				po.stim += (s.ptim - s2.ptim) / po.conf.speed
//test1
//					flg = s2.time < s.time	// go backwards
				s = s2
				t = po.stim + s.ptim / po.conf.speed
				s2 = null
			}

//fixme:--fix: jumping backwards often gives a part start
			  else

//console.log('   -- '+s.time+' '+s.typ.nm+' po.i_p:'+po.i_p
// +' part1:'+s.part1)

			if (s.part1) {		// if end of part
//				 && po.i_p != undefined) {
				po.repn = po.repe = 1
				po.repv = 0
//				po.reps = null
				s2 = s.part1.p_s[++po.i_p]	// next part

//if(s2)
//console.log('        part1 '+s.time+' '+s.typ.nm+' -> '+s2.time+' '+s2.typ.nm)

				// handle "2A"
				if (!s2 && !isNaN(+s.part1.parts[po.i_p])) {
//console.log('      start variant: '+s.part1.parts[po.i_p]
//+s.part1.parts[po.i_p+1])
					po.repn = +s.part1.parts[po.i_p]
					s2 = s.part1.p_s[++po.i_p]	// part
					po.norp = s.part1.p_s[po.i_p + 1]
							// next part for no repeat
				}
				if (!s2)
					s = po.s_end
//fixme: 24-02-22
else
	s2 = s2.ts_next
			}
		}

		// handle the flow marks
		if (s != po.s_end) {
//if(s.timark)
//console.log('   timark '+s.time+' '+s.typ.nm+' '+s.timark)
//else
//console.log(' '+s.time+' '+s.typ.nm)
			switch (s.timark) {		// any flow mark?
			case "coda":
				if (!po.tidone)		// if no D.C./D.S. yet
					break
				for (s2 = s.next; s2 != po.s_end; s2 = s2.next) {
					if (s2.timark == "coda")
						break
				}
				if (s2 == po.s_end)	// if no 2nd coda
					s2 = s		// just continue
//fixme 24/02/22
//				else
//					s2 = s2.ts_prev
				else
					po.tidone = 0
				break
			case "fine":
				if (po.tidone)		// if 2nd time
					s = po.s_end	// stop here
				break
//			case "sgno":
//				break
			case "dcap":
			case "dsgn":
				if (po.tidone) {	// if 2nd time
					s = po.s_end	// stop here
					break
				}
				po.tidone++
				if (s.timark[1] == 's') {	// dal segno
//?? 24-02-09
//				for (s2 = s.prev; s2.prev; s2 = s2.prev) {
					for (s2 = s.ts_prev; s2.ts_prev; s2 = s2.ts_prev) {
						if (s2.timark == "sgno") {
							s2 = s2.ts_prev
							break
						}
					}
				} else {			// da capo
//					for (s2 = s.prev; s2.prev; s2 = s2.prev) {
//						if (!s2.time)
//							break
					s2 = s.p_v.sym
					while (s2.time < 0) // skip the starting beats
						s2 = s2.ts_next
//console.log(' capo '+s2.time+' '+s2.typ.nm)
				}
				break
			}
		}
		if (s2) {				// if time jump
			po.stim += (s.ptim - s2.ptim) / po.conf.speed
			s = s2
			t = po.stim + s.ptim / po.conf.speed
//			po.reps = s
			s2 = null
		}
//fixme 24/02/22
		else
		    while (1) {
			if (!s || s == po.s_end || !s.ts_next) {
				if (po.onend)
					setTimeout(po.onend,
						(t - now + d) * 1000,
						po.repn)
				po.s_cur = s
				po.reps = null
				return
			}
			s = s.ts_next
			if (!s.noplay)
				break
		}
		t = po.stim + s.ptim / po.conf.speed // next time
		if (t > maxt)
			break
	}
	po.s_cur = s

	// delay before next sound generation
	po.timouts.push(setTimeout(play_cont,
				(t - now) * 1000
					- 300,	// wake before end of playing
				po))
    } // play_cont()

    // search the index in the parts
    function get_part(po) {
    var	s, i, s_p

	for (s = po.s_cur; s; s = s.ts_prev) {
		if (s.parts) {
			po.i_p = -1
			return
		}
		s_p = s.part1
		if (!s_p || !s_p.p_s)
			continue
		for (i = 0; i < s_p.p_s.length; i++) {
			if (s_p.p_s[i] == s) {
				po.i_p = i	// index in the parts
				return
			}
		}
	}
    } // get_part()

    // --- play_next ---
	get_part(po)

	po.stim = po.get_time(po) + .3	// start time + 0.3s
			- po.s_cur.ptim * po.conf.speed
	po.p_v = []			// voice table for the MIDI controls
	if (po.s_cur.time <= 0) {	// if start from the beginning
		po.repn = po.repe = 1	// new repeat start
		po.repv = 0
//		po.reps = po.s_cur
		po.reps = null		// (not on %%staves)
		po.tidone = 0		// flow repeat times (0 or 1)
	}
	play_cont(po)			// start playing
} // play_next()
